/*-
* Copyright (c) 2008 Damien Bergamini <damien.bergamini@free.fr>
*
* Permission to use, copy, modify, and distribute this software for any
* purpose with or without fee is hereby granted, provided that the above
* copyright notice and this permission notice appear in all copies.
*
* THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
* WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
* ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
* WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
* ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
* OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef _HMAC_H_
#define _HMAC_H_

#include "amd_md5.h"
#include "amd_sha1.h"

typedef struct _AMD_HMAC_MD5_CTX {
  AMD_MD5_CTX		ctx;
  uint8_t	key[AMD_MD5_BLOCK_LENGTH];
  unsigned int		key_len;
} AMD_HMAC_MD5_CTX;

typedef struct _AMD_HMAC_SHA1_CTX {
  AMD_SHA1_CTX	ctx;
  uint8_t	key[AMD_SHA1_BLOCK_LENGTH];
  unsigned int		key_len;
} AMD_HMAC_SHA1_CTX;

#if 0
typedef struct _AMD_HMAC_SHA256_CTX {
  SHA2_CTX	ctx;
  uint8_t	key[SHA256_BLOCK_LENGTH];
  unsigned int		key_len;
} AMD_HMAC_SHA256_CTX;
#endif

void	 AMD_HMAC_MD5_Init(AMD_HMAC_MD5_CTX *, const uint8_t *, unsigned int);

void	 AMD_HMAC_MD5_Update(AMD_HMAC_MD5_CTX *, const uint8_t *, unsigned int);

void	 AMD_HMAC_MD5_Final(uint8_t [AMD_MD5_DIGEST_LENGTH], AMD_HMAC_MD5_CTX *);

void	 AMD_HMAC_SHA1_Init(AMD_HMAC_SHA1_CTX *, const uint8_t *, unsigned int);

void	 AMD_HMAC_SHA1_Update(AMD_HMAC_SHA1_CTX *, const uint8_t *, unsigned int);

void	 AMD_HMAC_SHA1_Final(uint8_t [AMD_SHA1_DIGEST_LENGTH], AMD_HMAC_SHA1_CTX *);

#if 0
void	 HMAC_SHA256_Init(HMAC_SHA256_CTX *, const uint8_t *, unsigned int);

void	 HMAC_SHA256_Update(HMAC_SHA256_CTX *, const uint8_t *, unsigned int);

void	 HMAC_SHA256_Final(uint8_t [SHA256_DIGEST_LENGTH], HMAC_SHA256_CTX *);
#endif

#endif	/* _HMAC_H_ */
