


#if (_MSC_VER >= 1700) && !defined(_USING_V110_SDK71_)
#include <winapifamily.h>
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#define DISABLE_DLLMAIN
#elif defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PC_APP)
#define DISABLE_DLLMAIN
#endif
#endif

#ifndef DISABLE_DLLMAIN

#ifdef WIN32
#include "ortp-config-win32.h"
#else
#include "ortp-config.h"
#endif
#include "ortp/ortp.h"	

#if defined(HAVE_OPENSSL_SSL_H) && defined(HAVE_SRTP)
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/conf.h>
#endif

#ifdef EXTERNAL_LOGGER
#include "logger.h"
#else
#define	RegisterLog(logVar, logString) do { } while (0)
#define	UnregisterLog(logVar, logString) do { } while (0)
#endif

extern DWORD dwoRTPLogLevel;

/* Check for memory leaks */
//#define ORTP_MEMORY_DEBUG
#ifdef ORTP_MEMORY_DEBUG
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>

static int __cdecl MyAllocHook(
  int      nAllocType,
  void   * pvData,
  size_t   nSize,
  int      nBlockUse,
  long     lRequest,
  const unsigned char * szFileName,
  int      nLine
)
{
  char *operation[] = { "", "allocating", "re-allocating", "freeing" };
  char *blockType[] = { "Free", "Normal", "CRT", "Ignore", "Client" };

  char tmp[200];
  WCHAR wUnicode[500];
  int size;

  snprintf(tmp, sizeof(tmp), "oRTP Memory operation in %s, line %d: %s a %d-byte '%s' block (# %ld)\n",
    szFileName, nLine, operation[nAllocType], (int)nSize,
    blockType[nBlockUse], lRequest);
  size = MultiByteToWideChar(CP_UTF8, 0, tmp, -1, wUnicode, 500 * 2);

  if (nBlockUse == _CRT_BLOCK)   // Ignore internal C runtime library allocations
    return(1);

  _ASSERT((nAllocType > 0) && (nAllocType < 4));
  _ASSERT((nBlockUse >= 0) && (nBlockUse < 5));

  if (pvData != NULL) {
    char tmp[200];
    WCHAR wUnicode[500];
    int size;

    snprintf(tmp, sizeof(tmp), "oRTP at %p\n", pvData);
    size = MultiByteToWideChar(CP_UTF8, 0, tmp, -1, wUnicode, 500 * 2);

    OutputDebugString(wUnicode);
  }

  return(1);         // Allow the memory operation to proceed
}

static int __cdecl MyReportHook(
  int   nRptType,
  char *szMsg,
  int  *retVal
)
{
  char *RptTypes[] = { "Warning", "Error", "Assert" };

  char tmp[200];
  WCHAR wUnicode[500];
  int size;

  snprintf(tmp, sizeof(tmp), "oRTP MyReportHook %s: %s", RptTypes[nRptType], szMsg);
  size = MultiByteToWideChar(CP_UTF8, 0, tmp, -1, wUnicode, 500 * 2);

  OutputDebugString(wUnicode);

  retVal = 0;

  return(1);         // Allow the report to be made as usual

}

static _CrtMemState s1, s2, s3;

static void dumpleak(void) {
  _CrtDumpMemoryLeaks();
}

#endif

void init_random_number_generator(void);

BOOL WINAPI DllMain(
  HINSTANCE hinstDLL,	// handle to DLL module
  DWORD fdwReason,		// reason for calling function
  LPVOID lpReserved		// reserved
)
{
  WORD	wVersionRequested;
  WSADATA wsaData;

  // Perform actions based on the reason for calling.
  switch (fdwReason)
  {
  case DLL_PROCESS_ATTACH:

    OutputDebugString(TEXT("--> dll_entry.c - oRTP.dll - DLL_PROCESS_ATTACH()\n"));
#ifdef ORTP_MEMORY_DEBUG
    /* _crtBreakAlloc = 823; */
    _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_CHECK_CRT_DF | _CRTDBG_CHECK_ALWAYS_DF | _CRTDBG_LEAK_CHECK_DF || _CRTDBG_DELAY_FREE_MEM_DF);
    _CrtSetAllocHook(MyAllocHook);
    _CrtSetReportHook(MyReportHook);
    _CrtSetReportFile(_CRT_WARN, _CRTDBG_FILE_STDOUT);
    _CrtSetReportFile(_CRT_ERROR, _CRTDBG_FILE_STDOUT);
    _CrtSetReportFile(_CRT_ASSERT, _CRTDBG_FILE_STDOUT);

    atexit(dumpleak);
    _CrtMemCheckpoint(&s1);
#endif		 
    wVersionRequested = MAKEWORD(2, 0);

    if (WSAStartup(wVersionRequested, &wsaData) != 0)
    {
      return FALSE;
    }


    RegisterLog(&dwoRTPLogLevel, "LOG_ORTP");
    init_random_number_generator();
    break;

  case DLL_THREAD_ATTACH:

    break;

  case DLL_THREAD_DETACH:
#if defined(HAVE_OPENSSL_SSL_H) && defined(HAVE_SRTP)
#if OPENSSL_VERSION_NUMBER < 0x10100000L
    //CRYPTO_cleanup_all_ex_data();
    ERR_remove_state(0);
    ERR_remove_thread_state(NULL);
#else
    OPENSSL_thread_stop();
#endif
#endif
    break;

  case DLL_PROCESS_DETACH:

    OutputDebugString(TEXT("--> dll_entry.c - oRTP.dll - Detaching\n"));

    ortp_exit();
#ifdef ORTP_MEMORY_DEBUG
    UnregisterLog(&dwoRTPLogLevel, "LOG_ORTP");
    _CrtMemDumpAllObjectsSince(&s1);
    _CrtMemCheckpoint(&s2);
    if (_CrtMemDifference(&s3, &s1, &s2))
      _CrtMemDumpStatistics(&s3);
#endif
    break;
  }

  return TRUE;  // Successful DLL_PROCESS_ATTACH.
}

#endif
