/*
  The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef rtpsession_priv_h
#define rtpsession_priv_h

#include "ortp/rtpsession.h"

typedef enum {
	RTP_SESSION_RECV_SYNC=1,	/* the rtp session is synchronising in the incoming stream */
	RTP_SESSION_FIRST_PACKET_DELIVERED=1<<1,
	RTP_SESSION_ENABLE_SENDRECV=1<<2, /* the bundle is enabled to allow send and recv from bundle thread */
	RTP_SESSION_BLOCKING_MODE=1<<3, /* in blocking mode */
	RTP_SESSION_RECV_NOT_STARTED=1<<4,	/* the application has not started to try to recv */
	RTP_SESSION_SEND_NOT_STARTED=1<<5,  /* the application has not started to send something */
	RTP_SESSION_USING_TRANSPORT=1<<10,
	RTP_SESSION_DTLS_START=1<<11
}RtpSessionFlags;

#define RFC5764BIS_BAD -1
#define RFC5764BIS_UNKNOWN 0
#define RFC5764BIS_STUN 1
#define RFC5764BIS_ZRTP 2
#define RFC5764BIS_DTLS 3
#define RFC5764BIS_TURN 4
#define RFC5764BIS_RTP  5
#define RFC5764BIS_RTCP 6

#define rtp_session_using_transport(s, stream) (((s)->flags & RTP_SESSION_USING_TRANSPORT) && (s->stream.tr != 0))

int rtp_session_update_payload_type(RtpSession * session, int pt);
void rtp_putq(queue_t *q, mblk_t *mp);
mblk_t * rtp_getq(RtpSession * session, queue_t *q, uint32_t ts, int *rejected, mblk_t **last_dropped);
int rtp_session_rtp_recv(RtpSession * session, int idx);
int rtp_session_rtcp_recv(RtpSession * session, int idx);
int rtp_session_rtp_send (RtpSession * session, mblk_t * m);
int rtp_session_rtcp_send (RtpSession * session, mblk_t * m);

RtpSession *rtp_session_demux_media(RtpBundle *bundle, RtpSession *session, mblk_t *mp, struct sockaddr *addr, socklen_t addrlen);
int rtp_session_rtp_parse(RtpSession *session, mblk_t *mp, struct sockaddr *addr, socklen_t addrlen, int initial_packet_size, int *rtp_overhead);
void rtp_session_rtcp_parse(RtpSession *session, mblk_t *mp);
void rtp_session_notify_inc_rtcp(RtpSession *session, mblk_t *m);

int close_socket(RtpSession *rtpsession, ortp_socket_t sock);

#if defined(WIN32) || defined(_WIN32_WCE)
#define sockisvalid(s) (s != ~0)
#else
#define sockisvalid(s) ((s) >= 0)
#endif

#endif
