 /*
  The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#ifdef HAVE_CONFIG_H
#include "ortp-config.h"
#endif

#ifndef _WIN32_WCE
#include <errno.h>
#endif

#include <assert.h>

#if defined(WIN32) || defined(_WIN32_WCE)
#include <winsock2.h>
#include <stdlib.h>
/* #include <io.h> */
#include <time.h>
#include <ctype.h> /*for isdigit() */
#else

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h> 
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <net/if.h>

#endif

#include "ortp/stun_udp.h"
#include "ortp/stun.h"
#include "ortp/turn.h"
#include "ortp/ortp.h"

void stunDecodeXorAddress(const StunIPAddress *xorAddress, StunIPAddress *destAddr, StunTrId *stunTrId)
{
  uint32_t cookie = 0x2112A442;
  uint16_t cookie16 = 0x2112A442 >> 16;
  destAddr->family = xorAddress->family;
  destAddr->addr.ipv4.port = xorAddress->addr.ipv4.port^cookie16;
  if (xorAddress->family == IPv4Family) {
    destAddr->addr.ipv4.addr = xorAddress->addr.ipv4.addr^cookie;
  } else if (xorAddress->family == IPv6Family) {
    uint8_t xorId[16];
    xorId[0]=0x21;
    xorId[1]=0x12;
    xorId[2]=0xA4;
    xorId[3]=0x42;
    memcpy(xorId+4, stunTrId, 12);
    destAddr->addr.ipv6.addr[0] = xorAddress->addr.ipv6.addr[0]^xorId[0];
    destAddr->addr.ipv6.addr[1] = xorAddress->addr.ipv6.addr[1]^xorId[1];
    destAddr->addr.ipv6.addr[2] = xorAddress->addr.ipv6.addr[2]^xorId[2];
    destAddr->addr.ipv6.addr[3] = xorAddress->addr.ipv6.addr[3]^xorId[3];
    destAddr->addr.ipv6.addr[4] = xorAddress->addr.ipv6.addr[4]^xorId[4];
    destAddr->addr.ipv6.addr[5] = xorAddress->addr.ipv6.addr[5]^xorId[5];
    destAddr->addr.ipv6.addr[6] = xorAddress->addr.ipv6.addr[6]^xorId[6];
    destAddr->addr.ipv6.addr[7] = xorAddress->addr.ipv6.addr[7]^xorId[7];
    destAddr->addr.ipv6.addr[8] = xorAddress->addr.ipv6.addr[8]^xorId[8];
    destAddr->addr.ipv6.addr[9] = xorAddress->addr.ipv6.addr[9]^xorId[9];
    destAddr->addr.ipv6.addr[10] = xorAddress->addr.ipv6.addr[10]^xorId[10];
    destAddr->addr.ipv6.addr[11] = xorAddress->addr.ipv6.addr[11]^xorId[11];
    destAddr->addr.ipv6.addr[12] = xorAddress->addr.ipv6.addr[12]^xorId[12];
    destAddr->addr.ipv6.addr[13] = xorAddress->addr.ipv6.addr[13]^xorId[13];
    destAddr->addr.ipv6.addr[14] = xorAddress->addr.ipv6.addr[14]^xorId[14];
    destAddr->addr.ipv6.addr[15] = xorAddress->addr.ipv6.addr[15]^xorId[15];
  } else {
    ortp_error("stun: stunDecodeXorAddress/stunEncodeXorAddress missing family");
  }
}

void stunEncodeXorAddress(const StunIPAddress *origAddress, StunIPAddress *xorAddress, StunTrId *stunTrId)
{
  stunDecodeXorAddress(origAddress, xorAddress, stunTrId);
}

#ifndef INET6_ADDRSTRLEN
#define INET6_ADDRSTRLEN 65
#endif

char *stunIPAddressStr(const StunIPAddress *addr, char *tmp, size_t size_of_tmp)
{
   tmp[0] = 0;
   if(addr->family == 0) {
     ortp_warning("stunIPAddressStr: missing family info");
   }
   if(addr->family == IPv4Family) {
     char ipbuf[INET6_ADDRSTRLEN];
     uint32_t val_addr = htonl(addr->addr.ipv4.addr);
     ortp_inet_ntop(AF_INET, &val_addr, ipbuf, INET6_ADDRSTRLEN);
     snprintf(tmp, size_of_tmp, "[%s]:%d", ipbuf, addr->addr.ipv4.port);
   } else if(addr->family == IPv6Family) {
     char ipbuf[INET6_ADDRSTRLEN];
     ortp_inet_ntop(AF_INET6, &addr->addr.ipv6.addr, ipbuf, INET6_ADDRSTRLEN);
     snprintf(tmp, size_of_tmp, "[%s]:%d", ipbuf,
           addr->addr.ipv4.port);
   }
   return tmp;
}

char *
stunIPAddress_ntop(const StunIPAddress *addr, char *tmp, size_t size_of_tmp)
{
   tmp[0] = 0;
   if(addr->family == 0) {
     ortp_warning("stunIPAddressStr: missing family info");
   }
   if(addr->family == IPv4Family) {
     uint32_t val_addr = htonl(addr->addr.ipv4.addr);
     ortp_inet_ntop(AF_INET, &val_addr, tmp, size_of_tmp);
   } else if(addr->family == IPv6Family) {
     ortp_inet_ntop(AF_INET6, &addr->addr.ipv6.addr, tmp, size_of_tmp);
   }
   return tmp;
}

static bool_t 
stunParseAtrAddress( char* body, unsigned int hdrLen,  StunIPAddress *result )
{
   if ( hdrLen != 8 && hdrLen != 20 )
   {
      ortp_error("stun: hdrLen wrong for Address");
      return FALSE;
   }
   body++; /* pad */
   result->family = *body++;
   if (result->family == IPv4Family)
   {
      uint16_t nport;
      uint32_t naddr;
      memcpy(&nport, body, 2); body+=2;
      result->addr.ipv4.port = ntohs(nport);

      memcpy(&naddr, body, 4); body+=4;
      result->addr.ipv4.addr = ntohl(naddr);
      return TRUE;
   }
   else if (result->family == IPv6Family)
   {
      uint16_t nport;
      memcpy(&nport, body, 2); body+=2;
      result->addr.ipv6.port = ntohs(nport);
      memcpy(&result->addr.ipv6.addr[0], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[1], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[2], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[3], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[4], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[5], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[6], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[7], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[8], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[9], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[10], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[11], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[12], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[13], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[14], body, 1); body+=1;
      memcpy(&result->addr.ipv6.addr[15], body, 1); body+=1;
     return TRUE;
   }
   else
   {
      ortp_error("stun: bad address family: %i", result->family);
   }
	
   return FALSE;
}

static bool_t 
stunParseAtrChangeRequest( char* body, unsigned int hdrLen,  StunAtrChangeRequest *result )
{
   if ( hdrLen != 4 )
   {
     /* ortp_error("stun: hdr length = %i expecting %i",hdrLen, sizeof(result)); */
		
      ortp_error("stun: Incorrect size for SA_CHANGEREQUEST");
      return FALSE;
   }
   else
   {
      memcpy(&result->value, body, 4);
      result->value = ntohl(result->value);
      return TRUE;
   }
}

static bool_t 
stunParseAtrError( char* body, unsigned int hdrLen,  StunAtrError *result )
{
   if ( hdrLen < 4 || hdrLen >= 128+4)
   {
      ortp_error("stun: Incorrect size for SA_ERRORCODE");
      return FALSE;
   }
   else
   {
      memcpy(&result->pad, body, 2); body+=2;
      result->pad = ntohs(result->pad);
      result->errorClass = *body++;
      result->number = *body++;
		
      result->sizeReason = hdrLen - 4;
      memcpy(&result->reason, body, result->sizeReason);
      result->reason[result->sizeReason] = 0;
      return TRUE;
   }
}

static bool_t 
stunParseAtrUnknown( char* body, unsigned int hdrLen,  StunAtrUnknown *result )
{
   if ( hdrLen >= sizeof(*result) )
   {
      ortp_error("stun: Incorrect size for SA_UNKNOWNATTRIBUTE");
      return FALSE;
   }
   else
   {
      int i;
      if (hdrLen % 4 != 0) return FALSE;
      result->numAttributes = hdrLen / 4;
      for (i=0; i<result->numAttributes; i++)
      {
         memcpy(&result->attrType[i], body, 2); body+=2;
         result->attrType[i] = ntohs(result->attrType[i]);
      }
      return TRUE;
   }
}

static bool_t 
stunParseAtrString( char* body, unsigned int hdrLen,  StunAtrString *result )
{
   if ( hdrLen >= STUN_MAX_STRING )
   {
      ortp_error("stun: String is too large");
      return FALSE;
   }
   else
   {
      result->sizeValue = hdrLen;
      memcpy(&result->value, body, hdrLen);
      result->value[hdrLen] = 0;
      return TRUE;
   }
}


static bool_t 
stunParseAtrIntegrity( char* body, unsigned int hdrLen,  StunAtrIntegrity *result )
{
   if ( hdrLen != 20)
   {
      ortp_error("stun: SA_MESSAGEINTEGRITY must be 20 bytes");
      return FALSE;
   }
   else
   {
      memcpy(&result->hash, body, hdrLen);
      return TRUE;
   }
}

static bool_t 
turnParseAtrChannelNumber( char* body, unsigned int hdrLen,  TurnAtrChannelNumber *result )
{
   if ( hdrLen >= sizeof(*result) )
   {
      ortp_error("stun: Incorrect size for TA_CHANNELNUMBER");
      return FALSE;
   }
   else
   {
      if (hdrLen % 4 != 0) return FALSE;
      memcpy(&result->channelNumber, body, 2);
      body+=2;
      result->channelNumber = ntohs(result->channelNumber);
      memcpy(&result->rffu, body, 2);
      body+=2;
      result->rffu = ntohs(result->rffu);
      return TRUE;
   }
}

static bool_t 
turnParseAtrLifetime( char* body, unsigned int hdrLen,  TurnAtrLifetime *result )
{
   if ( hdrLen != sizeof(*result) )
   {
      ortp_error("stun: Incorrect size for TA_LIFETIME");
      return FALSE;
   }
   else
   {
      memcpy(&result->lifetime, body, 4);
      result->lifetime = ntohl(result->lifetime);
      return TRUE;
   }
}

static bool_t
turnParseAtrRequestAddressFamily( char* body, unsigned int hdrLen, Rfc6156AtrRequestAddressFamily *result )
{
   if ( hdrLen != 4 )
   {
      ortp_error("stun: Incorrect size for RFC6156A_REQUESTADDRESSFAMILY");
      return FALSE;
   }
   result->family = *body++;
   result->pad1 = *body++;
   result->pad2 = *body++;
   result->pad3 = *body++;
   return TRUE;
}

static bool_t 
turnParseAtrEvenPort( char* body, unsigned int hdrLen, TurnAtrEvenPort *result )
{
   if ( hdrLen != 4 )
   {
      ortp_error("stun: Incorrect size for TA_EVENPORT");
      return FALSE;
   }
   result->value = *body++;
   result->pad1 = *body++;
   result->pad2 = *body++;
   result->pad3 = *body++;
   return TRUE;
}

static bool_t 
turnParseAtrData( char* body, unsigned int hdrLen,  TurnAtrData *result )
{
   if ( hdrLen >= 1500 )
   {
      ortp_error("stun: Incorrect size for TA_DATA");
      return FALSE;
   }
   else
   {
      result->sizeValue = hdrLen;
      memcpy(&result->value, body, hdrLen);
      result->value[hdrLen] = 0;
      return TRUE;
   }
}

static bool_t 
turnParseAtrRequestedTransport( char* body, unsigned int hdrLen,  TurnAtrRequestedTransport *result )
{
   if ( hdrLen != 4 )
   {
      ortp_error("stun: Incorrect size for TA_REQUESTEDTRANSPORT");
      return FALSE;
   }
   result->proto = *body++;
   result->pad1 = *body++;
   result->pad2 = *body++;
   result->pad3 = *body++;
   return TRUE;
}

#if defined(htonq)
#elif defined(ORTP_BIGENDIAN)
#define htonq(n) n
#define ntohq(n) n
#else /* little endian */
static inline uint64_t
htonq (uint64_t v)
{
  return htonl ((uint32_t) (v >> 32))
    | (uint64_t) htonl ((uint32_t) v) << 32;
}
static inline uint64_t
ntohq (uint64_t v)
{
  return ntohl ((uint32_t) (v >> 32))
    | (uint64_t) ntohl ((uint32_t) v) << 32;
}
#endif /* little endian */

static bool_t 
turnParseAtrReservationToken( char* body, unsigned int hdrLen,  TurnAtrReservationToken *result )
{
  if ( hdrLen != 8 )
  {
    ortp_error("stun: Incorrect size for TA_RESERVATIONTOKEN");
    return FALSE;
  }
  memcpy(&result->value, body, 8);
  result->value = ntohq(result->value);
  return TRUE;
}

static bool_t 
stunParseAtrFingerprint( char* body, unsigned int hdrLen, StunAtrFingerprint *result )
{
  if ( hdrLen != 4 )
  {
    ortp_error("stun: Incorrect size for SA_FINGERPRINT");
    return FALSE;
  }

  memcpy(&result->fingerprint, body, 4);
  result->fingerprint = ntohl(result->fingerprint);
  return TRUE;
}

static bool_t 
iceParseAtrPriority( char* body, unsigned int hdrLen, IceAtrPriority *result )
{
  if ( hdrLen != 4 )
  {
    ortp_error("stun: Incorrect size for ICEA_PRIORITY");
    return FALSE;
  }

  memcpy(&result->priority, body, 4);
  result->priority = ntohl(result->priority);
  return TRUE;
}

static bool_t 
iceParseAtrIceControll( char* body, unsigned int hdrLen, IceAtrIceControll *result )
{
  if ( hdrLen != 8 )
  {
    ortp_error("stun: Incorrect size for ICEA_ICECONTROLLED/ICEA_ICECONTROLLING");
    return FALSE;
  }
  memcpy(&result->value, body, 8);
  result->value = ntohq(result->value);
  return TRUE;
} 

static bool_t
iceParseAtrNetworkCost( char* body, unsigned int hdrLen, IceAtrNetworkCost *result )
{
  if ( hdrLen != 4 )
  {
    ortp_error("stun: Incorrect size for ICE_NETWORK_COST");
    return FALSE;
  }

  memcpy(&result->networkid, body, 2);
  body+=2;
  result->networkid = ntohs(result->networkid);
  memcpy(&result->networkcost, body, 2);
  body+=2;
  result->networkcost = ntohs(result->networkcost);
  return TRUE;
}

bool_t
stunParseMessage( char* buf, unsigned int bufLen, StunMessage *msg)
{
   char tmp[512];
   char* body;
   unsigned int size;
	 ortp_debug("stun: Received stun message: %i bytes", bufLen);
   memset(msg, 0, sizeof(*msg));
	
   if (sizeof(StunMsgHdr) > bufLen)
   {
      ortp_warning("stun: message too short");
      return FALSE;
   }

   memcpy(&msg->msgHdr, buf, sizeof(StunMsgHdr));
   msg->msgHdr.msgType = ntohs(msg->msgHdr.msgType);
   msg->msgHdr.msgLength = ntohs(msg->msgHdr.msgLength);

   if (msg->msgHdr.msgLength + sizeof(StunMsgHdr) != bufLen)
   {
      ortp_warning("stun: Message header length doesn't match message size: %i - %i", msg->msgHdr.msgLength, bufLen);
      return FALSE;
   }

   body = buf + sizeof(StunMsgHdr);
   size = msg->msgHdr.msgLength;
	
   /*ortp_message("stun: bytes after header = %i", size); */
	
   while ( size > 0 )
   {
      /* !jf! should check that there are enough bytes left in the buffer */
		
      StunAtrHdr* attr = (StunAtrHdr*)body; /*reinterpret_cast<StunAtrHdr*>(body);*/
		
      unsigned int attrLen = ntohs(attr->length);
      int atrType = ntohs(attr->type);
		
      if ( attrLen+4 > size ) 
      {
         ortp_error("stun: claims attribute is larger than size of message (attribute type=%i)", atrType);
         return FALSE;
      }
		
      body += 4; /* skip the length and type in attribute header */
      size -= 4;
		
      if (atrType == SA_MAPPEDADDRESS)
      {
            msg->hasMappedAddress = TRUE;
            if ( stunParseAtrAddress(  body,  attrLen,  &msg->mappedAddress )== FALSE )
            {
               ortp_error("stun: problem parsing SA_MAPPEDADDRESS");
               return FALSE;
            }
            else
            {
               stunIPAddressStr(&msg->mappedAddress, tmp, sizeof(tmp));
               ortp_debug("stun: SA_MAPPEDADDRESS = %s", tmp);
            }
					
      }
      else if (atrType == SA_RESPONSEADDRESS)
      {
            msg->hasResponseAddress = TRUE;
            if ( stunParseAtrAddress(  body,  attrLen,  &msg->responseAddress )== FALSE )
            {
               ortp_error("stun: problem parsing SA_RESPONSEADDRESS");
               return FALSE;
            }
            else
            {
               stunIPAddressStr(&msg->responseAddress, tmp, sizeof(tmp));
               ortp_debug("stun: SA_RESPONSEADDRESS = %s", tmp);
            }
      }
      else if (atrType == SA_CHANGEREQUEST)
      {
            msg->hasChangeRequest = TRUE;
            if (stunParseAtrChangeRequest( body, attrLen, &msg->changeRequest) == FALSE)
            {
               ortp_error("stun: problem parsing SA_CHANGEREQUEST");
               return FALSE;
            }
            else
            {
               ortp_debug("stun: SA_CHANGEREQUEST = %i", msg->changeRequest.value);
            }
      }
      else if (atrType == SA_SOURCEADDRESS)
      {
            msg->hasSourceAddress = TRUE;
            if ( stunParseAtrAddress(  body,  attrLen,  &msg->sourceAddress )== FALSE )
            {
               ortp_error("stun: problem parsing SA_SOURCEADDRESS");
               return FALSE;
            }
            else
            {
               stunIPAddressStr(&msg->sourceAddress, tmp, sizeof(tmp));
               ortp_debug("stun: SA_SOURCEADDRESS = %s", tmp );
            }
      }
      else if (atrType == SA_CHANGEDADDRESS)
      {
            msg->hasChangedAddress = TRUE;
            if ( stunParseAtrAddress(  body,  attrLen,  &msg->changedAddress )== FALSE )
            {
               ortp_error("stun: problem parsing SA_CHANGEDADDRESS");
               return FALSE;
            }
            else
            {
               stunIPAddressStr(&msg->changedAddress, tmp, sizeof(tmp));
               ortp_debug("stun: SA_CHANGEDADDRESS = %s", tmp);
            }
      }
      else if (atrType == SA_USERNAME)
      {
            msg->hasUsername = TRUE;
            if (stunParseAtrString( body, attrLen, &msg->username) == FALSE)
            {
               ortp_error("stun: problem parsing SA_USERNAME");
               return FALSE;
            }
            else
            {
               ortp_debug("stun: SA_USERNAME = %s", msg->username.value );
            }					
      }
      else if (atrType == SA_PASSWORD)
      {
            msg->hasPassword = TRUE;
            if (stunParseAtrString( body, attrLen, &msg->password) == FALSE)
            {
               ortp_error("stun: problem parsing SA_PASSWORD");
               return FALSE;
            }
            else
            {
               ortp_debug("stun: SA_PASSWORD = %s", msg->password.value );
            }
      }
      else if (atrType == SA_MESSAGEINTEGRITY)
      {
            msg->hasMessageIntegrity = TRUE;
            if (stunParseAtrIntegrity( body, attrLen, &msg->messageIntegrity) == FALSE)
            {
               ortp_error("stun: problem parsing SA_MESSAGEINTEGRITY");
               return FALSE;
            }					
      }
      else if (atrType == SA_ERRORCODE)
      {
            msg->hasErrorCode = TRUE;
            if (stunParseAtrError(body, attrLen, &msg->errorCode) == FALSE)
            {
               ortp_error("stun: problem parsing SA_ERRORCODE");
               return FALSE;
            }
            else
            {
               ortp_debug("stun: SA_ERRORCODE = %i %i %s",
                                   msg->errorCode.errorClass ,
                                   msg->errorCode.number ,
                                   msg->errorCode.reason );
            }
					 
      }
      else if (atrType == SA_UNKNOWNATTRIBUTE)
      {
           msg->hasUnknownAttributes = TRUE;
            if (stunParseAtrUnknown(body, attrLen, &msg->unknownAttributes) == FALSE)
            {
               ortp_error("stun: problem parsing SA_UNKNOWNATTRIBUTE");
               return FALSE;
            }
      }
      else if (atrType == SA_REFLECTEDFROM)
      {
            msg->hasReflectedFrom = TRUE;
            if ( stunParseAtrAddress(  body,  attrLen,  &msg->reflectedFrom ) == FALSE )
            {
               ortp_error("stun: problem parsing SA_REFLECTEDFROM");
               return FALSE;
            }
      }
      else if (atrType == SA_REALM)
      {
            msg->hasRealm = TRUE;
            if (stunParseAtrString( body, attrLen, &msg->realmName) == FALSE)
            {
               ortp_error("stun: problem parsing SA_REALM");
               return FALSE;
            }
            else
            {
               ortp_debug("stun: SA_REALM = %s", msg->realmName.value );
            }
      }
      else if (atrType == SA_NONCE)
      {
            msg->hasNonce = TRUE;
            if (stunParseAtrString( body, attrLen, &msg->nonceName) == FALSE)
            {
               ortp_error("stun: problem parsing SA_NONCE");
               return FALSE;
            }
            else
            {
               ortp_debug("stun: SA_NONCE = %s", msg->nonceName.value );
            }
      }
      else if (atrType == SA_XORMAPPEDADDRESS || atrType == SA_XORMAPPEDADDRESS2)
      { 
           msg->hasXorMappedAddress = TRUE;
            if ( stunParseAtrAddress(  body,  attrLen,  &msg->xorMappedAddress ) == FALSE )
            {
               ortp_error("stun: problem parsing SA_XORMAPPEDADDRESS");
               return FALSE;
            }
            else
            {
               stunIPAddressStr(&msg->xorMappedAddress, tmp, sizeof(tmp));
               ortp_debug("stun: SA_XORMAPPEDADDRESS = %s", tmp );
            }
      }
      else if (atrType == SA_XORONLY)
      {
            ortp_warning("stun: SA_XORONLY - non standard extension ignored" );
      }
      else if (atrType == SA_SECONDARYADDRESS)
      {
            ortp_debug("stun: SA_SECONDARYADDRESS - non standard extension ignored" );
      }
      else if (atrType == SA_SOFTWARE)
      {
            msg->hasSoftware = TRUE;
            if (stunParseAtrString( body, attrLen, &msg->softwareName) == FALSE)
            {
               ortp_error("stun: problem parsing SA_SOFTWARE");
               return FALSE;
            }
            else
            {
               ortp_debug("stun: SA_SOFTWARE = %s", msg->softwareName.value );
            }
      }
      else if (atrType == TA_CHANNELNUMBER)
      {
           msg->hasChannelNumberAttributes = TRUE;
            if (turnParseAtrChannelNumber(body, attrLen, &msg->channelNumberAttributes) == FALSE)
            {
               ortp_error("stun: problem parsing TA_CHANNELNUMBER");
               return FALSE;
            }					
      }
      else if (atrType == TA_LIFETIME)
      {
           msg->hasLifetimeAttributes = TRUE;
            if (turnParseAtrLifetime(body, attrLen, &msg->lifetimeAttributes) == FALSE)
            {
               ortp_error("stun: problem parsing TA_LIFETIME");
               return FALSE;
            }					
      }
      else if (atrType == TA_DEPRECATEDBANDWIDTH)
      {
           ortp_warning("stun: deprecated attribute TA_DEPRECATEDBANDWIDTH");
      }
      else if (atrType == TA_XORPEERADDRESS)
      {
            msg->hasXorPeerAddress = TRUE;
            if ( stunParseAtrAddress(  body,  attrLen,  &msg->xorPeerAddress )== FALSE )
            {
               ortp_error("stun: problem parsing SA_XORPEERADDRESS");
               return FALSE;
            }
            else
            {
               stunIPAddressStr(&msg->xorPeerAddress, tmp, sizeof(tmp));
               ortp_debug("stun: SA_XORPEERADDRESS = %s", tmp);
            }
      }
      else if (atrType == TA_DATA)
      {
            msg->hasData = TRUE;
            if (turnParseAtrData( body, attrLen, &msg->data) == FALSE)
            {
               ortp_error("stun: problem parsing TA_DATA");
               return FALSE;
            }
            else
            {
            }
      }
      else if (atrType == TA_XORRELAYEDADDRESS)
      {
            msg->hasXorRelayedAddress = TRUE;
            if ( stunParseAtrAddress(  body,  attrLen,  &msg->xorRelayedAddress )== FALSE )
            {
               ortp_error("stun: problem parsing TA_XORRELAYEDADDRESS");
               return FALSE;
            }
            else
            {
               stunIPAddressStr(&msg->xorRelayedAddress, tmp, sizeof(tmp));
               ortp_debug("stun: TA_XORRELAYEDADDRESS = %s", tmp);
            }
      }
      else if (atrType == RFC6156A_REQUESTADDRESSFAMILY)
      {
            msg->hasRequestAddressFamily = TRUE;
            if ( turnParseAtrRequestAddressFamily(  body,  attrLen,  &msg->requestAddressFamily)== FALSE )
            {
               ortp_error("stun: problem parsing RFC6156A_REQUESTADDRESSFAMILY");
               return FALSE;
            }
      }
      else if (atrType == TA_EVENPORT)
      {
            msg->hasEvenPort = TRUE;
            if ( turnParseAtrEvenPort(  body,  attrLen,  &msg->evenPort)== FALSE )
            {
               ortp_error("stun: problem parsing TA_EVENPORT");
               return FALSE;
            }
      }
      else if (atrType == TA_REQUESTEDTRANSPORT)
      {
            msg->hasRequestedTransport = TRUE;
            if ( turnParseAtrRequestedTransport(  body,  attrLen,  &msg->requestedTransport )== FALSE )
            {
               ortp_error("stun: problem parsing TA_REQUESTEDTRANSPORT");
               return FALSE;
            }
      }
      else if (atrType == TA_DONTFRAGMENT)
      {
            msg->hasDontFragment = TRUE;
      }
      else if (atrType == TA_DEPRECATEDTIMERVAL)
      {
           ortp_warning("stun: deprecated attribute TA_DEPRECATEDTIMERVAL");
      }
      else if (atrType == TA_RESERVATIONTOKEN)
      {
            msg->hasReservationToken = TRUE;
            if ( turnParseAtrReservationToken(  body,  attrLen,  &msg->reservationToken)== FALSE )
            {
               ortp_error("stun: problem parsing TA_RESERVATIONTOKEN");
               return FALSE;
            }
      }
      else if (atrType == SA_FINGERPRINT)
      {
            msg->hasFingerprint = TRUE;
            if ( stunParseAtrFingerprint(  body,  attrLen,  &msg->fingerprint)== FALSE )
            {
               ortp_error("stun: problem parsing SA_FINGERPRINT");
               return FALSE;
            }
      }
      else if (atrType == ICEA_PRIORITY)
      {
            msg->hasPriority = TRUE;
            if (iceParseAtrPriority(body, attrLen, &msg->priority) == FALSE)
            {
               ortp_error("stun: problem parsing ICEA_PRIORITY");
               return FALSE;
            }					
      }
      else if (atrType == ICEA_USECANDIDATE)
      {
            msg->hasUseCandidate = TRUE;
      }
      else if (atrType == ICEA_ICECONTROLLED)
      {
           msg->hasIceControlled = TRUE;
           if (iceParseAtrIceControll(body, attrLen, &msg->iceControlled) == FALSE)
            {
               ortp_error("stun: problem parsing ICEA_ICECONTROLLED");
               return FALSE;
            }
      }
      else if (atrType == ICEA_ICECONTROLLING)
      {
           msg->hasIceControlling = TRUE;
           if (iceParseAtrIceControll(body, attrLen, &msg->iceControlling) == FALSE)
            {
               ortp_error("stun: problem parsing ICEA_ICECONTROLLING");
               return FALSE;
            }
      }
      else if (atrType == RFC5780_RESPONSE_ORIGIN)
      {
            msg->hasResponseOrigin = TRUE;
            if ( stunParseAtrAddress(  body,  attrLen,  &msg->responseOrigin )== FALSE )
            {
               ortp_error("stun: problem parsing RFC5780_RESPONSE_ORIGIN");
               return FALSE;
            }
            else
            {
               stunIPAddressStr(&msg->responseOrigin, tmp, sizeof(tmp));
               ortp_debug("stun: RFC5780_RESPONSE_ORIGIN = %s", tmp);
            }
      }
      else if (atrType == RFC5780_OTHER_ADDRESS)
      {
           ortp_warning("stun: TODO: RFC5780_OTHER_ADDRESS", atrType );
      }
      else if (atrType == ICE_NETWORK_COST)
      {
           msg->hasNetworkCost = TRUE;
           if (iceParseAtrNetworkCost(body, attrLen, &msg->networkCost) == FALSE)
            {
               ortp_error("stun: problem parsing ICE_NETWORK_COST");
               return FALSE;
            }
        /* https://tools.ietf.org/html/draft-thatcher-ice-network-cost-00 */
        /* ICE Network Cost: Dynamically selecting ICE candidate pairs based on relative cost of network interfaces */
        /* ortp_warning("stun: TODO: DRAFT_NETWORK_COST", atrType ); */
      }
      else
      {
            if ( atrType <= 0x7FFF ) 
            {
              ortp_error("stun: Unknown Comprehension-Required attribute: %04x", atrType );
              return FALSE;
            }
            else
              ortp_warning("stun: Unknown attribute: %04x", atrType );
      }
		
      if (attrLen%4>0)
        attrLen += (4-(attrLen%4));
      
      body += attrLen;
      size -= attrLen;
   }
    
   return TRUE;
}


static char* 
encode16(char* buf, uint16_t data)
{
   uint16_t ndata = htons(data);
   memcpy(buf, &ndata, sizeof(uint16_t));
   return buf + sizeof(uint16_t);
}

static char* 
encode32(char* buf, uint32_t data)
{
   uint32_t ndata = htonl(data);
   memcpy(buf, &ndata, sizeof(uint32_t));
   return buf + sizeof(uint32_t);
}

static char* 
encode64(char* buf, uint64_t data)
{
   uint64_t ndata = htonq(data);
   memcpy(buf, &ndata, sizeof(uint64_t));
   return buf + sizeof(uint64_t);
}

static char* 
encode(char* buf, const char* data, unsigned int length)
{
   memcpy(buf, data, length);
   return buf + length;
}


static char* 
encodeAtrAddress4(char* ptr, uint16_t type, const StunIPAddress *atr)
{
  if (atr->family == IPv4Family) {
    ptr = encode16(ptr, type);
    ptr = encode16(ptr, 8);
    *ptr++ = 0;
    *ptr++ = IPv4Family;
    ptr = encode16(ptr, atr->addr.ipv4.port);
    ptr = encode32(ptr, atr->addr.ipv4.addr);
  } else if (atr->family == IPv6Family) {
    ptr = encode16(ptr, type);
    ptr = encode16(ptr, 20);
    *ptr++ = 0;
    *ptr++ = IPv6Family;
    ptr = encode16(ptr, atr->addr.ipv6.port);

    *ptr++ = atr->addr.ipv6.addr[0];
    *ptr++ = atr->addr.ipv6.addr[1];
    *ptr++ = atr->addr.ipv6.addr[2];
    *ptr++ = atr->addr.ipv6.addr[3];
    *ptr++ = atr->addr.ipv6.addr[4];
    *ptr++ = atr->addr.ipv6.addr[5];
    *ptr++ = atr->addr.ipv6.addr[6];
    *ptr++ = atr->addr.ipv6.addr[7];
    *ptr++ = atr->addr.ipv6.addr[8];
    *ptr++ = atr->addr.ipv6.addr[9];
    *ptr++ = atr->addr.ipv6.addr[10];
    *ptr++ = atr->addr.ipv6.addr[11];
    *ptr++ = atr->addr.ipv6.addr[12];
    *ptr++ = atr->addr.ipv6.addr[13];
    *ptr++ = atr->addr.ipv6.addr[14];
    *ptr++ = atr->addr.ipv6.addr[15];
  }
   return ptr;
}

static char* 
encodeAtrChangeRequest(char* ptr, const StunAtrChangeRequest *atr)
{
   ptr = encode16(ptr, SA_CHANGEREQUEST);
   ptr = encode16(ptr, 4);
   ptr = encode32(ptr, atr->value);
   return ptr;
}

static char* 
encodeAtrError(char* ptr, const StunAtrError *atr)
{
   int padding;
   int i;

   ptr = encode16(ptr, SA_ERRORCODE);
   ptr = encode16(ptr, 4 + atr->sizeReason);
   ptr = encode16(ptr, atr->pad);
   *ptr++ = atr->errorClass;
   *ptr++ = atr->number;
   ptr = encode(ptr, atr->reason, atr->sizeReason);

   padding = (atr->sizeReason+4) % 4;
   if (padding>0)
   {
     for (i=0;i<4-padding;i++)
     {
       *ptr++ = 0;
     }
   }
   return ptr;
}


static char* 
encodeAtrUnknown(char* ptr, const StunAtrUnknown *atr)
{
   int i;
   ptr = encode16(ptr, SA_UNKNOWNATTRIBUTE);
   ptr = encode16(ptr, 2+2*atr->numAttributes);
   for (i=0; i<atr->numAttributes; i++)
   {
      ptr = encode16(ptr, atr->attrType[i]);
   }
   return ptr;
}

static char* 
encodeAtrString(char* ptr, uint16_t type, const StunAtrString *atr)
{
   int padding;
   int i;
	
   ptr = encode16(ptr, type);
   ptr = encode16(ptr, atr->sizeValue);
   ptr = encode(ptr, atr->value, atr->sizeValue);

   padding = atr->sizeValue % 4;
   if (padding>0)
   {
     for (i=0;i<4-padding;i++)
     {
       *ptr++ = 0;
     }
   }
   return ptr;
}

static char* 
encodeAtrData(char* ptr, const TurnAtrData *atr)
{
   int padding;
   int i;
	
   ptr = encode16(ptr, TA_DATA);
   ptr = encode16(ptr, atr->sizeValue);
   ptr = encode(ptr, atr->value, atr->sizeValue);

   padding = atr->sizeValue % 4;
   if (padding>0)
   {
     for (i=0;i<4-padding;i++)
     {
       *ptr++ = 0;
     }
   }
   return ptr;
}

static char* 
encodeAtrIntegrity(char* ptr, const StunAtrIntegrity *atr)
{
   ptr = encode16(ptr, SA_MESSAGEINTEGRITY);
   ptr = encode16(ptr, 20);
   ptr = encode(ptr, atr->hash, sizeof(atr->hash));
   return ptr;
}

static char*
encodeAtrFingerprint(char* ptr, const StunAtrFingerprint *atr)
{
	uint32_t val;
	ptr = encode16(ptr, SA_FINGERPRINT);
	ptr = encode16(ptr, 4);

	val = atr->fingerprint;
	val ^= 0x5354554E;
	ptr = encode32(ptr, val);
	return ptr;
}

static char*
encodeAtrChannelNumber(char* ptr, const TurnAtrChannelNumber *atr)
{
   ptr = encode16(ptr, TA_CHANNELNUMBER);
   ptr = encode16(ptr, 4);
   ptr = encode16(ptr, atr->channelNumber);
   ptr = encode16(ptr, atr->rffu);
   return ptr;
}

static char*
encodeAtrReservationToken(char* ptr, const TurnAtrReservationToken *atr)
{
   ptr = encode16(ptr, TA_RESERVATIONTOKEN);
   ptr = encode16(ptr, 8);
   ptr = encode64(ptr, atr->value);
   return ptr;
}

static char* 
encodeAtrRequestAddressFamily(char* ptr, const Rfc6156AtrRequestAddressFamily *atr)
{
   ptr = encode16(ptr, RFC6156A_REQUESTADDRESSFAMILY);
   ptr = encode16(ptr, 4);
   *ptr++ = atr->family;
   *ptr++ = atr->pad1;
   *ptr++ = atr->pad2;
   *ptr++ = atr->pad3;
   return ptr;
}

static char*
encodeAtrEvenPort(char* ptr, const TurnAtrEvenPort *atr)
{
   ptr = encode16(ptr, TA_EVENPORT);
   ptr = encode16(ptr, 4);
   *ptr++ = atr->value;
   *ptr++ = atr->pad1;
   *ptr++ = atr->pad2;
   *ptr++ = atr->pad3;
   return ptr;
}

static char* 
encodeAtrRequestedTransport(char* ptr, const TurnAtrRequestedTransport *atr)
{
   ptr = encode16(ptr, TA_REQUESTEDTRANSPORT);
   ptr = encode16(ptr, 4);
   *ptr++ = atr->proto;
   *ptr++ = atr->pad1;
   *ptr++ = atr->pad2;
   *ptr++ = atr->pad3;
   return ptr;
}

static char* 
encodeAtrLifeTime(char* ptr, const TurnAtrLifetime *atr)
{
   ptr = encode16(ptr, TA_LIFETIME);
   ptr = encode16(ptr, 4);
   ptr = encode32(ptr, atr->lifetime);
   return ptr;
}

static char* 
encodeAtrDontFragment(char* ptr)
{
   ptr = encode16(ptr, TA_DONTFRAGMENT);
   ptr = encode16(ptr, 0);
   return ptr;
}

static char* 
encodeAtrUseCandidate(char* ptr)
{
   ptr = encode16(ptr, ICEA_USECANDIDATE);
   ptr = encode16(ptr, 0);
   return ptr;
}

static char* 
encodeAtrPriority(char* ptr, const IceAtrPriority *atr)
{
   ptr = encode16(ptr, ICEA_PRIORITY);
   ptr = encode16(ptr, 4);
   ptr = encode32(ptr, atr->priority);
   return ptr;
}

static char* 
encodeAtrIceControll(char* ptr, uint16_t type, const IceAtrIceControll *atr)
{
   ptr = encode16(ptr, type);
   ptr = encode16(ptr, 8);
   ptr = encode64(ptr, atr->value);
   return ptr;
}

static char*
encodeAtrNetworkCost(char* ptr, const IceAtrNetworkCost *atr)
{
   ptr = encode16(ptr, ICE_NETWORK_COST);
   ptr = encode16(ptr, 4);
   ptr = encode16(ptr, atr->networkid);
   ptr = encode16(ptr, atr->networkcost);
   return ptr;
}

unsigned int
stunEncodeMessage( const StunMessage *msg, 
                   char* buf, 
                   unsigned int bufLen, 
                   const StunAtrString *password)
{
   char tmp[512];
   char* ptr = buf;
   char* lengthp;
   ptr = encode16(ptr, msg->msgHdr.msgType);
   lengthp = ptr;
   ptr = encode16(ptr, 0);
   ptr = encode32(ptr, msg->msgHdr.magic_cookie);
   ptr = encode(ptr, (const char*)msg->msgHdr.tr_id.octet, sizeof(msg->msgHdr.tr_id));
	
   ortp_debug("stun: Encoding stun message: ");

   if (msg->hasRequestedTransport)
   {
      ortp_debug("stun: Encoding TA_REQUESTEDTRANSPORT: %i", msg->requestedTransport.proto );
      ptr = encodeAtrRequestedTransport (ptr, &msg->requestedTransport);
   }
   if (msg->hasRequestAddressFamily)
   {
      ortp_debug("stun: Encoding RFC6156A_REQUESTADDRESSFAMILY: %i", msg->requestAddressFamily.family);
      ptr = encodeAtrRequestAddressFamily (ptr, &msg->requestAddressFamily);
   }
   if (msg->hasEvenPort) 
   {
      ortp_debug("stun: Encoding TA_EVENPORT: %i", msg->evenPort.value);
      ptr = encodeAtrEvenPort(ptr, &msg->evenPort);
   }
   if (msg->hasReservationToken)
   {
      ortp_debug("stun: Encoding TA_RESERVATIONTOKEN");
      ptr = encodeAtrReservationToken(ptr, &msg->reservationToken);
   }
   if (msg->hasLifetimeAttributes)
   {
      ortp_debug("stun: Encoding TA_LIFETIME: %i", msg->lifetimeAttributes.lifetime );
      ptr = encodeAtrLifeTime (ptr, &msg->lifetimeAttributes);
   }
   if (msg->hasDontFragment)
   {
      ortp_debug("stun: Encoding TA_DONTFRAGMENT: DF");
      ptr = encodeAtrDontFragment (ptr);
   }		  
   if (msg->hasMappedAddress)
   {
      stunIPAddressStr(&msg->mappedAddress, tmp, sizeof(tmp));
      ortp_debug("stun: Encoding SA_MAPPEDADDRESS: %s", tmp );
      ptr = encodeAtrAddress4 (ptr, SA_MAPPEDADDRESS, &msg->mappedAddress);
   }
   if (msg->hasResponseAddress)
   {
      stunIPAddressStr(&msg->responseAddress, tmp, sizeof(tmp));
      ortp_debug("stun: Encoding SA_RESPONSEADDRESS: %s", tmp );
      ptr = encodeAtrAddress4(ptr, SA_RESPONSEADDRESS, &msg->responseAddress);
   }
   if (msg->hasChangeRequest)
   {
      ortp_debug("stun: Encoding SA_CHANGEREQUEST: %i", msg->changeRequest.value );
      ptr = encodeAtrChangeRequest(ptr, &msg->changeRequest);
   }
   if (msg->hasSourceAddress)
   {
      stunIPAddressStr(&msg->sourceAddress, tmp, sizeof(tmp));
      ortp_debug("stun: Encoding SA_SOURCEADDRESS: %s", tmp );
      ptr = encodeAtrAddress4(ptr, SA_SOURCEADDRESS, &msg->sourceAddress);
   }
   if (msg->hasChangedAddress)
   {
      stunIPAddressStr(&msg->changedAddress, tmp, sizeof(tmp));
      ortp_debug("stun: Encoding SA_CHANGEDADDRESS: %s", tmp );
      ptr = encodeAtrAddress4(ptr, SA_CHANGEDADDRESS, &msg->changedAddress);
   }
   if (msg->hasUsername)
   {
     if (STUN_IS_SUCCESS_RESP(msg->msgHdr.msgType) || STUN_IS_ERR_RESP(msg->msgHdr.msgType)) {
       /* username must not be included in response */
     } else {
      ortp_debug("stun: Encoding SA_USERNAME: %s", msg->username.value );
      ptr = encodeAtrString(ptr, SA_USERNAME, &msg->username);
     }
   }
   //if (msg->hasPassword)
   //{
   //   ortp_debug("stun: Encoding SA_PASSWORD: %s", msg->password.value );
   //   ptr = encodeAtrString(ptr, SA_PASSWORD, &msg->password);
   //}
   if (msg->hasErrorCode)
   {
      ortp_debug("stun: Encoding SA_ERRORCODE: class=%i number=%i reason=%s" 
                          , msg->errorCode.errorClass 
                          , msg->errorCode.number
                          , msg->errorCode.reason );
      
      ptr = encodeAtrError(ptr, &msg->errorCode);
   }
   if (msg->hasUnknownAttributes)
   {
      ortp_debug("stun: Encoding SA_UNKNOWNATTRIBUTE: ???");
      ptr = encodeAtrUnknown(ptr, &msg->unknownAttributes);
   }
   if (msg->hasReflectedFrom)
   {
      stunIPAddressStr(&msg->reflectedFrom, tmp, sizeof(tmp));
      ortp_debug("stun: Encoding SA_REFLECTEDFROM: %s", tmp );
      ptr = encodeAtrAddress4(ptr, SA_REFLECTEDFROM, &msg->reflectedFrom);
   }
   if (msg->hasNonce)
   {
      ortp_debug("stun: Encoding SA_NONCE: %s", msg->nonceName.value );
      ptr = encodeAtrString(ptr, SA_NONCE, &msg->nonceName);
   }
   if (msg->hasRealm)
   {
      ortp_debug("stun: Encoding SA_REALM: %s", msg->realmName.value );
      ptr = encodeAtrString(ptr, SA_REALM, &msg->realmName);
   }
   
   if (msg->hasXorMappedAddress)
   {
      stunIPAddressStr(&msg->xorMappedAddress, tmp, sizeof(tmp));
      ortp_debug("stun: Encoding SA_XORMAPPEDADDRESS: %s", tmp );
      ptr = encodeAtrAddress4 (ptr, SA_XORMAPPEDADDRESS, &msg->xorMappedAddress);
   }
   
   if (msg->hasPriority)
   {	   
      ortp_debug("stun: Encoding ICEA_PRIORITY");
      ptr = encodeAtrPriority (ptr, &msg->priority);
   }
   if (msg->hasUseCandidate)
   {	   
      ortp_debug("stun: Encoding ICEA_USECANDIDATE");
      ptr = encodeAtrUseCandidate (ptr);
   }
   if (msg->hasIceControlled)
   {	   
      ortp_debug("stun: Encoding ICEA_ICECONTROLLED");
      ptr = encodeAtrIceControll (ptr, ICEA_ICECONTROLLED, &msg->iceControlled);
   }
   if (msg->hasIceControlling)
   {	   
      ortp_debug("stun: Encoding ICEA_ICECONTROLLING");
      ptr = encodeAtrIceControll (ptr, ICEA_ICECONTROLLING, &msg->iceControlling);
   }

   if (msg->hasSoftware)
   {
      ortp_debug("stun: Encoding SA_SOFTWARE: %s", msg->softwareName.value );
      ptr = encodeAtrString(ptr, SA_SOFTWARE, &msg->softwareName);
   }

   if (msg->hasChannelNumberAttributes)
   {
      ortp_debug("stun: Encoding TA_CHANNELNUMBER: chan=%d, rffu=%d", 
                                 msg->channelNumberAttributes.channelNumber, 
                                 msg->channelNumberAttributes.rffu);
      ptr = encodeAtrChannelNumber(ptr, &msg->channelNumberAttributes);
   }

   if (msg->hasXorPeerAddress)
   {
      stunIPAddressStr(&msg->xorPeerAddress, tmp, sizeof(tmp));
      ortp_debug("stun: Encoding TA_XORPEERADDRESS: %s", tmp );
      ptr = encodeAtrAddress4 (ptr, TA_XORPEERADDRESS, &msg->xorPeerAddress);
   }

   if (msg->hasNetworkCost)
   {
      ptr = encodeAtrNetworkCost (ptr, &msg->networkCost);
   }

   if (msg->hasData) {
      ortp_debug("stun: Encoding TA_DATA: -----");
      ptr = encodeAtrData (ptr, &msg->data);
   }

   if (msg->hasMessageIntegrity
     &&password!=NULL && password->sizeValue > 0
     &&msg->username.sizeValue>0
     &&msg->realmName.sizeValue>0)
   {
      StunAtrIntegrity integrity;
      //ortp_debug("stun: HMAC with password: %s", password->value );

      encode16(lengthp, (uint16_t)(ptr - buf - sizeof(StunMsgHdr)+24));
      stunCalculateIntegrity_longterm(integrity.hash, buf, (int)(ptr-buf) ,
        msg->username.value, msg->realmName.value, password->value);
      ptr = encodeAtrIntegrity(ptr, &integrity);
   }
   else if (msg->hasMessageIntegrity
     &&password!=NULL && password->sizeValue > 0
     &&msg->username.sizeValue>0)
   {
      StunAtrIntegrity integrity;
      //ortp_debug("stun: HMAC with password: %s", password->value );

      encode16(lengthp, (uint16_t)(ptr - buf - sizeof(StunMsgHdr)+24));
      stunCalculateIntegrity_shortterm(integrity.hash, buf, (int)(ptr-buf) ,
        password->value);
      ptr = encodeAtrIntegrity(ptr, &integrity);
   }

   if (msg->hasFingerprint)
   {
      StunAtrFingerprint fingerprint;
      //ortp_debug("stun: HMAC with password: %s", password->value );

      encode16(lengthp, (uint16_t)(ptr - buf - sizeof(StunMsgHdr)+8));
      fingerprint.fingerprint = stunCalculateFingerprint(buf, (int)(ptr-buf));
      ptr = encodeAtrFingerprint(ptr, &fingerprint);
   }
   encode16(lengthp, (uint16_t)(ptr - buf - sizeof(StunMsgHdr)));
   return (int)(ptr - buf);
}

/* return a random number to use as a port */
static int
randomPort()
{
   int min=0x4000;
   int max=0x7FFF;
	
   int ret = ortp_random();
   ret = ret|min;
   ret = ret&max;
	
   return ret;
}

#include "amd_hmac.h"

void
stunCalculateIntegrity_longterm(char* hmac, const char* input, int length,
                     const char *username, const char *realm, const char *password)
{
	unsigned char HA1[16];
	char HA1_text[1024];
	AMD_HMAC_SHA1_CTX ctx;
	AMD_MD5_CTX Md5Ctx;

	snprintf(HA1_text, sizeof(HA1_text), "%s:%s:%s", username, realm, password);
	AMD_MD5Init (&Md5Ctx);
	AMD_MD5Update (&Md5Ctx, (unsigned char *) HA1_text, (unsigned int) strlen (HA1_text));
	AMD_MD5Final ((unsigned char *) HA1, &Md5Ctx);

	AMD_HMAC_SHA1_Init(&ctx, (uint8_t*)HA1, 16);
	AMD_HMAC_SHA1_Update(&ctx, (uint8_t*)input, length);
	AMD_HMAC_SHA1_Final((uint8_t*)hmac, &ctx);
}

void
stunCalculateIntegrity_shortterm(char* hmac, const char* input, int length, const char* key)
{
	AMD_HMAC_SHA1_CTX ctx;

	AMD_HMAC_SHA1_Init(&ctx, (uint8_t*)key, strlen(key));
	AMD_HMAC_SHA1_Update(&ctx, (uint8_t*)input, length);
	AMD_HMAC_SHA1_Final((uint8_t*)hmac, &ctx);
}

uint32_t
stunCalculateFingerprint(const char* input, int length)
{
	/*-
	2  *  COPYRIGHT (C) 1986 Gary S. Brown.  You may use this program, or
	3  *  code or tables extracted from it, as desired without restriction.
	4  */
	static uint32_t crc32_tab[] = {
		0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f,
		0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
		0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91, 0x1db71064, 0x6ab020f2,
		0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
		0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9,
		0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
		0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b, 0x35b5a8fa, 0x42b2986c,
		0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
		0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423,
		0xcfba9599, 0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
		0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190, 0x01db7106,
		0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
		0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d,
		0x91646c97, 0xe6635c01, 0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
		0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
		0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
		0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7,
		0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
		0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa,
		0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
		0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81,
		0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
		0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683, 0xe3630b12, 0x94643b84,
		0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
		0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
		0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
		0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5, 0xd6d6a3e8, 0xa1d1937e,
		0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
		0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55,
		0x316e8eef, 0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
		0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe, 0xb2bd0b28,
		0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
		0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f,
		0x72076785, 0x05005713, 0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
		0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
		0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
		0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69,
		0x616bffd3, 0x166ccf45, 0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
		0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc,
		0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
		0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693,
		0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
		0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
	};
	const uint8_t *p = (uint8_t*)input;
	uint32_t crc;

	crc = ~0U;
	while (length--)
		crc = crc32_tab[(crc ^ *p++) & 0xFF] ^ (crc >> 8);
	return crc ^ ~0U;
}

uint64_t
stunGetSystemTimeSecs(void)
{
   uint64_t time=0;
#if	defined(_WIN32) || defined(_WIN32_WCE)
   SYSTEMTIME t;
   /*  CJ TODO - this probably has bug on wrap around every 24 hours */
   GetSystemTime( &t );
   time = (t.wHour*60+t.wMinute)*60+t.wSecond; 
#else
   struct timeval now;
   gettimeofday( &now , NULL );
   /* assert( now ); */
   time = now.tv_sec;
#endif
   return time;
}

void
stunBuildReqSimple( StunMessage* msg,
                    const StunAtrString *username,
                    bool_t changePort, bool_t changeIp, unsigned int id )
{
   int i;
   /* assert( msg ); */
   memset( msg , 0 , sizeof(*msg) );
	
   msg->msgHdr.msgType = (STUN_METHOD_BINDING|STUN_REQUEST);
	
   msg->msgHdr.magic_cookie = 0x2112A442;
   for ( i=0; i<12; i=i+4 )
   {
      /* assert(i+3<16); */
      int r = ortp_random();
      msg->msgHdr.tr_id.octet[i+0]= r>>0;
      msg->msgHdr.tr_id.octet[i+1]= r>>8;
      msg->msgHdr.tr_id.octet[i+2]= r>>16;
      msg->msgHdr.tr_id.octet[i+3]= r>>24;
   }
	
   if ( id != 0 )
   {
      msg->msgHdr.tr_id.octet[0] = id; 
   }
	
   if (changePort==TRUE || changeIp==TRUE)
   {
     msg->hasChangeRequest = TRUE;
     msg->changeRequest.value =(changeIp?ChangeIpFlag:0) | 
        (changePort?ChangePortFlag:0);
   }

   if ( username!=NULL && username->sizeValue > 0 )
   {
      msg->hasUsername = TRUE;
      /* msg->username = username; */
      memcpy(&msg->username, username, sizeof(StunAtrString));
   }
}


static void 
stunSendTest( Socket myFd, StunIPAddress *dest,
              const StunAtrString *username, const StunAtrString *password, 
              int testNum )
{ 
   /* assert( dest.addr != 0 ); */
   /* assert( dest.port != 0 ); */
	
   bool_t changePort=FALSE;
   bool_t changeIP=FALSE;
#if 0
   bool_t discard=FALSE;
#endif
   
   StunMessage req;
   char buf[STUN_MAX_MESSAGE_SIZE];
   int len = STUN_MAX_MESSAGE_SIZE;
   
   switch (testNum)
   {
      case 1:
      case 10:
      case 11:
         break;
      case 2:
         /* changePort=TRUE; */
         changeIP=TRUE;
         break;
      case 3:
         changePort=TRUE;
         break;
      case 4:
         changeIP=TRUE;
         break;
#if 0
      case 5:
         discard=TRUE;
         break;
#endif
      default:
         ortp_error("stun: Test %i is unkown", testNum);
         return ; /* error */
   }
   
   memset(&req, 0, sizeof(StunMessage));
	
   stunBuildReqSimple( &req, username, 
                       changePort , changeIP , 
                       testNum );
	
   len = stunEncodeMessage( &req, buf, len, password );
	
   //ortp_debug("stun: About to send msg of len %i to xxx", len);
	
   sendMessage( myFd, buf, len, dest );
	
   /* add some delay so the packets don't get sent too quickly */
#if defined(_WIN32_WCE) 
   Sleep (10);
#elif defined(WIN32)/* !cj! TODO - should fix this up in windows */
   {
       clock_t now = clock();
       /* assert( CLOCKS_PER_SEC == 1000 ); */
       while ( clock() <= now+10 ) { };
   }
#else
		 usleep(10*1000);
#endif

}

NatType
stunNatType( StunIPAddress *dest,
             bool_t* preservePort, /* if set, is return for if NAT preservers ports or not */
             bool_t* hairpin,  /* if set, is the return for if NAT will hairpin packets */
             StunIPAddress* sAddr /* NIC to use */
   )
{ 
   /* assert( dest.addr != 0 ); */
   /* assert( dest.port != 0 ); */
   Socket myFd1;
   Socket myFd2;

   bool_t respTestI=FALSE;
   bool_t isNat=TRUE;
#if 0
   StunIPAddress testIchangedAddr;
#endif
   StunIPAddress testImappedAddr;
   bool_t respTestI2=FALSE; 
   bool_t mappedIpSame = TRUE;
   StunIPAddress testI2mappedAddr;
   /* StunIPAddress testI2dest=dest; */
   StunIPAddress testI2dest;
   bool_t respTestII=FALSE;
   bool_t respTestIII=FALSE;
   bool_t respTestHairpin=FALSE;
   StunAtrString username;
   StunAtrString password;
   int count=0;
   uint64_t second_started;
   uint64_t second_elapsed;
   Socket s;

   if ( hairpin ) 
   {
      *hairpin = FALSE;
   }
	
   if ( sAddr->addr.ipv4.port == 0 )
   {
      sAddr->addr.ipv4.port = randomPort();
   }

   myFd1 = openPort(sAddr);

   sAddr->addr.ipv4.port++;
   myFd2 = openPort(sAddr);
   sAddr->addr.ipv4.port--;

   if ( ( myFd1 == INVALID_SOCKET) || ( myFd2 == INVALID_SOCKET) )
   {
      ortp_error("stun: Some problem opening port/interface to send on");
      return StunTypeFailure; 
   }

   /* assert( myFd1 != INVALID_SOCKET ); */
   /* assert( myFd2 != INVALID_SOCKET ); */
   
   memcpy(&testI2dest, dest, sizeof(StunIPAddress));

   memset(&testImappedAddr,0,sizeof(testImappedAddr));
   
   username.sizeValue = 0;
   password.sizeValue = 0;

   /* stunSendTest( myFd1, dest, username, password, 1 ); */

   
   second_started = stunGetSystemTimeSecs();
   second_elapsed = 1;

   while ( count < 3 && second_elapsed < 5)
   {
      struct timeval tv;
      fd_set fdSet; 
      int err;
      int e;

#if defined(WIN32) || defined(_WIN32_WCE)
      unsigned int fdSetSize;
#else
      int fdSetSize;
#endif

      second_elapsed = stunGetSystemTimeSecs() - second_started ;

      FD_ZERO(&fdSet); fdSetSize=0;
      FD_SET(myFd1,&fdSet); fdSetSize = (myFd1+1>fdSetSize) ? myFd1+1 : fdSetSize;
      FD_SET(myFd2,&fdSet); fdSetSize = (myFd2+1>fdSetSize) ? myFd2+1 : fdSetSize;
      tv.tv_sec=0;
      tv.tv_usec=500*1000; /* 150 ms */
      if ( count == 0 ) tv.tv_usec=0;
		
      err = select(fdSetSize, &fdSet, NULL, NULL, &tv);
      e = getErrno();
      if ( err == SOCKET_ERROR )
      {
         /* error occured */
#if !defined(_WIN32_WCE)
         ortp_error("stun: Error %i %s in select", e, ortp_strerror(e));
#else
         ortp_error("stun: Error %i in select", e);
#endif
		 closesocket(myFd1); /* AMD */
         closesocket(myFd2); /* AMD */
         return StunTypeFailure;
     }
      else if ( err == 0 )
      {
         /* timeout occured */
         count++;
         if ( !respTestI ) 
         {
            stunSendTest( myFd1, dest, &username, &password, 1 );
         }         
			
         if ( (!respTestI2) && respTestI ) 
         {
            /* check the address to send to if valid */
            if (  ( testI2dest.addr.ipv4.addr != 0 ) &&
                  ( testI2dest.addr.ipv4.port != 0 ) )
            {
               stunSendTest( myFd1, &testI2dest, &username, &password, 10 );
            }
         }
			
         if ( !respTestII )
         {
            stunSendTest( myFd2, dest, &username, &password, 2 );
         }
			
         if ( !respTestIII )
         {
            stunSendTest( myFd2, dest, &username, &password, 3 );
         }
			
         if ( respTestI && (!respTestHairpin) )
         {
            if (  ( testImappedAddr.addr.ipv4.addr != 0 ) &&
                  ( testImappedAddr.addr.ipv4.port != 0 ) )
            {
               stunSendTest( myFd1, &testImappedAddr, &username, &password, 11 );
            }
         }

      }
      else
      {
         int i;
         /* data is avialbe on some fd */
			
         for ( i=0; i<2; i++)
         {
            Socket myFd;
            if ( i==0 ) 
            {
               myFd=myFd1;
            }
            else
            {
               myFd=myFd2;
            }
				
            if ( myFd!=INVALID_SOCKET ) 
            {					
               if ( FD_ISSET(myFd,&fdSet) )
               {
                  char msg[STUN_MAX_MESSAGE_SIZE];
                  int msgLen = sizeof(msg);
                  						
                  StunIPAddress from;
                  StunMessage resp;

                  getMessage( myFd,
                              msg,
                              &msgLen,
                              &from );
                  
                  memset(&resp, 0, sizeof(StunMessage));
						
                  stunParseMessage( msg,msgLen, &resp );
						
                  //ortp_debug("stun: Received message of type %i id=%i",
                          //resp.msgHdr.msgType,
                          //(int)(resp.msgHdr.tr_id.octet[0]) );
						
                  switch( resp.msgHdr.tr_id.octet[0] )
                  {
                     case 1:
                     {
                        if ( !respTestI )
                        {
#if 0
                           testIchangedAddr.addr = resp.changedAddress.ipv4.addr;
                           testIchangedAddr.port = resp.changedAddress.ipv4.port;
#endif
                           memcpy(&testImappedAddr, &resp.mappedAddress, sizeof(testImappedAddr));
									
                           if ( preservePort )
                           {
                              *preservePort = ( testImappedAddr.addr.ipv4.port == sAddr->addr.ipv4.port );
                           }								
									
                           testI2dest.family = resp.changedAddress.family;
                           memcpy(&testI2dest.addr.ipv6.addr, &resp.changedAddress.addr.ipv6.addr, 16);
									
                           if (sAddr)
                           {
                              memcpy(sAddr, &testImappedAddr, sizeof(StunIPAddress));
                           }
									
                           count = 0;
                        }		
                        respTestI=TRUE;
                     }
                     break;
                     case 2:
                     {  
                        respTestII=TRUE;
                     }
                     break;
                     case 3:
                     {
                        respTestIII=TRUE;
                     }
                     break;
                     case 10:
                     {
                        if ( !respTestI2 )
                        {
                           testI2mappedAddr.addr.ipv4.addr = resp.mappedAddress.addr.ipv4.addr;
                           testI2mappedAddr.addr.ipv4.port = resp.mappedAddress.addr.ipv4.port;
								
                           mappedIpSame = FALSE;
                           if ( (testI2mappedAddr.addr.ipv4.addr  == testImappedAddr.addr.ipv4.addr ) &&
                                (testI2mappedAddr.addr.ipv4.port == testImappedAddr.addr.ipv4.port ))
                           { 
                              mappedIpSame = TRUE;
                           }
								
							
                        }
                        respTestI2=TRUE;
                     }
                     break;
                     case 11:
                     {
							
                        if ( hairpin ) 
                        {
                           *hairpin = TRUE;
                        }
                        respTestHairpin = TRUE;
                     }
                     break;
                  }
               }
            }
         }
      }
   }
	
   closesocket(myFd1); /* AMD */
   closesocket(myFd2); /* AMD */

   /* see if we can bind to this address */
   /* cerr << "try binding to " << testImappedAddr ); */
   {
     StunIPAddress srcAddrWithPort;
     memcpy(&srcAddrWithPort, &testImappedAddr, sizeof(StunIPAddress));
     srcAddrWithPort.addr.ipv4.port = 0; /*use ephemeral*/
     s = openPort( &srcAddrWithPort );
   }
   if ( s != INVALID_SOCKET )
   {
      isNat = FALSE;
      /* cerr << "binding worked"); */
   }
   else
   {
      isNat = TRUE;
      /* cerr << "binding failed"); */
   }

   closesocket(s); /* AMD */
	
   //ortp_debug("stun: test I = %i", respTestI );
   //ortp_debug("stun: test II = %i", respTestII );
   //ortp_debug("stun: test III = %i", respTestIII );
   //ortp_debug("stun: test I(2) = %i", respTestI2 );
   ortp_debug("stun: is nat  = %i", isNat);
   ortp_debug("stun: mapped IP same = %i", mappedIpSame );
	
   /* implement logic flow chart from draft RFC */
   if ( respTestI )
   {
      if ( isNat )
      {
         if (respTestII)
         {
            return StunTypeConeNat;
         }
         else
         {
            if ( mappedIpSame )
            {
               if ( respTestIII )
               {
                  return StunTypeRestrictedNat;
               }
               else
               {
                  return StunTypePortRestrictedNat;
               }
            }
            else
            {
               return StunTypeSymNat;
            }
         }
      }
      else
      {
         if (respTestII)
         {
            return StunTypeOpen;
         }
         else
         {
            return StunTypeSymFirewall;
         }
      }
   }
   else
   {
      return StunTypeBlocked;
   }
	
   return StunTypeUnknown;
}

int
stunOpenSocket( StunIPAddress *dest, StunIPAddress* mapAddr,
                StunIPAddress* srcAddr )
{
   /* assert( dest.addr != 0 ); */
   /* assert( dest.port != 0 ); */
   /* assert( mapAddr );*/
   Socket myFd;
   char msg[STUN_MAX_MESSAGE_SIZE];
   int msgLen = sizeof(msg);
	
   StunAtrString username;
   StunAtrString password;

   StunIPAddress from;
   StunMessage resp;
   bool_t ok;
   StunIPAddress mappedAddr;

   if ( srcAddr->addr.ipv4.port == 0 )
   {
      srcAddr->addr.ipv4.port = randomPort();
   }

   myFd = openPort(srcAddr);
   if (myFd == INVALID_SOCKET)
   {
      return myFd;
   }
   
   username.sizeValue = 0;
   password.sizeValue = 0;

   stunSendTest(myFd, dest, &username, &password, 1 );
	
   getMessage( myFd, msg, &msgLen, &from );

   memset(&resp, 0, sizeof(StunMessage));
	
   ok = stunParseMessage( msg, msgLen, &resp );
   if (!ok)
   {
       closesocket(myFd);
       return -1;
   }

   if (resp.hasXorMappedAddress==TRUE)
   {
     stunDecodeXorAddress(&resp.xorMappedAddress, &mappedAddr, &resp.msgHdr.tr_id);
   }
   else
     mappedAddr = resp.mappedAddress;
	
   /*
     ortp_message("stun: --- stunOpenSocket --- ");
     ortp_message("stun: \treq  id=" << req.id );
     ortp_message("stun: \tresp id=" << id );
     ortp_message("stun: \tmappedAddr=" << mappedAddr );
   */

   *mapAddr = mappedAddr;
	
   return myFd;
}


bool_t
stunOpenSocketPair(StunIPAddress *dest,
                   StunIPAddress* mapAddr_rtp,
                   StunIPAddress* mapAddr_rtcp,
                   int* fd1, int* fd2, StunIPAddress* srcAddr )
{
   /* assert( dest.addr!= 0 ); */
   /* assert( dest.port != 0 ); */
   /* assert( mapAddr ); */
   
   const int NUM=2;
   char msg[STUN_MAX_MESSAGE_SIZE];
   int msgLen =sizeof(msg);
	
   StunIPAddress from;
   int fd[2/*NUM*/];
   int i;
	
   StunAtrString username;
   StunAtrString password;
	
   StunIPAddress mappedAddr[2/*NUM*/];
   StunIPAddress srcAddrWithPort;

   *fd1=-1;
   *fd2=-1;

   if ( srcAddr->addr.ipv4.port == 0 )
   {
      srcAddr->addr.ipv4.port = randomPort();
   }

   for( i=0; i<NUM; i++)
   {
      memcpy(&srcAddrWithPort, srcAddr, sizeof(StunIPAddress));
      if (srcAddrWithPort.addr.ipv4.port>0)
        srcAddrWithPort.addr.ipv4.port+=i;

      fd[i] = openPort( &srcAddrWithPort );
      if (fd[i] < 0) 
      {
         while (i > 0)
         {
            closesocket(fd[--i]);
         }
         return FALSE;
      }
   }
	
   username.sizeValue = 0;
   password.sizeValue = 0;

   for( i=0; i<NUM; i++)
   {
      stunSendTest(fd[i], dest, &username, &password, 1/*testNum*/ );
   }
   
   for( i=0; i<NUM; i++)
   {
      StunMessage resp;
      bool_t ok;
      msgLen = sizeof(msg)/sizeof(*msg);
      getMessage( fd[i],
                  msg,
                  &msgLen,
                  &from);
      
      memset(&resp, 0, sizeof(StunMessage));
		
      ok = stunParseMessage( msg, msgLen, &resp );
      if (!ok) 
      {  
          for( i=0; i<NUM; i++)
          {
              closesocket(fd[i]);
          }
         return FALSE;
      }
	  
      if (resp.hasXorMappedAddress==TRUE)
      {
        stunDecodeXorAddress(&resp.xorMappedAddress, &mappedAddr[i], &resp.msgHdr.tr_id);
      }
      else
        mappedAddr[i] = resp.mappedAddress;
   }
	
    *mapAddr_rtp = mappedAddr[0];
    *mapAddr_rtcp = mappedAddr[1];
    *fd1 = fd[0];
    *fd2 = fd[1];

   for( i=0; i<NUM; i++)
   {
      closesocket( fd[i] );
   }
	
   return TRUE;
}

/* Local Variables:
   mode:c
   c-file-style:"ellemtel"
   c-file-offsets:((case-label . +))
   indent-tabs-mode:nil
   End:
*/

