 /*
  The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#if !defined(WIN32) && !defined(_WIN32_WCE)
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h> 
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <net/if.h>
#endif
#include <assert.h>

#include <time.h>

#if defined(WIN32) || defined(_WIN32_WCE)

#include <winsock2.h>
#include <stdlib.h>
/* #include <io.h> */

#else

#include <arpa/inet.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>

#endif

#include <string.h>

#include "ortp/stun_udp.h"
#include "ortp/ortp.h"

#if !defined(WIN32) && !defined(_WIN32_WCE)
int getErrno() { return errno; }
#else
int getErrno() { return WSAGetLastError(); }
#undef ENOTSOCK
#define ENOTSOCK      WSAENOTSOCK
#undef EAFNOSUPPORT
#define EAFNOSUPPORT  WSAEAFNOSUPPORT
#undef EADDRINUSE
#define EADDRINUSE    WSAEADDRINUSE
#undef EADDRNOTAVAIL
#define EADDRNOTAVAIL WSAEADDRNOTAVAIL
#undef ECONNRESET
#define ECONNRESET    WSAECONNRESET
#undef ECONNREFUSED
#define ECONNREFUSED  WSAECONNREFUSED
#undef EHOSTDOWN
#define EHOSTDOWN     WSAEHOSTDOWN
#undef EHOSTUNREACH
#define EHOSTUNREACH  WSAEHOSTUNREACH

#endif

bool_t
stunIPToStunIPAddress( const char* ipaddr, int srcport, StunIPAddress *cAddr)
{
  int counter=0;
  int res;
  struct addrinfo *addrinfo = NULL;
  struct addrinfo *curinfo;
  int family=AF_INET;
  if (strchr(ipaddr, ':')!=NULL)
    family=AF_INET6;
  
  res = stunGetaddrinfo (&addrinfo, ipaddr, srcport, 0, family);
  if (res)
    return FALSE;

  for (curinfo = addrinfo; curinfo; curinfo = curinfo->ai_next) {

    if (curinfo->ai_protocol!=0 && curinfo->ai_protocol != IPPROTO_UDP && curinfo->ai_protocol != IPPROTO_TCP) {
      continue;
    }
    if (curinfo->ai_family != AF_INET && curinfo->ai_family != AF_INET6) {
      continue;
    }
    if (curinfo->ai_family == AF_INET) {
      struct in_addr *sin_addr = &((struct sockaddr_in*) curinfo->ai_addr)->sin_addr;
      cAddr->family = IPv4Family;
      cAddr->addr.ipv4.addr = ntohl( sin_addr->s_addr );
      cAddr->addr.ipv4.port = srcport;
      counter++;
      break;
    }

    if (curinfo->ai_family == AF_INET6) {
      struct in6_addr *sin_addr = &((struct sockaddr_in6*) curinfo->ai_addr)->sin6_addr;
      cAddr->family = IPv6Family;
      memcpy(cAddr->addr.ipv6.addr , &sin_addr->s6_addr, sizeof(cAddr->addr.ipv6.addr) );
      cAddr->addr.ipv6.port = srcport;
      counter++;
      break;
    }
  }
  freeaddrinfo(addrinfo);
  if (counter>0)
    return TRUE;
  return FALSE;
}

bool_t
stunParseServerName( const char* name, StunIPAddress *addr_ipv4_udp, StunIPAddress *addr_ipv6_udp, StunIPAddress *addr_ipv4_tcp, StunIPAddress *addr_ipv6_tcp)
{
  char host[512];
  int port=3478;
  int counter=0;

  memset(host, 0, sizeof(host));

  if (name[0] == '[') {
    int port;
    const char * p = strchr(name, ']');
    size_t len = p ? (size_t) (p - name - 1): 0;
    if (p==NULL || len>sizeof(host))
      return FALSE;
    memcpy(host, name + 1, len);

    p++;
    if (p[0]==':') {
      p++;
      port = atoi(p);
      if (port>65535 || port<0)
        port=3478;
    }
  } else {
    const char * p = strchr(name, ':');
    size_t len = p ? (size_t) (p - name): strlen(name);
    if (len>sizeof(host))
      return FALSE;
    memcpy(host, name, len);
    if (p!=NULL) {
      p++;
      port = atoi(p);
      if (port>65535 || port<0)
        port=3478;
    }
  }
	
  {
    int res;
    struct addrinfo *addrinfo = NULL;
    struct addrinfo *curinfo;
    res = stunGetaddrinfo (&addrinfo, host, port, 0, PF_UNSPEC);
    if (res)
      return FALSE;

    for (curinfo = addrinfo; curinfo; curinfo = curinfo->ai_next) {

      if (curinfo->ai_protocol!=0 && curinfo->ai_protocol != IPPROTO_UDP && curinfo->ai_protocol != IPPROTO_TCP) {
        continue;
      }
      if (curinfo->ai_family != AF_INET && curinfo->ai_family != AF_INET6) {
        continue;
      }
      if (addr_ipv4_udp!=NULL && curinfo->ai_family == AF_INET) {
        struct in_addr *sin_addr = &((struct sockaddr_in*) curinfo->ai_addr)->sin_addr;
        addr_ipv4_udp->family = IPv4Family;
        addr_ipv4_udp->addr.ipv4.addr = ntohl( sin_addr->s_addr );
        addr_ipv4_udp->addr.ipv4.port = port;
        counter++;
      }
      if (addr_ipv4_tcp!=NULL && curinfo->ai_family == AF_INET) {
        struct in_addr *sin_addr = &((struct sockaddr_in*) curinfo->ai_addr)->sin_addr;
        addr_ipv4_tcp->family = IPv4Family;
        addr_ipv4_tcp->addr.ipv4.addr = ntohl( sin_addr->s_addr );
        addr_ipv4_tcp->addr.ipv4.port = port;
        counter++;
      }
      if (addr_ipv6_udp!=NULL && curinfo->ai_family == AF_INET6) {
        struct in6_addr *sin_addr = &((struct sockaddr_in6*) curinfo->ai_addr)->sin6_addr;
        addr_ipv6_udp->family = IPv6Family;
        memcpy(addr_ipv6_udp->addr.ipv6.addr , &sin_addr->s6_addr, sizeof(addr_ipv6_udp->addr.ipv6.addr) );
        addr_ipv6_udp->addr.ipv6.port = port;
        counter++;
      }
      if (addr_ipv6_tcp!=NULL && curinfo->ai_family == AF_INET6) {
        struct in6_addr *sin_addr = &((struct sockaddr_in6*) curinfo->ai_addr)->sin6_addr;
        addr_ipv6_tcp->family = IPv6Family;
        memcpy(addr_ipv6_tcp->addr.ipv6.addr , &sin_addr->s6_addr, sizeof(addr_ipv6_tcp->addr.ipv6.addr) );
        addr_ipv6_tcp->addr.ipv6.port = port;
        counter++;
      }
    }
    freeaddrinfo(addrinfo);
  }
  if (counter>0)
    return TRUE;
  return FALSE;
}

int
stunGetaddrinfo (struct addrinfo **addrinfo, const char *hostname, int port, int protocol, int family)
{
  struct addrinfo hints;
  char portbuf[10];
  int error;

  *addrinfo=NULL;
  if (hostname == NULL)
    return -1;

  snprintf (portbuf, sizeof (portbuf), "%i", port);

  memset (&hints, 0, sizeof (hints));

  hints.ai_flags = 0;

  hints.ai_family = family;
  //hints.ai_family = PF_UNSPEC;  /* ipv6 only support */
  //hints.ai_family = PF_INET6;  /* ipv6 only support */
  //hints.ai_family = PF_INET;  /* ipv4 only support */

  if (protocol == IPPROTO_UDP)
    hints.ai_socktype = SOCK_DGRAM;
  else
    hints.ai_socktype = SOCK_STREAM;

  hints.ai_protocol = protocol; /* IPPROTO_UDP or IPPROTO_TCP */
  error = getaddrinfo (hostname, portbuf, &hints, addrinfo);

  if (error || *addrinfo == NULL) {
#if defined(HAVE_RESOLV_H)
    /* When a DNS server has changed after roaming to a new network. The
       new one should be automatically used. However, a few system are not
       doing this automatically so, when the DNS server is not accessible,
       we force getaddrinfo to use the new one after the first failure. */
    if (error == EAI_AGAIN)
      res_init ();
#endif
    ortp_error ("STUN/getaddrinfo failure. [%s]:%s (%d)", hostname, portbuf, error);
    return -1;
  }
  else {
    struct addrinfo *elem;

    char tmp[INET6_ADDRSTRLEN];

    char porttmp[10];

    for (elem = *addrinfo; elem != NULL; elem = elem->ai_next) {
      getnameinfo (elem->ai_addr, (socklen_t)elem->ai_addrlen, tmp, sizeof (tmp), porttmp, sizeof (porttmp), NI_NUMERICHOST | NI_NUMERICSERV);
    }
  }

  return 0;
}

Socket
openPort( StunIPAddress *ipaddr )
{
  Socket fd;
  struct addrinfo *addrinfo;
  struct addrinfo *curinfo;
  char tmpip[512];
  int res;
  int family = 0;

  if (ipaddr->family==IPv4Family)
    family = AF_INET;
  else if (ipaddr->family==IPv6Family)
    family = AF_INET6;
  else
    return INVALID_SOCKET;

  fd = socket(family, SOCK_DGRAM, IPPROTO_UDP);
  if ( fd == INVALID_SOCKET )
  {
    ortp_error("stun_udp: Could not create a UDP socket");
    return INVALID_SOCKET;
  }

  stunIPAddress_ntop(ipaddr, tmpip, sizeof(tmpip));
  res = stunGetaddrinfo (&addrinfo, tmpip, ipaddr->addr.ipv4.port, IPPROTO_UDP, family);
  if (res) {
    closesocket(fd);
    return INVALID_SOCKET;
  }
  for (curinfo = addrinfo; curinfo; curinfo = curinfo->ai_next) {

    if (curinfo->ai_protocol != IPPROTO_UDP) {
      continue;
    }
    if (curinfo->ai_family != family) {
      continue;
    }

    if ( bind( fd,(struct sockaddr*)curinfo->ai_addr, curinfo->ai_addrlen) != 0 )
    {
      int e = getErrno();

      switch (e)
      {
      case 0:
        {
          ortp_error("stun_udp: Could not bind socket");
        }
        break;
      case EADDRINUSE:
        {
          ortp_error("stun_udp: Port %i for receiving UDP is in use", ipaddr->addr.ipv4.port);
        }
        break;
      case EADDRNOTAVAIL:
        {
          ortp_error("stun_udp: Cannot assign requested address");
        }
        break;
      default:
        {
#if !defined(_WIN32_WCE)
          ortp_error("stun_udp: Could not bind UDP receive port Error=%i %s",
            e, ortp_strerror(e));
#else
          ortp_error("stun_udp: Could not bind UDP receive port Error=%i",
            e);
#endif
        }
        break;
      }
      closesocket(fd);
      freeaddrinfo(addrinfo);
      return INVALID_SOCKET;
    }


    break;
  }


  ortp_debug("stun: opened port %i with fd %i", port, fd);
  freeaddrinfo(addrinfo);
  return fd;
}


bool_t 
getMessage( Socket fd, char* buf, int* len, StunIPAddress *fromIP)
{
   /* assert( fd != INVALID_SOCKET ); */
	
   int originalSize = *len;
   struct sockaddr_storage from;
   int fromLen = sizeof(struct sockaddr_storage);
   	

   int err;
   struct timeval tv;
   fd_set fdSet; 
#if defined(WIN32) || defined(_WIN32_WCE)
   unsigned int fdSetSize;
#else
   int fdSetSize;
#endif

   if (originalSize <= 0)
   {
	   return FALSE;
   }

   tv.tv_sec=1;
   tv.tv_usec=0; /* 150 ms */
   FD_ZERO(&fdSet); fdSetSize=0;
   FD_SET(fd,&fdSet); fdSetSize = fd+1;

   err = select(fdSetSize, &fdSet, NULL, NULL, &tv);
   if ( err == SOCKET_ERROR )
   {
      int e = getErrno();
      switch (e)
      {
         case ENOTSOCK:
            ortp_error("stun_udp: Error fd not a socket");
            break;
         case ECONNRESET:
            ortp_error("stun_udp: Error connection reset - host not reachable");
            break;
				
         default:
            ortp_error("stun_udp: Socket Error=%i", e);
      }
      return FALSE;
   }

    if (err==0)
    {
        ortp_error("stun_udp: Connection timeout with stun server!");
        *len = 0;
        return FALSE;
    }

    if (FD_ISSET (fd, &fdSet))
    {
        *len = (int)recvfrom(fd,
                        buf,
                        originalSize,
                        0,
                        (struct sockaddr *)&from,
                        (socklen_t*)&fromLen);
        	
        if ( *len == SOCKET_ERROR )
        {
            int e = getErrno();
        		
            switch (e)
            {
                case ENOTSOCK:
                    ortp_error("stun_udp: Error fd not a socket");
                    break;
                case ECONNRESET:
                    ortp_error("stun_udp: Error connection reset - host not reachable");
                    break;
        				
                default:
                  ortp_error("stun_udp: Socket Error=%i %s", e, ortp_strerror(e));
            }
        		
            return FALSE;
        }
        	
        if ( *len < 0 )
        {
            ortp_error("stun_udp: socket closed? negative len");
            return FALSE;
        }
            
        if ( *len == 0 )
        {
            ortp_error("stun_udp: socket closed? zero len");
            return FALSE;
        }

        if (from.ss_family==AF_INET) {
          struct in_addr *sin_addr = &((struct sockaddr_in*) &from)->sin_addr;
          fromIP->family = IPv4Family;
          fromIP->addr.ipv4.addr = ntohl( sin_addr->s_addr );
          fromIP->addr.ipv4.port = ntohs( ((struct sockaddr_in*) &from)->sin_port);
        } else {
          struct in6_addr *sin_addr = &((struct sockaddr_in6*) &from)->sin6_addr;
          fromIP->family = IPv6Family;
          memcpy(fromIP->addr.ipv6.addr , &sin_addr->s6_addr, sizeof(fromIP->addr.ipv6.addr) );
          fromIP->addr.ipv6.port = ntohs( ((struct sockaddr_in6*) &from)->sin6_port);
        }

        if ( (*len)+1 >= originalSize )
        {
            ortp_error("stun_udp: Received a message that was too large");
            return FALSE;
        }
        buf[*len]=0;
            
        return TRUE;
    }
    return FALSE;
}


bool_t 
sendMessage( Socket fd, char* buf, int l, StunIPAddress *dest)
             //unsigned int dstIp, unsigned short dstPort)
{
   int s;

   if (fd == INVALID_SOCKET)
	   return FALSE;

   if ( dest->addr.ipv4.port == 0 )
   {
      /* sending on a connected port */
      s = (int)send(fd,buf,l,0);
   }
   else
   {
      struct addrinfo *addrinfo;
      char tmpip[512];
      int res;

      stunIPAddress_ntop(dest, tmpip, sizeof(tmpip));
      res = stunGetaddrinfo (&addrinfo, tmpip, dest->addr.ipv4.port, IPPROTO_UDP, (dest->family==0x01)?AF_INET:AF_INET6);
      if (res) {
        return FALSE;
      }
        
      s = (int)sendto(fd, buf, l, 0,(struct sockaddr*)addrinfo->ai_addr, addrinfo->ai_addrlen);
   }
    
   if ( s == SOCKET_ERROR )
   {
      int e = getErrno();
      switch (e)
      {
         case ECONNREFUSED:
         case EHOSTDOWN:
         case EHOSTUNREACH:
         {
            /* quietly ignore this */
         }
         break;
         case EAFNOSUPPORT:
         {
            ortp_error("stun_udp: err EAFNOSUPPORT in send");
         }
         break;
         default:
         {
#if !defined(_WIN32_WCE)
            ortp_error("stun_udp: err %i %s in send", e, ortp_strerror(e));
#else
            ortp_error("stun_udp: err %i in send", e);
#endif
         }
      }
      return FALSE;
   }
    
   if ( s == 0 )
   {
      ortp_error("stun_udp: no data sent in send");
      return FALSE;
   }
    
   if ( s != l )
   {
      ortp_error("stun_udp: only %i out of %i bytes sent", s, l);
      return FALSE;
   }
    
   return TRUE;
}


void
initNetwork()
{
#if defined(WIN32) || defined(_WIN32_WCE)
   WORD wVersionRequested = MAKEWORD( 2, 0 );
   WSADATA wsaData;
   int err;
	
   err = WSAStartup( wVersionRequested, &wsaData );
   if ( err != 0 ) 
   {
      /* could not find a usable WinSock DLL */
      ortp_error("stun_udp: Could not load winsock");
   }
    
   /* Confirm that the WinSock DLL supports 2.2.*/
   /* Note that if the DLL supports versions greater    */
   /* than 2.2 in addition to 2.2, it will still return */
   /* 2.2 in wVersion since that is the version we      */
   /* requested.                                        */
    
   if ( LOBYTE( wsaData.wVersion ) != 2 ||
        HIBYTE( wsaData.wVersion ) != 2 ) 
   {
      /* Tell the user that we could not find a usable */
      /* WinSock DLL.                                  */
      WSACleanup( );
      ortp_error("stun_udp: Wrong winsock (!= 2.2) version");
   }    
#endif
}


/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

/* Local Variables:
   mode:c
   c-file-style:"ellemtel"
   c-file-offsets:((case-label . +))
   indent-tabs-mode:nil
   End:
*/
