/*
The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
Copyright (C) 2011-2012 Aymeric MOIZARD amoizard_at_osip.org

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ====================================================================
* The Vovida Software License, Version 1.0 
* 
* Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in
*    the documentation and/or other materials provided with the
*    distribution.
* 
* 3. The names "VOCAL", "Vovida Open Communication Application Library",
*    and "Vovida Open Communication Application Library (VOCAL)" must
*    not be used to endorse or promote products derived from this
*    software without prior written permission. For written
*    permission, please contact vocal@vovida.org.
*
* 4. Products derived from this software may not be called "VOCAL", nor
*    may "VOCAL" appear in their name, without prior written
*    permission of Vovida Networks, Inc.
* 
* THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
* NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
* NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
* IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
* 
* ====================================================================
* 
* This software consists of voluntary contributions made by Vovida
* Networks, Inc. and many individuals on behalf of Vovida Networks,
* Inc.  For more information on Vovida Networks, Inc., please see
* <http://www.vovida.org/>.
*
*/

#if defined(WIN32) || defined(_WIN32_WCE)
#include "ortp-config-win32.h"
#elif HAVE_CONFIG_H
#include "ortp-config.h"
#endif

#if (_MSC_VER >= 1700) && !defined(_USING_V110_SDK71_)
#include <winapifamily.h>
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#include <iostream>
#include <chrono>
#include <thread>
#endif
#endif

#include "ortp/ortp.h"
#include "utils.h"
#include "ortp/stun.h"
#include "ortp/turn.h"
#include "rtpsession_priv.h"

#include <assert.h>

#define TCP_PAD_BYTES (sizeof(uint16_t))
#ifdef _MSC_VER
#define FUNC __FUNCTION__
#else
#define FUNC __func__
#endif

#if !defined(WIN32) && !defined(_WIN32_WCE)
#include <netinet/tcp.h>
#include <netdb.h>
#ifndef closesocket
#define closesocket close
#endif
#endif


#if defined(WIN32) || defined(_WIN32_WCE)
#ifndef _WIN32_WCE
#include <errno.h>
#endif
#ifdef EWOULDBLOCK
#define is_would_block_error(errnum)	(errnum==WSAEWOULDBLOCK || errnum==EWOULDBLOCK)
#else
#define is_would_block_error(errnum)	(errnum==WSAEWOULDBLOCK)
#endif
#else
#define is_would_block_error(errnum)	(errnum==EWOULDBLOCK || errnum==EAGAIN)
#endif

#if 0	// Enable to get a detailed packet receive debugging log
#if !defined(WIN32) && !defined(_WIN32_WCE)
#define TRACEFILE "/tmp/recv.log"
#else
#define TRACEFILE "c:\\tmp\\recv.log"
#endif
#endif

#ifdef TRACEFILE
static struct timeval last_tv;
struct timeval tv, diff;
static FILE *f;
#endif

#ifdef __MACH__
#include <sys/types.h>
#include <sys/timeb.h>
#endif

static uint64_t turn_get_cur_time(void *unused){
#if defined(_WIN32_WCE)
  DWORD timemillis = GetTickCount();
  return timemillis;
#elif defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#if 1
  std::chrono::monotonic_clock::time_point val = std::chrono::monotonic_clock::now();
  std::chrono::milliseconds ms = std::chrono::duration_cast<std::chrono::milliseconds>(val.time_since_epoch());
  return ms.count();
#else
  return GetTickCount64();
#endif
#elif defined(WIN32)
  static uint32_t rollover_count=0;
  static DWORD last_t=0;
  DWORD t = timeGetTime();

  if ( last_t > t && (last_t - t > 100000) )
  {
    /* timeGetTime() may returns lower time than before because
    we are receiving values from different processors.
    */
    ortp_warning("turn_get_cur_time: rollover detected last_t = %u", last_t);
    ortp_warning("turn_get_cur_time: rollover detected cur_t  = %u", t);
    ++rollover_count;
  }
  last_t = t;
  return ((uint64_t)rollover_count<<32) + t;
#elif defined(__MACH__) && defined(__GNUC__) && (__GNUC__ >= 3)
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return (tv.tv_sec*1000LL) + (tv.tv_usec/1000LL);
#elif defined(__MACH__)
  struct timespec ts;
  struct timeb time_val;

  ftime (&time_val);
  ts.tv_sec = time_val.time;
  ts.tv_nsec = time_val.millitm * 1000000;
  return (ts.tv_sec*1000LL) + (ts.tv_nsec/1000000LL);
#else
  struct timespec ts;
  if (clock_gettime(CLOCK_MONOTONIC,&ts)<0){
    ortp_fatal("clock_gettime() doesn't work: %s",ortp_strerror(errno));
  }
  return (ts.tv_sec*1000LL) + (ts.tv_nsec/1000000LL);
#endif
}

// States during which we can read/write suceesfully
#define TurnSessionCanRecv(x) ((x)->state >= TurnSessionSentChanBind && (x)->state <= TurnSessionSentBindRefresh)
#define TurnSessionCanSend(x) ((x)->state >= TurnSessionStable && (x)->state <= TurnSessionSentBindRefresh)

// Determine how soon before the expiration we refresh allocations/channels
#define TURN_ALLOCATION_REFRESH_INTERVAL	60  /* seconds */
#define TURN_CHANNEL_REFRESH_INTERVAL		60  /* seconds */

static int turn_allocate(struct TurnSocketData *data, TurnAtrReservationToken* resToken, bool_t reservePortPair, int requestIpv6);
static int turn_flush_sendq(struct TurnSocketData *data, int flags);

static int
  turn_queue_refresh(struct TurnSocketData *data, int lifetime)
{
  StunAtrString username, *pUsername = NULL;
  StunAtrString password, *pPassword = NULL;
  StunAtrString nonce, *pNonce = NULL;
  StunAtrString realm, *pRealm = NULL;

  mblk_t *new_blk;

  ortp_warning("Refreshing allocation on fd %d...", data->fd);

  if (data->auth_username[0] != '\0') {
    snprintf(username.value, sizeof(username.value), "%s", data->auth_username);
    username.sizeValue = (uint16_t)strlen(username.value);
    pUsername = &username;
  }

  if (data->auth_password[0] != '\0') {
    snprintf(password.value, sizeof(password.value), "%s", data->auth_password);
    password.sizeValue = (uint16_t)strlen(password.value);
    pPassword = &password;
  }

  if (data->auth_realm[0] != '\0') {
    snprintf(realm.value, sizeof(realm.value), "%s", data->auth_realm);
    realm.sizeValue = (uint16_t)strlen(realm.value);
    pRealm = &realm;
  }

  if (data->auth_nonce[0] != '\0') {
    snprintf(nonce.value, sizeof(nonce.value), "%s", data->auth_nonce);
    nonce.sizeValue = (uint16_t)strlen(nonce.value);
    pNonce = &nonce;
  }

  data->refresh.requestlen = turnBuildRefresh(data->refresh.request, sizeof(data->refresh.request), pUsername, pPassword, pRealm, pNonce, lifetime, &data->refresh.tid);
  ortp_debug("%s: About to send Refresh msg of len %i to fd %d 'expires:%i", FUNC, data->refresh.requestlen, data->fd, lifetime);

  data->state = TurnSessionSentRefresh;

  new_blk = allocb(data->refresh.requestlen, 0);
  if (new_blk == NULL) {
    ortp_warning("%s: PANIC: out of memory for local buffers of %d bytes", FUNC, data->refresh.requestlen);
    data->state = TurnSessionFailed;
    return -1;
  }

  msgappend(new_blk, data->refresh.request, data->refresh.requestlen, FALSE);
  putq(&data->send_q, new_blk);
  return 1;
}

static int
  turn_queue_createpermission(struct TurnSocketData *data, int use_channelbind, unsigned int channel_bind, const char *rem_cand_conn_addr, int rem_cand_conn_port, struct StunTransaction *output)
{
  StunAtrString username, *pUsername = NULL;
  StunAtrString password, *pPassword = NULL;
  StunAtrString nonce, *pNonce = NULL;
  StunAtrString realm, *pRealm = NULL;

  StunIPAddress peerAddress;

  mblk_t *new_blk;
  int ret;

  ortp_warning("Refreshing channel binding on fd %d channel %x...", data->fd, channel_bind);

  if (data->fd==-1) {
    return -1;
  }
  if (channel_bind==0)
  {
    ortp_error("TODO ...");
    return -1;
  }
  if (data->auth_username[0] != '\0') {
    snprintf(username.value, sizeof(username.value), "%s", data->auth_username);
    username.sizeValue = (uint16_t)strlen(username.value);
    pUsername = &username;
  }

  if (data->auth_password[0] != '\0') {
    snprintf(password.value, sizeof(password.value), "%s", data->auth_password);
    password.sizeValue = (uint16_t)strlen(password.value);
    pPassword = &password;
  }

  if (data->auth_realm[0] != '\0') {
    snprintf(realm.value, sizeof(realm.value), "%s", data->auth_realm);
    realm.sizeValue = (uint16_t)strlen(realm.value);
    pRealm = &realm;
  }

  if (data->auth_nonce[0] != '\0') {
    snprintf(nonce.value, sizeof(nonce.value), "%s", data->auth_nonce);
    nonce.sizeValue = (uint16_t)strlen(nonce.value);
    pNonce = &nonce;
  }

  memset(&peerAddress, 0, sizeof(peerAddress));
  ret = stunIPToStunIPAddress(rem_cand_conn_addr, rem_cand_conn_port, &peerAddress);
  if (ret != TRUE) {
    ortp_error("turn: bad peer address for channel bind");
    return -1;
  }

  output->requestlen = turnBuildChannelBindOrPermission(use_channelbind, output->request, sizeof(output->request), pUsername, pPassword, 
    channel_bind, &peerAddress,
    pRealm, pNonce, &output->tid);
  ortp_debug("%s: About to send ChannelBind msg of len %i to fd %d", FUNC, output->requestlen, data->fd);
  data->state = TurnSessionSentBindRefresh;

  new_blk = allocb(output->requestlen, 0);
  if (new_blk == NULL) {
    ortp_warning("%s: PANIC: out of memory for local buffers of %d bytes", FUNC, output->requestlen);
    data->state = TurnSessionFailed;
    return -1;
  }

  msgappend(new_blk, output->request, output->requestlen, FALSE);
  putq(&data->send_q, new_blk);
  return 1;
}

static void 
  ice_sendtest( struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, struct CandidatePair *remote_candidate, Socket myFd, StunIPAddress *dest,
  const StunAtrString *username, const StunAtrString *password, 
  StunTrId *tid)
{	
  StunMessage req;
  bool_t res;

  memset(&req, 0, sizeof(StunMessage));

  stunBuildReqSimple( &req, username, FALSE, FALSE, 1);
  req.hasMessageIntegrity=TRUE;

  /* 7.1.1.1
  The attribute MUST be set equal to the priority that would be
  assigned, based on the algorithm in Section 4.1.2, to a peer
  reflexive candidate, should one be learned as a consequence of this
  check */
  req.hasPriority = TRUE;

  req.priority.priority = (110 << 24) | (255 << 16) | (255 << 8)
    | (256 - remote_candidate->remote_candidate.component_id);

  /* TODO: put this parameter only for the candidate selected */
  if (remote_candidate->nominated_flag==1 && remote_candidate->rem_controlling==0)
    req.hasUseCandidate = TRUE;

  if (remote_candidate->rem_controlling==1)
  {
    req.hasIceControlled = TRUE;
    req.iceControlled.value = checklist->tiebreak_value;
  }
  else
  {
    req.hasIceControlling = TRUE;
    req.iceControlling.value	= checklist->tiebreak_value;
  }

  /* TODO: not yet implemented? */
  req.hasFingerprint = TRUE;

  remote_candidate->connectivitycheck.requestlen = stunEncodeMessage( &req, remote_candidate->connectivitycheck.request, STUN_MAX_MESSAGE_SIZE, password );

  memcpy(tid , &(req.msgHdr.tr_id), sizeof(req.msgHdr.tr_id));

  if (strcasecmp(remote_candidate->local_candidate.cand_type, "relay")==0) {
    /* send indication */
    int len;
    mblk_t *output;
    if (remote_candidate->connectivitycheck.requestlen>sizeof(req.data.value)) {
      ortp_error("ice.c: data doesn't fit into send indication");
      return;
    }
    memset(&req, 0, sizeof(StunMessage));

    stunBuildReqSimple( &req, NULL, FALSE , FALSE , 0 );
    req.msgHdr.msgType= (TURN_INDICATION_SEND|STUN_INDICATION);

    req.hasXorPeerAddress=TRUE;
    stunEncodeXorAddress(dest, &req.xorPeerAddress, &req.msgHdr.tr_id);
    req.hasData = TRUE;
    memcpy( req.data.value, remote_candidate->connectivitycheck.request, remote_candidate->connectivitycheck.requestlen);
    req.data.sizeValue = remote_candidate->connectivitycheck.requestlen;
    len = (remote_candidate->connectivitycheck.requestlen + 51) & ~3; /* 20 (STUN HEADER) + 24 (XORPAd) + 4 (TA_DATA+len) + 3 (POSS. PADDING) = 51 */
    output = allocb(len, 0);
    len = stunEncodeMessage( &req, (char *)output->b_rptr, len, NULL );
    output->b_wptr = output->b_rptr+len;
    putq(&data->send_q, output);
    ortp_debug("ice.c: STUN REQ sent via using relay -> x:x id=%i", (int)(req.msgHdr.tr_id.octet[0]) );
    remote_candidate->num_occ_req_sent++;
    return;
  }

  res = sendMessage( myFd, remote_candidate->connectivitycheck.request, remote_candidate->connectivitycheck.requestlen, dest );
  if (res==TRUE) {
#define IP_UDP_OVERHEAD (20+8)
    struct timeval tv;
    gettimeofday(&tv, NULL);
    BandwidthStatistics_update(&session->rtp.sent_total_stun_bitrate, remote_candidate->connectivitycheck.requestlen+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);

    remote_candidate->num_occ_req_sent++;
  }
}

static ortp_socket_t _stun_choose_socket(RtpSession *session, bool_t isrtcp, uint8_t family) {
  ortp_socket_t *tables;
  int i;

  tables = session->rtp.socketv4;
  if (isrtcp==TRUE && !session->rtcp_mux)
    tables = session->rtcp.socketv4;

  for (i=0;i<10;i++) {
    if (sockisvalid(tables[i])) {
      struct sockaddr_storage ai_addr;
      socklen_t len = sizeof (struct sockaddr_storage);
      int res = getsockname (tables[i], (struct sockaddr *) &ai_addr, &len);
      if (res == 0) {
        if (ai_addr.ss_family == AF_INET && family==1)
          return tables[i];
        else if (ai_addr.ss_family == AF_INET6 && family==2)
          return tables[i];
        continue;
      }
    }
  }
  return tables[0];
}

static int ice_process_waiting_state(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, bool_t isrtcp, uint64_t ctime, struct CandidatePair *cand_pair) {

  StunIPAddress stunServerAddr;
  StunAtrString username;
  StunAtrString password;
  bool_t res;

  ortp_socket_t media_socket_tmp;

  cand_pair->connectivitycheck.state = ICE_IN_PROGRESS;
  cand_pair->connectivitycheck.nbretransmission=0;
  cand_pair->connectivitycheck.timeout=ctime+checklist->RTO;
  /* keep same rem_controlling for retransmission */
  cand_pair->rem_controlling = checklist->rem_controlling;

  username.sizeValue = 0;
  password.sizeValue = 0;

  /* username comes from "ice-ufrag" (rfrag:lfrag) */
  /* ufrag and pwd are in first row only */
  if (checklist->rem_ice_ufrag[0]=='\0')
    snprintf(username.value, sizeof(username.value), "iceu:%s", checklist->loc_ice_ufrag);
  else
    snprintf(username.value, sizeof(username.value), "%s:%s", checklist->rem_ice_ufrag, checklist->loc_ice_ufrag);
  username.sizeValue = (uint16_t)strlen(username.value);

  if (checklist->rem_ice_pwd[0]=='\0')
    snprintf(password.value, sizeof(password.value), "%s", "iced");
  else
    snprintf(password.value, sizeof(password.value), "%s", checklist->rem_ice_pwd);
  password.sizeValue = (uint16_t)strlen(password.value);

  res = stunIPToStunIPAddress(cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &stunServerAddr);
  if ( res == TRUE )
  {
    media_socket_tmp = _stun_choose_socket(session, isrtcp, stunServerAddr.family);
    ice_sendtest(data, session, checklist, cand_pair, media_socket_tmp, &stunServerAddr, &username, &password, &(cand_pair->connectivitycheck.tid));
    //if (cand_pair->nominated_flag==1 && cand_pair->rem_controlling==0) {
      ortp_message("ice.c(%x): STUN REQ (WAITING) (%8X) (%s) -> (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
        session,
        ntohl(*((uint32_t*)(cand_pair->connectivitycheck.tid.octet))),
        (cand_pair->nominated_flag==1 && cand_pair->rem_controlling==0)==FALSE?"":"USE-CANDIDATE",
        cand_pair->local_candidate.conn_addr,
        cand_pair->local_candidate.conn_port,
        (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
        cand_pair->local_candidate.rel_port,
        cand_pair->local_candidate.cand_type,
        cand_pair->remote_candidate.conn_addr,
        cand_pair->remote_candidate.conn_port,
        (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
        cand_pair->remote_candidate.rel_port,
        cand_pair->remote_candidate.cand_type,
        cand_pair->nominated_flag==0?"NO":"YES");
    //}
  }
  return 0;
}

static int ice_process_inprogress_state(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, bool_t isrtcp, uint64_t ctime, struct CandidatePair *cand_pair) {

  StunIPAddress destAddr;
  bool_t res;

  if (ctime<cand_pair->connectivitycheck.timeout) {
    return 0; /* not yet! */
  }

  if (cand_pair->connectivitycheck.nbretransmission>=7) {
    ortp_message("ice.c(%x): ICE_FAILED for candidate pair (NO ANSWER)! (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s)",
      session,
      cand_pair->local_candidate.conn_addr,
      cand_pair->local_candidate.conn_port,
      (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
      cand_pair->local_candidate.rel_port,
      cand_pair->local_candidate.cand_type,
      cand_pair->remote_candidate.conn_addr,
      cand_pair->remote_candidate.conn_port,
      (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
      cand_pair->remote_candidate.rel_port,
      cand_pair->remote_candidate.cand_type);

    cand_pair->connectivitycheck.state = ICE_FAILED;
    return 0;
  }

  cand_pair->connectivitycheck.nbretransmission++;
  cand_pair->connectivitycheck.timeout=ctime+checklist->RTO;

  res = stunIPToStunIPAddress(cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &destAddr);
  if ( res == TRUE ) {
    if (strcasecmp(cand_pair->local_candidate.cand_type, "relay")==0) {
      /* send indication */
      StunMessage req;
      int len;
      mblk_t *output;
      if (cand_pair->connectivitycheck.requestlen>sizeof(req.data.value)) {
        ortp_error("ice.c: data doesn't fit into send indication");
        return -1;
      }
      memset(&req, 0, sizeof(StunMessage));

      stunBuildReqSimple( &req, NULL, FALSE , FALSE , 0 );
      req.msgHdr.msgType= (TURN_INDICATION_SEND|STUN_INDICATION);

      req.hasXorPeerAddress=TRUE;
      stunEncodeXorAddress(&destAddr, &req.xorPeerAddress, &req.msgHdr.tr_id);
      req.hasData = TRUE;
      memcpy( req.data.value, cand_pair->connectivitycheck.request, cand_pair->connectivitycheck.requestlen);
      req.data.sizeValue = cand_pair->connectivitycheck.requestlen;
      len = (cand_pair->connectivitycheck.requestlen + 51) & ~3; /* 20 (STUN HEADER) + 24 (XORPAd) + 4 (TA_DATA+len) + 3 (POSS. PADDING) = 51 */
      output = allocb(len, 0);
      len = stunEncodeMessage( &req, (char *)output->b_rptr, len, NULL );
      output->b_wptr = output->b_rptr+len;
      putq(&data->send_q, output);
      ortp_debug("ice.c: STUN REQ sent via using relay -> x:x");
    } else {
      ortp_socket_t media_socket_tmp = _stun_choose_socket(session, isrtcp, destAddr.family);
      res = sendMessage( media_socket_tmp, cand_pair->connectivitycheck.request, cand_pair->connectivitycheck.requestlen, &destAddr);
      if (res==TRUE) {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        BandwidthStatistics_update(&session->rtp.sent_total_stun_bitrate, cand_pair->connectivitycheck.requestlen+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
      }
    }

    ortp_message("ice.c(%x): STUN REQ (INPROGRESS) (%8X) (%s) -> (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
      session,
      ntohl(*((uint32_t*)(cand_pair->connectivitycheck.tid.octet))),
      (cand_pair->nominated_flag==1 && cand_pair->rem_controlling==0)==FALSE?"":"USE-CANDIDATE",
      cand_pair->local_candidate.conn_addr,
      cand_pair->local_candidate.conn_port,
      (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
      cand_pair->local_candidate.rel_port,
      cand_pair->local_candidate.cand_type,
      cand_pair->remote_candidate.conn_addr,
      cand_pair->remote_candidate.conn_port,
      (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
      cand_pair->remote_candidate.rel_port,
      cand_pair->remote_candidate.cand_type,
      cand_pair->nominated_flag==0?"NO":"YES");
  }
  return 1;
}

static int ice_process_succeeded_state(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, bool_t isrtcp, uint64_t ctime, struct CandidatePair *cand_pair) {

  StunIPAddress stunServerAddr;
  StunAtrString username;
  StunAtrString password;
  bool_t res;

  if (cand_pair->nominated_flag == 0 && cand_pair->rem_controlling==0)
    cand_pair->nominated_flag = 1;
  if (cand_pair->nominated_flag == 0) {
    ortp_message("ice.c(%x): (SUCCEEDED/remote side hasn't sent a USE-CANDIDATE) (--) -> (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
      session,
      cand_pair->local_candidate.conn_addr,
      cand_pair->local_candidate.conn_port,
      (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
      cand_pair->local_candidate.rel_port,
      cand_pair->local_candidate.cand_type,
      cand_pair->remote_candidate.conn_addr,
      cand_pair->remote_candidate.conn_port,
      (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
      cand_pair->remote_candidate.rel_port,
      cand_pair->remote_candidate.cand_type,
      cand_pair->nominated_flag==0?"NO":"YES");
  }
  checklist->keepalive_time=ctime+15*1000;

  cand_pair->connectivitycheck.state = ICE_IN_PROGRESS;
  cand_pair->connectivitycheck.nbretransmission=0;
  cand_pair->connectivitycheck.timeout=ctime+checklist->RTO;
  /* keep same rem_controlling for retransmission */
  cand_pair->rem_controlling = checklist->rem_controlling;

  username.sizeValue = 0;
  password.sizeValue = 0;

  /* username comes from "ice-ufrag" (rfrag:lfrag) */
  /* ufrag and pwd are in first row only */
  if (checklist->rem_ice_ufrag[0]=='\0')
    snprintf(username.value, sizeof(username.value), "iceu:%s", checklist->loc_ice_ufrag);
  else
    snprintf(username.value, sizeof(username.value), "%s:%s", checklist->rem_ice_ufrag, checklist->loc_ice_ufrag);
  username.sizeValue = (uint16_t)strlen(username.value);

  if (checklist->rem_ice_pwd[0]=='\0')
    snprintf(password.value, sizeof(password.value), "%s", "iced");
  else
    snprintf(password.value, sizeof(password.value), "%s", checklist->rem_ice_pwd);
  password.sizeValue = (uint16_t)strlen(password.value);

  res = stunIPToStunIPAddress(cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &stunServerAddr);
  if ( res == TRUE )
  {
    uint32_t *val;
    ortp_socket_t media_socket_tmp = _stun_choose_socket(session, isrtcp, stunServerAddr.family);
    ice_sendtest(data, session, checklist, cand_pair, media_socket_tmp, &stunServerAddr, &username, &password, &(cand_pair->connectivitycheck.tid));
    val = (uint32_t*)(cand_pair->connectivitycheck.tid.octet);
    ortp_message("ice.c(%x): STUN REQ (SUCCEEDED) (%8X) (%s) -> (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
      session,
      ntohl(*val),
      (cand_pair->nominated_flag==1 && cand_pair->rem_controlling==0)==FALSE?"":"USE-CANDIDATE",
      cand_pair->local_candidate.conn_addr,
      cand_pair->local_candidate.conn_port,
      (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
      cand_pair->local_candidate.rel_port,
      cand_pair->local_candidate.cand_type,
      cand_pair->remote_candidate.conn_addr,
      cand_pair->remote_candidate.conn_port,
      (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
      cand_pair->remote_candidate.rel_port,
      cand_pair->remote_candidate.cand_type,
      cand_pair->nominated_flag==0?"NO":"YES");
  }
  return 0;
}

static int ice_process_pruned_state(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, bool_t isrtcp, uint64_t ctime, struct CandidatePair *cand_pair) {
  return ice_process_waiting_state(data, session, checklist, isrtcp, ctime, cand_pair);
}

static int ice_send_stun_request(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, bool_t isrtcp, uint64_t ctime)
{
  OList *candidate_pair_table;
  OList *elem;

  struct CandidatePair *cand_pair;
  bool_t res;
  int idx;

  if (checklist==NULL)
    return 0;

  /* prepare ONCE tie-break value */
  if (checklist->tiebreak_value==0) {
    checklist->tiebreak_value = random() * (0x7fffffffffffffffLL /0x7fff);
  }

  candidate_pair_table = checklist->cand_pairs;
  if (candidate_pair_table==NULL) {
    if (data->channelbind_number == 0) {
#if defined(_WIN32_WCE)
      time_t now = timeGetTime()/1000;
#else
      time_t now = time(NULL);
#endif
      data->channelbind_state=2;
      data->channelbind_number = 0x4000;
      data->channel_expires = now + TURN_PERMISSION_LIFETIME;
      turn_queue_createpermission(data, 1, 0x4000, checklist->remote_conn_ip, checklist->remote_conn_port, &data->createpermission);
    }
    return 0;
  }

#if 1
  if (data->fd!=-1) {
    cand_pair=NULL;
    idx=0;
    for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem),idx++){
      cand_pair=(struct CandidatePair*) elem->data;

      if (strcasecmp(cand_pair->local_candidate.cand_type, "relay")!=0)
        continue;
      if (cand_pair->channelbind_number == 0)
      {
#if defined(_WIN32_WCE)
        time_t now = timeGetTime()/1000;
#else
        time_t now = time(NULL);
#endif
        data->max_channelbind_number++;
        cand_pair->channelbind_number=data->max_channelbind_number;
        cand_pair->channelbind_state=1;
        cand_pair->channel_expires = now + TURN_PERMISSION_LIFETIME;
        turn_queue_createpermission(data, 0, cand_pair->channelbind_number, cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &cand_pair->createpermission);
        return 0; /* make it progressive: avoid burst */
      }
    }
  }
#endif

  cand_pair=NULL;
  idx=0;
  for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem),idx++){
    cand_pair=(struct CandidatePair*) elem->data;
    if (cand_pair->connectivitycheck.state == ICE_SUCCEEDED) {
      if (checklist->rem_controlling==1) {
        if (cand_pair->nominated_flag==1 && (checklist->nominated_pair_index<0 || checklist->nominated_pair_index>idx)) {
          checklist->nominated_pair_index = idx;
          ice_process_succeeded_state(data, session, checklist, isrtcp, ctime, cand_pair);
          break;
        }
#if 0
        else if (cand_pair->sccwuc_received){
          ice_process_succeeded_state(data, session, checklist, isrtcp, ctime, cand_pair);
          cand_pair->sccwuc_received = FALSE;
          break;
        }
#endif
      } else if (checklist->nominated_pair_index<0 || checklist->nominated_pair_index>idx) {
        checklist->nominated_pair_index = idx;
        ice_process_succeeded_state(data, session, checklist, isrtcp, ctime, cand_pair);
        break;
      }
    }
  }

  res=FALSE;
  cand_pair=NULL;
  for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem)){
    cand_pair=(struct CandidatePair*) elem->data;
    if (cand_pair->connectivitycheck.state == ICE_IN_PROGRESS)
    {
      ice_process_inprogress_state(data, session, checklist, isrtcp, ctime, cand_pair);
      res=TRUE;
    }
  }

  cand_pair=NULL;
  for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem)){
    cand_pair=(struct CandidatePair*) elem->data;
    if (cand_pair->connectivitycheck.state == ICE_WAITING)
    {
      ice_process_waiting_state(data, session, checklist, isrtcp, ctime, cand_pair);
      return 0; /* make it progressive: avoid burst */
    }
  }

  if (res==FALSE) {
    /* restart ICE_PRUNED request */
    cand_pair=NULL;
    idx=0;
    for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem),idx++){
      cand_pair=(struct CandidatePair*) elem->data;

      if (cand_pair->connectivitycheck.state == ICE_PRUNED)
      {
        ice_process_pruned_state(data, session, checklist, isrtcp, ctime, cand_pair);
      }
      if (checklist->nominated_pair_index>=0) {
        if (idx>=checklist->nominated_pair_index)
          break;
      }
    }
  }

  if (checklist->nominated_pair_index<0)
    return 0;

  /* send STUN indication each 15 seconds: keepalive */
  if (ctime>checklist->keepalive_time)
  {
    checklist->keepalive_time=ctime+15*1000;
    for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem)){
      cand_pair=(struct CandidatePair*) elem->data;
      if (cand_pair->connectivitycheck.state == ICE_SUCCEEDED)
      {
        StunIPAddress stunServerAddr;
        res = stunIPToStunIPAddress(cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &stunServerAddr);
        if ( res == TRUE )
        {
          StunMessage req;
          char buf[STUN_MAX_MESSAGE_SIZE];
          int len = STUN_MAX_MESSAGE_SIZE;
          memset(&req, 0, sizeof(StunMessage));
          stunBuildReqSimple( &req, NULL, FALSE, FALSE, 1);
          req.msgHdr.msgType = (STUN_METHOD_BINDING|STUN_INDICATION);
          req.hasFingerprint = TRUE;
          len = stunEncodeMessage( &req, buf, len, NULL);

          if (strcasecmp(cand_pair->local_candidate.cand_type, "relay")==0) {
            /* send indication */
            mblk_t *output;
            if (len>sizeof(req.data.value)) {
              ortp_error("ice.c: data doesn't fit into send indication");
              return 0;
            }
            memset(&req, 0, sizeof(StunMessage));

            stunBuildReqSimple( &req, NULL, FALSE , FALSE , 0 );
            req.msgHdr.msgType= (TURN_INDICATION_SEND|STUN_INDICATION);

            req.hasXorPeerAddress=TRUE;
            stunEncodeXorAddress(&stunServerAddr, &req.xorPeerAddress, &req.msgHdr.tr_id);
            req.hasData = TRUE;
            memcpy( req.data.value, buf, len);
            req.data.sizeValue = len;
            len = (len + 51) & ~3; /* 20 (STUN HEADER) + 24 (XORPAd) + 4 (TA_DATA+len) + 3 (POSS. PADDING) = 51 */
            output = allocb(len, 0);
            len = stunEncodeMessage( &req, (char *)output->b_rptr, len, NULL );
            output->b_wptr = output->b_rptr+len;
            putq(&data->send_q, output);
            ortp_debug("ice.c: STUN INDICATION sent via using relay -> x:x");
          } else {
            ortp_socket_t media_socket_tmp = _stun_choose_socket(session, isrtcp, stunServerAddr.family);
            res = sendMessage( media_socket_tmp, buf, len, &stunServerAddr );
            if (res==TRUE) {
              struct timeval tv;
              gettimeofday(&tv, NULL);
              BandwidthStatistics_update(&session->rtp.sent_total_stun_bitrate, len+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
            }
          }
        }
      }
    }
  }

  return 0;
}

static void
  _ice_createErrorResponse(StunMessage *response, int cl, int number, const char* msg)
{
  response->msgHdr.msgType = (STUN_METHOD_BINDING | STUN_ERR_RESP);
  response->hasErrorCode = TRUE;
  response->errorCode.errorClass = cl;
  response->errorCode.number = number;
  strncpy(response->errorCode.reason, msg, sizeof(response->errorCode.reason));
  response->errorCode.reason[sizeof(response->errorCode.reason)-1] = 0;
  response->errorCode.sizeReason = (uint16_t)strlen(response->errorCode.reason);
  response->hasFingerprint = TRUE;
}

static int ice_send_err_response(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, ortp_socket_t rtp_socket, StunMessage *msg, int cl, int number, const char *error_msg, StunIPAddress *remote_addr) {

  StunMessage resp;
  StunAtrString password;
  char buf[STUN_MAX_MESSAGE_SIZE];
  int len = sizeof(buf);
  int i;
  uint32_t *val = (uint32_t*)(msg->msgHdr.tr_id.octet);

  memset( &resp, 0 , sizeof(resp));

  resp.msgHdr.magic_cookie = ntohl(msg->msgHdr.magic_cookie);
  for (i=0; i<12; i++ )
  {
    resp.msgHdr.tr_id.octet[i] = msg->msgHdr.tr_id.octet[i];
  }
  memset(&password, 0, sizeof(StunAtrString));
  if (msg->hasUsername && msg->hasMessageIntegrity) {
    resp.hasMessageIntegrity = TRUE;
    resp.hasFingerprint = TRUE;
    resp.hasUsername = TRUE;
    memcpy(resp.username.value, msg->username.value, msg->username.sizeValue );
    resp.username.sizeValue = msg->username.sizeValue;

    snprintf(password.value, sizeof(password.value), "%s", checklist->loc_ice_pwd);
    password.sizeValue = (uint16_t)strlen(password.value);
  }
  {
    const char serverName[] = "antisipmediastreamer2 " STUN_VERSION;
    resp.hasSoftware = TRUE;
    memcpy( resp.softwareName.value, serverName, sizeof(serverName));
    resp.softwareName.sizeValue = sizeof(serverName);
  }
  ortp_error("ice.c(%x): (%8X) STUN REQ <- connectivity check error = %s", session, ntohl(*val), error_msg);
  _ice_createErrorResponse(&resp, cl, number, error_msg);
  len = stunEncodeMessage(&resp, buf, len, &password );
  if (len<=0)
    return -1;
  if (rtp_socket!=0) {
    bool_t res = sendMessage( rtp_socket, buf, len, remote_addr);
    if (res==TRUE) {
      struct timeval tv;
      gettimeofday(&tv, NULL);
      BandwidthStatistics_update(&session->rtp.sent_total_stun_bitrate, len+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
    }
    return 0;
  } else {
    /* send indication */
    StunMessage req;
    mblk_t *output;
    if (len>sizeof(req.data.value)) {
      ortp_error("ice.c: data doesn't fit into send indication");
      return -1;
    }
    memset(&req, 0, sizeof(StunMessage));

    stunBuildReqSimple( &req, NULL, FALSE , FALSE , 0 );
    req.msgHdr.msgType= (TURN_INDICATION_SEND|STUN_INDICATION);

    req.hasXorPeerAddress=TRUE;
    stunEncodeXorAddress(remote_addr, &req.xorPeerAddress, &req.msgHdr.tr_id);
    req.hasData = TRUE;
    memcpy( req.data.value, buf, len);
    req.data.sizeValue = len;
    len = (len + 51) & ~3; /* 20 (STUN HEADER) + 24 (XORPAd) + 4 (TA_DATA+len) + 3 (POSS. PADDING) = 51 */
    output = allocb(len, 0);
    len = stunEncodeMessage( &req, (char *)output->b_rptr, len, NULL );
    output->b_wptr = output->b_rptr+len;
    putq(&data->send_q, output);
  }

  return 0;
}

static int _ice_select_connection(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, ortp_socket_t rtp_socket, struct sockaddr *from, socklen_t fromlen) {
  /* choose to send from "default" ice candidate (the one from local sdp connection) */
  int idx=0;
  OList *elem;
  for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
    struct CandidatePair *cand_pair=(struct CandidatePair*) elem->data;
    if (cand_pair->nominated_flag!=0 && cand_pair->connectivitycheck.state == ICE_SUCCEEDED)
    {
      /* a selected connection has been choosen */
      if (strcasecmp(cand_pair->local_candidate.cand_type, "relay")==0) {
        int ret;
        data->use_relay=1;
        if (cand_pair->channelbind_number==0) {
          /* we need to assign it now, in order to use the channel (prflx relay address detected...) */
#if defined(_WIN32_WCE)
          time_t now = timeGetTime()/1000;
#else
          time_t now = time(NULL);
#endif
          data->max_channelbind_number++;
          cand_pair->channelbind_number=data->max_channelbind_number;
          cand_pair->channelbind_state=1;
          cand_pair->channel_expires = now + TURN_PERMISSION_LIFETIME;
          turn_queue_createpermission(data, 0, cand_pair->channelbind_number, cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &cand_pair->createpermission);
        }

        if (data->channelbind_number!=cand_pair->channelbind_number) {
          ortp_message("ice.c(%x): switching to selected pair for sending channel=%04X -> (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
            session,
            cand_pair->channelbind_number,
            (cand_pair->local_candidate.conn_addr[0]=='\0')?"-":cand_pair->local_candidate.conn_addr,
            cand_pair->local_candidate.conn_port,
            (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
            cand_pair->local_candidate.rel_port,
            cand_pair->local_candidate.cand_type,
            cand_pair->remote_candidate.conn_addr,
            cand_pair->remote_candidate.conn_port,
            (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
            cand_pair->remote_candidate.rel_port,
            cand_pair->remote_candidate.cand_type,
            cand_pair->nominated_flag==0?"NO":"YES");
          data->channelbind_number = cand_pair->channelbind_number;

          ret = stunIPToStunIPAddress(cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &data->peerAddress);
          if (ret != TRUE) {
            ortp_error("turn: bad peer address for channel bind");
            return -1;
          }
        }
      } else {

        if (rtp_socket==session->rtp.socketv4[0] || rtp_socket==session->rtp.socketv4[1]) {
          if (fromlen!=session->rtp.rem_addrlen
            || memcmp(&session->rtp.rem_addr,from, fromlen)!=0
            || data->use_relay==1)
            ortp_message("ice.c(%x): switching to selected pair for sending RTP -> (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
              session,
              (cand_pair->local_candidate.conn_addr[0]=='\0')?"-":cand_pair->local_candidate.conn_addr,
              cand_pair->local_candidate.conn_port,
              (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
              cand_pair->local_candidate.rel_port,
              cand_pair->local_candidate.cand_type,
              cand_pair->remote_candidate.conn_addr,
              cand_pair->remote_candidate.conn_port,
              (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
              cand_pair->remote_candidate.rel_port,
              cand_pair->remote_candidate.cand_type,
              cand_pair->nominated_flag==0?"NO":"YES");
          if (fromlen!=session->rtp.rem_addrlen
            || memcmp(&session->rtp.rem_addr,from, fromlen)!=0)
          {
            memcpy(&session->rtp.rem_addr, from, fromlen);
            session->rtp.rem_addrlen=fromlen;

            if (rtp_socket==session->rtp.socketv4[0])
              rtp_session_set_dscp(session, 0, -1);
            else if (rtp_socket==session->rtp.socketv4[1])
              rtp_session_set_dscp(session, 1, -1);
          }
        } else {
          if (fromlen!=session->rtcp.rem_addrlen
            || memcmp(&session->rtcp.rem_addr,from, fromlen)!=0
            || data->use_relay==1)
          {
            ortp_message("ice.c(%x): switching to selected pair for sending RTCP -> (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
              session,
              (cand_pair->local_candidate.conn_addr[0]=='\0')?"-":cand_pair->local_candidate.conn_addr,
              cand_pair->local_candidate.conn_port,
              (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
              cand_pair->local_candidate.rel_port,
              cand_pair->local_candidate.cand_type,
              cand_pair->remote_candidate.conn_addr,
              cand_pair->remote_candidate.conn_port,
              (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
              cand_pair->remote_candidate.rel_port,
              cand_pair->remote_candidate.cand_type,
              cand_pair->nominated_flag==0?"NO":"YES");
          }
          if (fromlen!=session->rtcp.rem_addrlen
            || memcmp(&session->rtcp.rem_addr,from, fromlen)!=0)
          {
            memcpy(&session->rtcp.rem_addr, from, fromlen);
            session->rtcp.rem_addrlen=fromlen;

            if (rtp_socket==session->rtcp.socketv4[0])
              rtp_session_set_dscp(session, 0, -1);
            else if (rtp_socket==session->rtcp.socketv4[1])
              rtp_session_set_dscp(session, 1, -1);
          }
        }
        if (data->use_relay==1) {
          ortp_message("ice.c(%x): switching and turning off relay");
        }
        data->use_relay=0;
      }
      rtp_session_set_flag (session, RTP_SESSION_DTLS_START);
      break;
    }
  }
  return 0;
}

static int ice_process_stun_request(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, ortp_socket_t rtp_socket, StunMessage *msg, mblk_t *mp, int msglen, struct sockaddr *from, socklen_t fromlen, char *src6host, int recvport)
{
  struct _OList *all_pairs=NULL;
  struct CandidatePair *cand_pair=NULL;
  OList *elem;

  StunMessage resp;
  StunIPAddress remote_addr;
  int i;

  uint32_t *val = (uint32_t*)(msg->msgHdr.tr_id.octet);

  memset( &resp, 0 , sizeof(resp));
  if (from->sa_family == AF_INET) {
    remote_addr.family = IPv4Family;
    remote_addr.addr.ipv4.addr = ntohl(((struct sockaddr_in *) from)->sin_addr.s_addr);
    remote_addr.addr.ipv4.port = ntohs(((struct sockaddr_in *) from)->sin_port);
  } else {
    struct in6_addr *sin_addr = &((struct sockaddr_in6*)from)->sin6_addr;
    remote_addr.family = IPv6Family;
    memcpy(remote_addr.addr.ipv6.addr , &sin_addr->s6_addr, sizeof(remote_addr.addr.ipv6.addr) );
    remote_addr.addr.ipv4.port = ntohs(((struct sockaddr_in6 *) from)->sin6_port);
  }
  resp.msgHdr.magic_cookie = ntohl(msg->msgHdr.magic_cookie);
  for (i=0; i<12; i++ )
  {
    resp.msgHdr.tr_id.octet[i] = msg->msgHdr.tr_id.octet[i];
  }

  /* check mandatory params */
  if (!msg->hasPriority)
  {
    ice_send_err_response(data, session, checklist, rtp_socket, msg, 4, 00, "Missing PRIORITY attribute", &remote_addr);
    return -1;
  }

  if (!msg->hasUsername)
  {
    ice_send_err_response(data, session, checklist, rtp_socket, msg, 4, 32, "Missing USERNAME attribute", &remote_addr);
    return -1;
  }
  if (!msg->hasMessageIntegrity)
  {
    ice_send_err_response(data, session, checklist, rtp_socket, msg, 4, 1, "Missing MESSAGEINTEGRITY attribute", &remote_addr);
    return -1;
  }

  /*
  The password associated with that transport address ID is used to verify
  the MESSAGE-INTEGRITY attribute, if one was present in the request.
  */
  {
    char hmac[20];
    /* remove length of fingerprint if present */
    if (msg->hasFingerprint==TRUE)
    {
      char *lenpos = (char *)mp->b_rptr + sizeof(uint16_t);
      uint16_t newlen = htons(msg->msgHdr.msgLength-8); /* remove fingerprint size */
      memcpy(lenpos, &newlen, sizeof(uint16_t));
      stunCalculateIntegrity_shortterm(hmac, (char*)mp->b_rptr, (int)(msglen-24-8), checklist->loc_ice_pwd);
    }
    else
      stunCalculateIntegrity_shortterm(hmac, (char*)mp->b_rptr, (int)(msglen-24), checklist->loc_ice_pwd);
    if (memcmp(msg->messageIntegrity.hash, hmac, 20)!=0)
    {
      ortp_warning("Wrong MESSAGEINTEGRITY attribute: skip request");
      /* Chrome is replying to the first STUN binding, without icepassword and ice-login */
      /* thus, the stun request has wrong integrity. */
      /* then chrome disable this binding... we temporarily accept the message until google fix it */
      /* ice_send_err_response(data, session, checklist, rtp_socket, msg, 4, 1, "Wrong MESSAGEINTEGRITY attribute", &remote_addr); */
      //return -1;
    }
    if (msg->hasFingerprint==TRUE)
    {
      char *lenpos = (char *)mp->b_rptr + sizeof(uint16_t);
      uint16_t newlen = htons(msg->msgHdr.msgLength); /* add back fingerprint size */
      memcpy(lenpos, &newlen, sizeof(uint16_t));
    }
  }

  resp.hasUsername = TRUE;
  memcpy(resp.username.value, msg->username.value, msg->username.sizeValue );
  resp.username.sizeValue = msg->username.sizeValue;

  resp.hasMessageIntegrity = TRUE;

  resp.hasFingerprint = TRUE;

  /* 7.2.1.1. Detecting and Repairing Role Conflicts */
  /* TODO */
  if (!msg->hasIceControlling && !msg->hasIceControlled)
  {
    ice_send_err_response(data, session, checklist, rtp_socket, msg, 4, 87, "Missing either ICE-CONTROLLING or ICE-CONTROLLED attribute", &remote_addr);
    return -1;
  }

  if (checklist->rem_controlling==0 && msg->hasIceControlling) {
    /* If the agent's tie-breaker is larger than or equal
    to the contents of the ICE-CONTROLLING attribute
    -> send 487, and do not change ROLE */
    if (checklist->tiebreak_value >= msg->iceControlling.value) {
      ice_send_err_response(data, session, checklist, rtp_socket, msg, 4, 87, "Role Conflict", &remote_addr);
      return -1;
    }
    else {
      for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
        uint64_t G;
        uint64_t D;
        cand_pair=(struct CandidatePair*) elem->data;
        /* controller agent */
        G = cand_pair->remote_candidate.priority;
        /* controlled agent */
        D = cand_pair->local_candidate.priority;
        cand_pair->pair_priority = (MIN(G, D))<<32 | (MAX(G, D))<<1 | (G>D?1:0);
      }
      checklist->rem_controlling = 1;
      /* reset all to initial WAITING state? */
      ortp_message("ice.c(%x): (%8X) STUN REQ <- tiebreaker -> reset all to ICE_WAITING state", session, ntohl(*val));
      for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
        cand_pair=(struct CandidatePair*) elem->data;
        if (cand_pair->connectivitycheck.state == ICE_PRUNED)
          continue;
        cand_pair->connectivitycheck.state = ICE_WAITING;
        memset(&cand_pair->connectivitycheck.tid , 0, sizeof(cand_pair->connectivitycheck.tid));
        cand_pair->connectivitycheck.timeout = 0;
        cand_pair->connectivitycheck.nbretransmission = 0;
      }
    }
  }

  if (checklist->rem_controlling==1 && msg->hasIceControlled) {

    /* If the agent's tie-breaker is larger than or equal
    to the contents of the ICE-CONTROLLED attribute
    -> change ROLE */
    if (checklist->tiebreak_value >= msg->iceControlled.value) {
      for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
        uint64_t G;
        uint64_t D;
        cand_pair=(struct CandidatePair*) elem->data;
        /* controller agent */
        G = cand_pair->local_candidate.priority;
        /* controlled agent */
        D = cand_pair->remote_candidate.priority;
        cand_pair->pair_priority = (MIN(G, D))<<32 | (MAX(G, D))<<1 | (G>D?1:0);
      }
      checklist->rem_controlling = 0;
      /* reset all to initial WAITING state? */
      ortp_message("ice.c(%x): (%8X) STUN REQ <- tiebreaker -> reset all to ICE_WAITING state", session, ntohl(*val));
      for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
        cand_pair=(struct CandidatePair*) elem->data;
        if (cand_pair->connectivitycheck.state == ICE_PRUNED)
          continue;
        cand_pair->connectivitycheck.state = ICE_WAITING;
        memset(&cand_pair->connectivitycheck.tid , 0, sizeof(cand_pair->connectivitycheck.tid));
        cand_pair->connectivitycheck.timeout = 0;
        cand_pair->connectivitycheck.nbretransmission = 0;
      }
    }
    else {
      ice_send_err_response(data, session, checklist, rtp_socket, msg, 4, 87, "Role Conflict", &remote_addr);
      return -1;
    }
  }

  cand_pair=NULL;

  {
    int idx;
    idx=0;
    for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
      cand_pair=(struct CandidatePair*) elem->data;
      /* connectivity check is coming from a known remote candidate?
      we should also check the port...
      */
      if (strcmp(cand_pair->remote_candidate.conn_addr, src6host)==0
        && cand_pair->remote_candidate.conn_port==recvport)
      {
        if (rtp_socket==0 && strcasecmp(cand_pair->local_candidate.cand_type, "relay")!=0) {
          /* STUN REQ coming from the relay */
          cand_pair=NULL;
          continue;
        } else if (rtp_socket!=0 && strcasecmp(cand_pair->local_candidate.cand_type, "relay")==0) {
          /* STUN REQ coming to local socket */
          cand_pair=NULL;
          continue;
        }

#if 0
        /* INTEROP WITH WEBRTC AGGRESSIVE NOMINATION: we must stop answering several STUN request all coming with USE-CANDIDATE: answer the best one only */
        if (msg->hasUseCandidate) {
          /* discard request with USE-CANDIDATE when we already received one with higher priority */
          struct CandidatePair *cand_pair_with_higher_priority;
          OList *elem_with_higher_priority;
          cand_pair_with_higher_priority=NULL;
          for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
            cand_pair_with_higher_priority=(struct CandidatePair*) elem->data;
            if (cand_pair_with_higher_priority==cand_pair)
              break;
            if (cand_pair_with_higher_priority->nominated_flag>0)
            {
              ortp_message("ice.c(%x): (%8X) STUN REQ (%s) <- %i (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s nc=%i:%i) from known peer: DISCARDED because of lower priority",
                session,
                ntohl(*val),
                msg->hasUseCandidate==0?"":"USE-CANDIDATE",
                idx,
                cand_pair->local_candidate.conn_addr,
                cand_pair->local_candidate.conn_port,
                (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
                cand_pair->local_candidate.rel_port,
                cand_pair->local_candidate.cand_type,
                cand_pair->remote_candidate.conn_addr,
                cand_pair->remote_candidate.conn_port,
                (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
                cand_pair->remote_candidate.rel_port,
                cand_pair->remote_candidate.cand_type,
                msg->hasNetworkCost==0?-1:msg->networkCost.networkid,
                msg->hasNetworkCost==0?-1:msg->networkCost.networkcost);
              return -1;
            }
          }
        }
#endif

        //if (msg->hasUseCandidate && !(session->flags & RTP_SESSION_DTLS_START)) {
          ortp_message("ice.c(%x): (%8X) STUN REQ (%s) <- %i (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s nc=%i:%i) from known peer",
            session,
            ntohl(*val),
            msg->hasUseCandidate==0?"":"USE-CANDIDATE",
            idx,
            cand_pair->local_candidate.conn_addr,
            cand_pair->local_candidate.conn_port,
            (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
            cand_pair->local_candidate.rel_port,
            cand_pair->local_candidate.cand_type,
            cand_pair->remote_candidate.conn_addr,
            cand_pair->remote_candidate.conn_port,
            (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
            cand_pair->remote_candidate.rel_port,
            cand_pair->remote_candidate.cand_type,
            msg->hasNetworkCost==0?-1:msg->networkCost.networkid,
            msg->hasNetworkCost==0?-1:msg->networkCost.networkcost);
        //}
        cand_pair->num_icc_req_received++;

        /* always enter "nominated_pair" when receiving USE-CANDIDATE and being controlled */
        if (msg->hasUseCandidate==TRUE && checklist->rem_controlling==1)
        {
          cand_pair->nominated_flag = 1;
          cand_pair->sccwuc_received = TRUE; /* send back a STUN REQUEST */
        }
        if (cand_pair->connectivitycheck.state==ICE_PRUNED
          || cand_pair->connectivitycheck.state==ICE_FROZEN
          || cand_pair->connectivitycheck.state==ICE_FAILED)
        {
          /* restart a new connectivity check */
          if (cand_pair->occ_is_wrong_pair) {
            /* skip anyway? */
          } else if (checklist->nominated_pair_index<0 || checklist->nominated_pair_index>idx)
            cand_pair->connectivitycheck.state = ICE_WAITING;
        }
        all_pairs = o_list_append(all_pairs, cand_pair);
      }
      cand_pair=NULL;
    }

#if 0
    /* fake support for ICE: just use symmetric signalling */
    if (checklist->rem_ice_ufrag[0]=='\0' || checklist->cand_pairs==NULL )
    {
      if (rtp_socket!=0) {
        if (rtp_socket==session->rtp.socket) {
          if (fromlen!=session->rtp.rem_addrlen
            || memcmp(&session->rtp.rem_addr,from, fromlen)!=0)
          {
            if (checklist->nominated_pair_index==0) {
              /* first one is selected, hopefully a good one? */
              ortp_message("ice.c(%x): (%8X) STUN REQ <- connectivity check received but remote app has ice disabled (symmetric RTP used) (%s:%i)", session, ntohl(*val), src6host, recvport);
              memcpy(&session->rtp.rem_addr, from, fromlen);
              session->rtp.rem_addrlen=fromlen;
              rtp_session_set_dscp(session, -1);
            }
            checklist->nominated_pair_index=-1;
          }
        } else {
          if (fromlen!=session->rtcp.rem_addrlen
            || memcmp(&session->rtcp.rem_addr,from, fromlen)!=0)
          {
            ortp_message("ice.c(%x): (%8X) STUN REQ <- connectivity check received but remote app has ice disabled (symmetric RTCP used) (%s:%i)", session, ntohl(*val), src6host, recvport);
            memcpy(&session->rtcp.rem_addr, from, fromlen);
            session->rtcp.rem_addrlen=fromlen;
            rtp_session_set_dscp(session, -1);
          }
        }
        ortp_message("ice.c(%x): (%8X) STUN REQ <- connectivity check received but remote app has ice disabled (disable relay) (%s:%i)", session, ntohl(*val), src6host, recvport);
        data->use_relay = 0;
      }
    }
#endif

    if (all_pairs==NULL && checklist->cand_pairs!=NULL) {
      struct CandidatePair *new_pair = ortp_new0(struct CandidatePair, 1);
      int idx;

      if (rtp_socket==0) {
        idx=0;
        for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
          cand_pair=(struct CandidatePair*) elem->data;
          if (strcasecmp(cand_pair->local_candidate.cand_type, "relay")==0)
            break;
          cand_pair=NULL;
        }
      }
      if (cand_pair==NULL)
        cand_pair=(struct CandidatePair*) checklist->cand_pairs->data;

      ortp_debug("ice.c(%x): (%8X) STUN REQ <- connectivity check received from an unknow candidate (%s:%i)", session, ntohl(*val), src6host, recvport);
      /* TODO: add the peer-reflexive candidate */

      if (rtp_socket!=0 && strcasecmp(cand_pair->local_candidate.cand_type, "host")!=0) {
        /* we have a peer candidate, but received on a (local?) candidate that wasn't advertised: */
        /* an rtpproxy may do this if we used the rtp_socket to send something there (STUN INDICATION FOR EXAMPLE) */
        /* let's skip it, because this use-case should not happen and there must be another path */
        ortp_message("ice.c(%x): (%8X) STUN REQ <- connectivity check received on wrong socket (%s:%i)", session, ntohl(*val), src6host, recvport);
        ortp_free(new_pair);
        return -1;
      }

      memcpy(&new_pair->local_candidate, &cand_pair->local_candidate, sizeof(new_pair->local_candidate));

      snprintf(new_pair->remote_candidate.foundation, sizeof(new_pair->remote_candidate.foundation), "%i", 12);
      new_pair->remote_candidate.component_id = cand_pair->remote_candidate.component_id;

      /* -> no known base address for peer */

      new_pair->remote_candidate.conn_port = recvport;
      snprintf(new_pair->remote_candidate.conn_addr, sizeof(new_pair->remote_candidate.conn_addr), "%s", src6host);

      /* take it from PRIORITY STUN attr */
      new_pair->remote_candidate.priority = msg->priority.priority;
      if (new_pair->remote_candidate.priority==0)
      {
        uint32_t type_preference = 110;
        uint32_t interface_preference = 255;
        uint32_t stun_priority=255;
        new_pair->remote_candidate.priority = (type_preference << 24) | (interface_preference << 16) | (stun_priority << 8)
          | (256 - new_pair->remote_candidate.component_id);
      }

      snprintf(new_pair->remote_candidate.cand_type, sizeof(new_pair->remote_candidate.cand_type), "prflx");
      snprintf (new_pair->remote_candidate.transport, sizeof (new_pair->remote_candidate.transport), "UDP");

      if (checklist->rem_controlling==0)
      {
        uint64_t G = new_pair->local_candidate.priority;
        /* controlled agent */
        uint64_t D = new_pair->remote_candidate.priority;
        new_pair->pair_priority = (MIN(G, D))<<32 | (MAX(G, D))<<1 | (G>D?1:0);
      }
      else
      {
        uint64_t G = new_pair->remote_candidate.priority;
        /* controlled agent */
        uint64_t D = new_pair->local_candidate.priority;
        new_pair->pair_priority = (MIN(G, D))<<32 | (MAX(G, D))<<1 | (G>D?1:0);
      }

      new_pair->connectivitycheck.state = ICE_SUCCEEDED;
      /* insert new pair candidate */

      if (msg->hasUseCandidate==TRUE && checklist->rem_controlling==1)
      {
        new_pair->nominated_flag = 1;
        new_pair->sccwuc_received = TRUE; /* send back a STUN REQUEST */
      }

      idx=0;
      for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
        cand_pair=(struct CandidatePair*) elem->data;

        /* if we already have a nominated pair: if the new pair is lower priority, should we mind about it?  */
        if (new_pair->pair_priority <= cand_pair->pair_priority) {

          if (elem->next!=NULL) {
            continue;
          }
          /* append at the end! **/
          idx++;
          checklist->cand_pairs = o_list_append(checklist->cand_pairs, new_pair);
        } else {
          /* move upper data */
          checklist->cand_pairs = o_list_insert(elem, new_pair);
        } 
        all_pairs = o_list_append(all_pairs, new_pair);
        if (checklist->nominated_pair_index>=idx)
          checklist->nominated_pair_index++;
        //if (msg->hasUseCandidate) {
          ortp_message("ice.c(%x): (%8X) STUN REQ (%s) <- %i (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s nc=%i:%i) new learned remote reflexive candidate",
            session,
            ntohl(*val),
            msg->hasUseCandidate==0?"":"USE-CANDIDATE",
            idx,
            new_pair->local_candidate.conn_addr,
            new_pair->local_candidate.conn_port,
            (new_pair->local_candidate.rel_addr[0]=='\0')?"-":new_pair->local_candidate.rel_addr,
            new_pair->local_candidate.rel_port,
            new_pair->local_candidate.cand_type,
            new_pair->remote_candidate.conn_addr,
            new_pair->remote_candidate.conn_port,
            "-",
            new_pair->remote_candidate.rel_port,
            new_pair->remote_candidate.cand_type,
            msg->hasNetworkCost==0?-1:msg->networkCost.networkid,
            msg->hasNetworkCost==0?-1:msg->networkCost.networkcost);

        //}
        break;
      }
      new_pair->num_icc_req_received++;
    }
  }

  resp.hasXorMappedAddress = TRUE;
  stunEncodeXorAddress(&remote_addr, &resp.xorMappedAddress, &resp.msgHdr.tr_id);

  resp.msgHdr.msgType = (STUN_METHOD_BINDING | STUN_SUCCESS_RESP);

  {
    const char serverName[] = "antisipmediastreamer2 " STUN_VERSION;
    resp.hasSoftware = TRUE;
    memcpy( resp.softwareName.value, serverName, sizeof(serverName));
    resp.softwareName.sizeValue = sizeof(serverName);
  }

  {
    StunAtrString password;
    char buf[STUN_MAX_MESSAGE_SIZE];
    int len = sizeof(buf);

    snprintf(password.value, sizeof(password.value), "%s", checklist->loc_ice_pwd);
    password.sizeValue = (uint16_t)strlen(password.value);

    len = stunEncodeMessage( &resp, buf, len, &password );
    if (len<=0) {
      if (all_pairs!=NULL) o_list_free(all_pairs);
      return -1;
    }

    if (rtp_socket!=0) {
      bool_t res;
      res = sendMessage( rtp_socket, buf, len, &remote_addr);
      if (res==TRUE) {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        BandwidthStatistics_update(&session->rtp.sent_total_stun_bitrate, len+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
      }
      if (res && all_pairs!=NULL) {
        int idx=0;
        for (elem=all_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
          cand_pair=(struct CandidatePair*) elem->data;
          cand_pair->num_icc_resp_sent++;
        }
        if (idx>1) {
          ortp_message("ice.c(%x): (%8X) STUN RESP -> sendMessage (apply to %i candidates) -> (%s:%i)", session, ntohl(*val),idx,  src6host, recvport);
        }
      }
      ortp_debug("ice.c(%x): (%8X) STUN RESP -> sendMessage (%s:%i)", session, ntohl(*val), src6host, recvport);
      if (all_pairs!=NULL) o_list_free(all_pairs);
      return 0;
    }else {
      /* send indication */
      StunMessage req;
      mblk_t *output;
      if (len>sizeof(req.data.value)) {
        ortp_error("ice.c: data doesn't fit into send indication");
        if (all_pairs!=NULL) o_list_free(all_pairs);
        return -1;
      }
      memset(&req, 0, sizeof(StunMessage));

      stunBuildReqSimple( &req, NULL, FALSE , FALSE , 0 );
      req.msgHdr.msgType= (TURN_INDICATION_SEND|STUN_INDICATION);

      req.hasXorPeerAddress=TRUE;
      stunEncodeXorAddress(&remote_addr, &req.xorPeerAddress, &req.msgHdr.tr_id);
      req.hasData = TRUE;
      memcpy( req.data.value, buf, len);
      req.data.sizeValue = len;
      len = (len + 51) & ~3; /* 20 (STUN HEADER) + 24 (XORPAd) + 4 (TA_DATA+len) + 3 (POSS. PADDING) = 51 */
      output = allocb(len, 0);
      len = stunEncodeMessage( &req, (char *)output->b_rptr, len, NULL );
      output->b_wptr = output->b_rptr+len;
      putq(&data->send_q, output);
      ortp_message("ice.c(%x): (%8X) STUN RESP -> using relay -> (%s:%i size=%i)", session, ntohl(*val), src6host, recvport, len);
      if (all_pairs!=NULL) {
        int idx=0;
        for (elem=all_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
          cand_pair=(struct CandidatePair*) elem->data;
          cand_pair->num_icc_resp_sent++;
        }
        if (idx>1) {
          ortp_message("ice.c(%x): (%8X) STUN RESP -> using relay (apply to %i candidates) -> (%s:%i size=%i)", session, ntohl(*val), idx, src6host, recvport, len);
        }
      }
    }
  }
  if (all_pairs!=NULL) o_list_free(all_pairs);
  return 0;
}

static int ice_process_stun_response(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, ortp_socket_t rtp_socket, StunMessage *resp, mblk_t *mp, int msglen, struct sockaddr *from, socklen_t fromlen, char *src6host, int recvport)
{
  OList *elem;
  /* set state to RECV-VALID or VALID */
  StunIPAddress mappedAddr;

  uint32_t *val = (uint32_t*)(resp->msgHdr.tr_id.octet);

  if (resp->hasXorMappedAddress!=TRUE)
  {
    ortp_error("ice.c(%x): (%8X) STUN RESP <- Missing XOR-MAPPED-ADDRESS in STUN answer.", session, ntohl(*val));
    return -1;
  }

  stunDecodeXorAddress(&resp->xorMappedAddress, &mappedAddr, &resp->msgHdr.tr_id);

  {
    char mapped_addr[66];
    struct CandidatePair *cand_pair=NULL;
    int idx;

    stunIPAddress_ntop(&mappedAddr, mapped_addr, sizeof(mapped_addr));

    idx=0;
    for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
      cand_pair=(struct CandidatePair*) elem->data;
      if (memcmp(&(cand_pair->connectivitycheck.tid), &(resp->msgHdr.tr_id), sizeof(resp->msgHdr.tr_id))==0)
      {
        break;
      }
      cand_pair = NULL;
    }

    if (cand_pair==NULL)
    {
      ortp_message("ice.c(%x): (%8X) STUN RESP <- no transaction for STUN answer?", session, ntohl(*val));
      return 0;
    }

    if (strcmp(src6host, cand_pair->remote_candidate.conn_addr)!=0
      || recvport!=cand_pair->remote_candidate.conn_port)
    {
      /* 7.1.2.2.  Success Cases
      -> must be a security issue: refuse non-symmetric answer */
      ortp_debug("ice.c(%x): (%8X) STUN RESP <- %i (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s refused because non-symmetric",
        session,
        ntohl(*val),
        idx,
        cand_pair->local_candidate.conn_addr,
        cand_pair->local_candidate.conn_port,
        (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
        cand_pair->local_candidate.rel_port,
        cand_pair->local_candidate.cand_type,
        cand_pair->remote_candidate.conn_addr,
        cand_pair->remote_candidate.conn_port,
        (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
        cand_pair->remote_candidate.rel_port,
        cand_pair->remote_candidate.cand_type,
        cand_pair->nominated_flag==0?"NO":"YES");
      cand_pair->connectivitycheck.state = ICE_FAILED;
      return 0;
    }

    /* Youhouhouhou */
    ortp_message("ice.c(%x): (%8X) STUN RESP <- %i (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
      session,
      ntohl(*val),
      idx,
      cand_pair->local_candidate.conn_addr,
      cand_pair->local_candidate.conn_port,
      (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
      cand_pair->local_candidate.rel_port,
      cand_pair->local_candidate.cand_type,
      cand_pair->remote_candidate.conn_addr,
      cand_pair->remote_candidate.conn_port,
      (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
      cand_pair->remote_candidate.rel_port,
      cand_pair->remote_candidate.cand_type,
      cand_pair->nominated_flag==0?"NO":"YES");
    cand_pair->num_occ_resp_received++;

    if (mappedAddr.addr.ipv4.port != cand_pair->local_candidate.conn_port
      || strcmp(mapped_addr, cand_pair->local_candidate.conn_addr)!=0)
    {
      /* a new srflx is discovered */
      struct CandidatePair *cand_pair2=NULL;

      cand_pair->occ_is_wrong_pair = TRUE;

      for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
        cand_pair2=(struct CandidatePair*) elem->data;
        if (mappedAddr.addr.ipv4.port == cand_pair2->local_candidate.conn_port
          && strcmp(mapped_addr, cand_pair2->local_candidate.conn_addr)==0
          && cand_pair->remote_candidate.conn_port == cand_pair2->remote_candidate.conn_port
          && strcmp(cand_pair->remote_candidate.conn_addr, cand_pair2->remote_candidate.conn_addr)==0)
        {
          if (cand_pair2->connectivitycheck.state==ICE_PRUNED
            ||cand_pair2->connectivitycheck.state==ICE_FROZEN
            ||cand_pair2->connectivitycheck.state==ICE_FAILED
            ||cand_pair2->connectivitycheck.state==ICE_IN_PROGRESS)
            cand_pair2->connectivitycheck.state = ICE_WAITING; /* trigger check */
          /* use-case: a srflx is talking to a srflx (and not a local to a srflx), so we disable the local->srflx and enable the srflx->srflx */
          ortp_message("ice.c(%x): (%8X) STUN RESP <- %i (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s XorMappedAddress matched another candidate -> enable it",
            session,
            ntohl(*val),
            idx,
            cand_pair2->local_candidate.conn_addr,
            cand_pair2->local_candidate.conn_port,
            (cand_pair2->local_candidate.rel_addr[0]=='\0')?"-":cand_pair2->local_candidate.rel_addr,
            cand_pair2->local_candidate.rel_port,
            cand_pair2->local_candidate.cand_type,
            cand_pair2->remote_candidate.conn_addr,
            cand_pair2->remote_candidate.conn_port,
            (cand_pair2->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair2->remote_candidate.rel_addr,
            cand_pair2->remote_candidate.rel_port,
            cand_pair2->remote_candidate.cand_type,
            cand_pair2->nominated_flag==0?"NO":"YES");

          if (cand_pair->nominated_flag > 0 && checklist->rem_controlling == 0) {
            if (checklist->nominated_pair_index == idx) {
              checklist->nominated_pair_index = -1; /* a new stun request will be sent on this pair */
              ortp_message("ice.c(%x) : (% 8X) restart nomination as the pair is false-positive",
                session,
                ntohl(*val),
                idx);
            }
          }
          break;
        }
        cand_pair2=NULL;
      }
      if (cand_pair2 == NULL && idx==0 && (strcasecmp(cand_pair->local_candidate.cand_type, "host") != 0)) {
        /* SPECIAL TRICK: if the first candidate is not "host", we are in relay only mode and we don't want to build a peer-reflexive candidate */
        ortp_warning("skip building a peer-reflexive candidate");
      } else if (cand_pair2 == NULL) {
        struct CandidatePair *new_pair = ortp_new0(struct CandidatePair, 1);
        int new_idx;

        /* 7.1.2.2.1.  Discovering Peer Reflexive Candidates */
        /* If IP & port were different than mappedAddr, there was A NAT
        between me and remote destination. */
        memcpy(&new_pair->remote_candidate, &cand_pair->remote_candidate, sizeof(new_pair->remote_candidate));

        snprintf(new_pair->local_candidate.foundation, sizeof(new_pair->local_candidate.foundation), "%i", 12);
        new_pair->local_candidate.component_id = cand_pair->local_candidate.component_id;

        /* what is my base address? */
        new_pair->local_candidate.rel_port = cand_pair->local_candidate.conn_port;
        snprintf(new_pair->local_candidate.rel_addr, sizeof(new_pair->local_candidate.rel_addr),
          "%s", cand_pair->local_candidate.conn_addr);

        new_pair->local_candidate.conn_port = mappedAddr.addr.ipv4.port;
        snprintf(new_pair->local_candidate.conn_addr, sizeof(new_pair->local_candidate.conn_addr),
          "%s", mapped_addr);

        new_pair->local_candidate.priority = (110 << 24) | (255 << 16) | (255 << 8)
          | (256 - new_pair->local_candidate.component_id);

        snprintf(new_pair->local_candidate.cand_type, sizeof(new_pair->local_candidate.cand_type),
          "prflx");
        snprintf (new_pair->local_candidate.transport,
          sizeof (new_pair->local_candidate.transport),
          "UDP");

        if (checklist->rem_controlling==0)
        {
          uint64_t G = new_pair->local_candidate.priority;
          /* controlled agent */
          uint64_t D = new_pair->remote_candidate.priority;
          new_pair->pair_priority = (MIN(G, D))<<32 | (MAX(G, D))<<1 | (G>D?1:0);
        }
        else
        {
          uint64_t G = new_pair->remote_candidate.priority;
          /* controlled agent */
          uint64_t D = new_pair->local_candidate.priority;
          new_pair->pair_priority = (MIN(G, D))<<32 | (MAX(G, D))<<1 | (G>D?1:0);
        }
        new_pair->connectivitycheck.state = ICE_WAITING;

        if (cand_pair->nominated_flag > 0 && checklist->rem_controlling == 0) {
          if (checklist->nominated_pair_index == idx) {
            checklist->nominated_pair_index = -1; /* a new stun request will be sent on this pair */
            ortp_message("ice.c(%x) : (% 8X) restart nomination as the pair is false-positive",
              session,
              ntohl(*val),
              idx);
            new_pair->connectivitycheck.state = ICE_SUCCEEDED; /* make it faster - keep nominated_flag=0 */
          }
        }
        new_idx=0;
        for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem), new_idx++){
          struct CandidatePair*cand_pair_tmp=(struct CandidatePair*) elem->data;

          if (new_pair->pair_priority <= cand_pair_tmp->pair_priority) {
            if (elem->next!=NULL) {
              continue;
            }
            /* append at the end! **/
            new_idx++;
            checklist->cand_pairs = o_list_append(checklist->cand_pairs, new_pair);
          } else {
            /* move upper data */
            checklist->cand_pairs = o_list_insert(elem, new_pair);
          }

          if (checklist->nominated_pair_index>= new_idx)
            checklist->nominated_pair_index++;
          ortp_message("ice.c(%x): (%8X) STUN RESP <- %i (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) new discovered local peer-reflexive candidate",
            session,
            ntohl(*val),
            new_idx,
            new_pair->local_candidate.conn_addr,
            new_pair->local_candidate.conn_port,
            (new_pair->local_candidate.rel_addr[0]=='\0')?"-":new_pair->local_candidate.rel_addr,
            new_pair->local_candidate.rel_port,
            new_pair->local_candidate.cand_type,
            new_pair->remote_candidate.conn_addr,
            new_pair->remote_candidate.conn_port,
            (new_pair->remote_candidate.rel_addr[0]=='\0')?"-":new_pair->remote_candidate.rel_addr,
            new_pair->remote_candidate.rel_port,
            new_pair->remote_candidate.cand_type);

          /* should we copy nominated_flag? */
          if (cand_pair->nominated_flag>0 && checklist->rem_controlling==1)
          {
            /* we previously received a USE-CANDIDATE request, but it wasn't accepted because xorMappedAddress was different. However, it was correct */
            new_pair->nominated_flag = 1;
            new_pair->sccwuc_received = TRUE; /* send back a STUN REQUEST */
          }
          break;
        }
      }
    }

    if (cand_pair->connectivitycheck.state != ICE_SUCCEEDED)
    {
      if (cand_pair->nominated_flag>0 && cand_pair->occ_is_wrong_pair==FALSE)
      {
        checklist->nominated_pair_index = idx;
        /* pair was previously nominated */
        ortp_message("ice.c(%x): (%8X) ICE CONCLUDED == %i (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
          session,
          ntohl(*val),
          idx,
          cand_pair->local_candidate.conn_addr,
          cand_pair->local_candidate.conn_port,
          (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
          cand_pair->local_candidate.rel_port,
          cand_pair->local_candidate.cand_type,
          cand_pair->remote_candidate.conn_addr,
          cand_pair->remote_candidate.conn_port,
          (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
          cand_pair->remote_candidate.rel_port,
          cand_pair->remote_candidate.cand_type,
          cand_pair->nominated_flag==0?"NO":"YES");
        cand_pair->connectivitycheck.state = ICE_SUCCEEDED;
        _ice_select_connection(data, session, checklist, rtp_socket, from, fromlen);
        return 0;
      } else if (checklist->loc_ice_pwd[0] == 'i' && checklist->loc_ice_pwd[1] == 'c' && checklist->loc_ice_pwd[2] == 'e' && checklist->loc_ice_pwd[3] == 'd' && (checklist->nominated_pair_index<0 || checklist->nominated_pair_index>idx)) {
        checklist->nominated_pair_index = idx;
        cand_pair->nominated_flag=1;
        /* pair was previously nominated */
        ortp_message("ice.c(%x): (%8X) ICE CONCLUDED(NRI) == %i (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s) nominated=%s",
          session,
          ntohl(*val),
          idx,
          cand_pair->local_candidate.conn_addr,
          cand_pair->local_candidate.conn_port,
          (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
          cand_pair->local_candidate.rel_port,
          cand_pair->local_candidate.cand_type,
          cand_pair->remote_candidate.conn_addr,
          cand_pair->remote_candidate.conn_port,
          (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
          cand_pair->remote_candidate.rel_port,
          cand_pair->remote_candidate.cand_type,
          cand_pair->nominated_flag==0?"NO":"YES");
        cand_pair->connectivitycheck.state = ICE_SUCCEEDED;
        _ice_select_connection(data, session, checklist, rtp_socket, from, fromlen);
        return 0;
      }

      cand_pair->connectivitycheck.state = ICE_FAILED;
      if (cand_pair->occ_is_wrong_pair==FALSE)
      {
        /* no peer-reflexive candidate was discovered */
        cand_pair->connectivitycheck.state = ICE_SUCCEEDED;
      }
    }
  }
  return 0;
}

static int ice_process_stun_error_response(RtpSession *session, struct IceCheckList *checklist, StunMessage *resp, mblk_t *mp, int msglen, struct sockaddr *from, socklen_t fromlen, char *src6host, int recvport)
{
  OList *elem;
  uint32_t *val = (uint32_t*)(resp->msgHdr.tr_id.octet);
  for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
    struct CandidatePair *cand_pair=(struct CandidatePair*) elem->data;

    if (memcmp(&(cand_pair->connectivitycheck.tid), &(resp->msgHdr.tr_id), sizeof(resp->msgHdr.tr_id))==0)
    {
      cand_pair->connectivitycheck.state = ICE_FAILED;
      ortp_message("ice.c(%x): (%8X) ERROR_RESPONSE: ICE_FAILED for candidate pair! (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s)",
        session,
        ntohl(*val),
        cand_pair->local_candidate.conn_addr,
        cand_pair->local_candidate.conn_port,
        (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
        cand_pair->local_candidate.rel_port,
        cand_pair->local_candidate.cand_type,
        cand_pair->remote_candidate.conn_addr,
        cand_pair->remote_candidate.conn_port,
        (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
        cand_pair->remote_candidate.rel_port,
        cand_pair->remote_candidate.cand_type);

      if (resp->hasErrorCode==TRUE && resp->errorCode.errorClass==4 && resp->errorCode.number==87)
      {
        if (cand_pair->rem_controlling==1)
        {
          for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
            struct CandidatePair *cand_pair2=(struct CandidatePair*) elem->data;
            /* controller agent */
            uint64_t G = cand_pair2->local_candidate.priority;
            /* controlled agent */
            uint64_t D = cand_pair2->remote_candidate.priority;
            cand_pair2->pair_priority = (MIN(G, D))<<32 | (MAX(G, D))<<1 | (G>D?1:0);
          }
          checklist->rem_controlling=0;
        }
        else
        {
          for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
            struct CandidatePair *cand_pair2=(struct CandidatePair*) elem->data;
            /* controller agent */
            uint64_t G = cand_pair2->remote_candidate.priority;
            /* controlled agent */
            uint64_t D = cand_pair2->local_candidate.priority;
            cand_pair2->pair_priority = (MIN(G, D))<<32 | (MAX(G, D))<<1 | (G>D?1:0);
          }
          checklist->rem_controlling=1;
        }
        /* reset all to initial WAITING state? */
        ortp_message("ice.c(%x): (%8X) ERROR_RESPONSE: 487 -> reset all to ICE_WAITING state", session, ntohl(*val));
        for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
          struct CandidatePair *cand_pair2=(struct CandidatePair*) elem->data;
          if (cand_pair2->connectivitycheck.state == ICE_PRUNED)
            continue;
          cand_pair2->connectivitycheck.state = ICE_WAITING;
          memset(&cand_pair2->connectivitycheck.tid , 0, sizeof(cand_pair2->connectivitycheck.tid));
          cand_pair2->connectivitycheck.timeout = 0;
          cand_pair2->connectivitycheck.nbretransmission = 0;
        }
      }
      break;
    }
  }
  return 0;
}

static int ice_process_stun_binding_indication(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, ortp_socket_t rtp_socket, struct sockaddr *from, socklen_t fromlen, char *src6host, int recvport)
{
  if (rtp_socket==0) {
    ortp_message("ice.c(%x): STUN INDICATION <- received from relay (%s:%i)", session, src6host, recvport);
  } else if (rtp_socket==session->rtp.socketv4[0] || rtp_socket==session->rtp.socketv4[1]) {
    /* remote side has no STUN support, but remote side was able
    to connect to us using the candidate we have provided.
    NOTE: ssrc is not validated here... */
    if (checklist->rem_ice_ufrag[0]=='\0' || checklist->cand_pairs==NULL )
    {
      if (data->stun_indication_received==0) {
        if (fromlen!=session->rtp.rem_addrlen
          || memcmp(&session->rtp.rem_addr, from, fromlen)!=0)
        {
          ortp_message("ice.c(%x): STUN INDICATION <- received but remote app has ice disabled (symmetric RTP used) (%s:%i)", session, src6host, recvport);
          memcpy(&session->rtp.rem_addr, from, fromlen);
          session->rtp.rem_addrlen=fromlen;
          rtp_session_set_dscp(session, 0, -1);
          data->stun_indication_received=1;
          data->use_relay = 0;
        }
      }
    }
  } else if (rtp_socket==session->rtcp.socketv4[0] || rtp_socket==session->rtcp.socketv4[1]) {
    /* remote side has no STUN support, but remote side was able
    to connect to us using the candidate we have provided.
    NOTE: ssrc is not validated here... */
    if (checklist->rem_ice_ufrag[0]=='\0' || checklist->cand_pairs==NULL )
    {
      if (data->stun_indication_received==0) {
        if (fromlen!=session->rtcp.rem_addrlen
          || memcmp(&session->rtcp.rem_addr, from, fromlen)!=0)
        {
          ortp_message("ice.c(%x): STUN INDICATION <- received but remote app has ice disabled (symmetric RTCP used) (%s:%i)", session, src6host, recvport);
          memcpy(&session->rtcp.rem_addr, from, fromlen);
          session->rtcp.rem_addrlen=fromlen;
          rtp_session_set_dscp(session, 0, -1);
          data->stun_indication_received=1;
          data->use_relay = 0;
        }
      }
    }
  }
  return 0;
}

static void dump_candidate_info(RtpSession *session, struct IceCheckList *checklist) {
  struct CandidatePair *cand_pair;
  OList *elem;
  int idx;
  cand_pair=NULL;
  idx=0;

  if (checklist->cand_pairs==NULL)
    return;

  ortp_message("ice.c(%x): CP checklist->nominated_pair_index=%i %s:%i", session, checklist->nominated_pair_index, checklist->remote_conn_ip, checklist->remote_conn_port);

  for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
    char status[32];
    cand_pair=(struct CandidatePair*) elem->data;

#define ICE_PRUNED -1
#define ICE_FROZEN 0
#define ICE_WAITING 1
#define ICE_IN_PROGRESS 2 /* STUN request was sent */
#define ICE_SUCCEEDED 3
#define ICE_FAILED 4 /* no answer or unrecoverable failure */

    if (cand_pair->connectivitycheck.state==ICE_PRUNED) {
      snprintf(status, sizeof(status), "%s", "ICE_PRUNED    ");
    } else if (cand_pair->connectivitycheck.state==ICE_FROZEN) {
      snprintf(status, sizeof(status), "%s", "ICE_FROZEN    ");
    } else if (cand_pair->connectivitycheck.state==ICE_WAITING) {
      snprintf(status, sizeof(status), "%s", "ICE_WAITING   ");
    } else if (cand_pair->connectivitycheck.state==ICE_IN_PROGRESS) {
      snprintf(status, sizeof(status), "%s", "ICE_INPROGRESS");
    } else if (cand_pair->connectivitycheck.state==ICE_SUCCEEDED) {
      snprintf(status, sizeof(status), "%s", "ICE_SUCCEEDED ");
    } else if (cand_pair->connectivitycheck.state==ICE_FAILED) {
      snprintf(status, sizeof(status), "%s", "ICE_FAILED    ");
    }

    ortp_message("ice.c(%x): CP chan=%04X %s nominated=%s occ=%4d:%4d icc=%4d:%4d (UC=%s) false-pos=%s (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s)",
      session,
      cand_pair->channelbind_number,
      status,
      cand_pair->nominated_flag==0?"NO ":"YES",
      cand_pair->num_occ_req_sent,
      cand_pair->num_occ_resp_received,
      cand_pair->num_icc_req_received,
      cand_pair->num_icc_resp_sent,
      cand_pair->sccwuc_received==FALSE?"NO ":"YES",
      cand_pair->occ_is_wrong_pair==0?"NO ":"YES",
      (cand_pair->local_candidate.conn_addr[0]=='\0')?"-":cand_pair->local_candidate.conn_addr,
      cand_pair->local_candidate.conn_port,
      (cand_pair->local_candidate.rel_addr[0]=='\0')?"-":cand_pair->local_candidate.rel_addr,
      cand_pair->local_candidate.rel_port,
      cand_pair->local_candidate.cand_type,
      cand_pair->remote_candidate.conn_addr,
      cand_pair->remote_candidate.conn_port,
      (cand_pair->remote_candidate.rel_addr[0]=='\0')?"-":cand_pair->remote_candidate.rel_addr,
      cand_pair->remote_candidate.rel_port,
      cand_pair->remote_candidate.cand_type
      );
  }

}
static int ice_process_stun_message(struct TurnSocketData *data, RtpSession *session, struct IceCheckList *checklist, ortp_socket_t rtp_socket, mblk_t *mp, int msglen, struct sockaddr *from, socklen_t fromlen)
{
  StunMessage msg;
  bool_t res;
  char src6host[NI_MAXHOST];
  int recvport = 0;
  int i;

  rtp_header_t *rtp=(rtp_header_t*)mp->b_rptr;
  uint16_t stunlen;
  if (rtp->version==2)
    return -10; /* rtp packet */

  /* try to see if it is a STUN packet */
  stunlen=*((uint16_t*)(mp->b_rptr + sizeof(uint16_t)));
  stunlen = ntohs(stunlen);
  if (stunlen+20!=msglen){
    return -12; /* not a stun packet */
  }

  memset( &msg, 0 , sizeof(msg) );
  res = stunParseMessage((char*)mp->b_rptr, msglen, &msg);
  if (!res)
  {
    ortp_error("ice.c(%x): Malformed STUN packet.", session);
    return -1;
  }

  if (checklist==NULL)
  {
    ortp_error("ice.c(%x): dropping STUN packet: ice is not configured", session);
    return -1;
  }

  /* prepare ONCE tie-break value */
  if (checklist->tiebreak_value==0) {
    checklist->tiebreak_value = random() * (0x7fffffffffffffffLL/0x7fff);
  }

  memset (src6host, 0, sizeof (src6host));

  {
    struct sockaddr_storage *aaddr = (struct sockaddr_storage *)from;
    if (aaddr->ss_family==AF_INET)
      recvport = ntohs (((struct sockaddr_in *) from)->sin_port);
    else
      recvport = ntohs (((struct sockaddr_in6 *) from)->sin6_port);
  }
  i = getnameinfo ((struct sockaddr*)from, fromlen,
    src6host, NI_MAXHOST,
    NULL, 0, NI_NUMERICHOST);
  if (i != 0)
  {
    ortp_error("ice.c(%x): Error with getnameinfo", session);
    return -1;
  }

  if (STUN_IS_INDICATION(msg.msgHdr.msgType)) {
    i = ice_process_stun_binding_indication(data, session, checklist, rtp_socket, from, fromlen, src6host, recvport);
    //dump_candidate_info(session, checklist);
    return i;
  }

  if (STUN_IS_REQUEST(msg.msgHdr.msgType)) {
    i = ice_process_stun_request(data, session, checklist, rtp_socket, &msg, mp, msglen, from, fromlen, src6host, recvport);
    dump_candidate_info(session, checklist);
    return i;
  }

  if (STUN_IS_SUCCESS_RESP(msg.msgHdr.msgType)) {
    i = ice_process_stun_response(data, session, checklist, rtp_socket, &msg, mp, msglen, from, fromlen, src6host, recvport);
    dump_candidate_info(session, checklist);
    return i;
  }

  if (STUN_IS_ERR_RESP(msg.msgHdr.msgType)) {
    i = ice_process_stun_error_response(session, checklist, &msg, mp, msglen, from, fromlen, src6host, recvport);
    dump_candidate_info(session, checklist);
    return i;
  }

  return 0;
}


#if 0
int
  rtp_turn_set_remote_addr(RtpTurnData *data, const char *raddr, int rport)
{
  if (turn_bind_chan(&data->rtp, raddr, rport, 0x4000)
    || turn_bind_chan(&data->rtcp, raddr, rport + 1, 0x4001))
    return -1;

  return 0;
}
#endif


static int
  turn_session_check_for_refresh(struct TurnSocketData *data, struct IceCheckList *checklist)
{
  OList *candidate_pair_table;
  OList *elem;
  int idx;

  time_t now;

  if (data == NULL)
    return -1;

#if defined(_WIN32_WCE)
  now = timeGetTime()/1000;
#else
  now = time(NULL);
#endif
  if (data->state == TurnSessionStable && data->alloc_expires - now < TURN_ALLOCATION_REFRESH_INTERVAL) {
    return turn_queue_refresh(data, TURN_DEFAULT_ALLOCATION_LIFETIME);
  }

  if (checklist==NULL)
    return -1;

  candidate_pair_table = checklist->cand_pairs;
  if (candidate_pair_table==NULL) {
    if (data->channelbind_state == 1 && data->channel_expires - now < TURN_CHANNEL_REFRESH_INTERVAL) {
      /* create permission */
      data->channel_expires = now + TURN_PERMISSION_LIFETIME;
      return turn_queue_createpermission(data, 0, 0x4000, checklist->remote_conn_ip, checklist->remote_conn_port, &data->createpermission);
    }
    else if (data->channelbind_state == 2 && data->channel_expires - now < TURN_CHANNEL_REFRESH_INTERVAL) {
      /* channel bind */
      data->channel_expires = now + TURN_PERMISSION_LIFETIME;
      return turn_queue_createpermission(data, 1, 0x4000, checklist->remote_conn_ip, checklist->remote_conn_port, &data->createpermission);
    }
    return 0;
  }

  /* else, do it for all candidate pair */
  idx=0;
  for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem),idx++){
    struct CandidatePair *cand_pair=(struct CandidatePair*) elem->data;

    if (strcasecmp(cand_pair->local_candidate.cand_type, "relay")!=0)
      continue;
    if (cand_pair->channelbind_state == 1 && cand_pair->channel_expires - now < TURN_CHANNEL_REFRESH_INTERVAL)
    {
      cand_pair->channel_expires = now + TURN_PERMISSION_LIFETIME;
      turn_queue_createpermission(data, 0, cand_pair->channelbind_number, cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &cand_pair->createpermission);
      return 0; /* make it progressive: avoid burst */
    }
    else if (cand_pair->channelbind_state == 2 && cand_pair->channel_expires - now < TURN_CHANNEL_REFRESH_INTERVAL)
    {
      cand_pair->channel_expires = now + TURN_PERMISSION_LIFETIME;
      turn_queue_createpermission(data, 1, cand_pair->channelbind_number, cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &cand_pair->createpermission);
      return 0; /* make it progressive: avoid burst */
    }
  }

  return 0;
}

static int
  turn_session_process_turn_message(struct TurnSocketData *data, struct IceCheckList *checklist, char* buf, int len, mblk_t *m, struct sockaddr *from, socklen_t *fromlen)
{
  int rc, err = -1;
  char* err_reason = NULL;
  StunMessage resp;

  if (!stunParseMessage(buf, len, &resp)) {
    ortp_warning("%s: stunParseMessage() failed: len = %d", FUNC, len);
    return -1;
  }

  if (STUN_IS_SUCCESS_RESP(resp.msgHdr.msgType) || STUN_IS_ERR_RESP(resp.msgHdr.msgType)) {
    uint32_t *val = (uint32_t*)(resp.msgHdr.tr_id.octet);

    if (resp.msgHdr.msgType==(TURN_METHOD_CHANNELBIND|STUN_SUCCESS_RESP)
      ||resp.msgHdr.msgType==(TURN_METHOD_CREATEPERMISSION|STUN_SUCCESS_RESP)) { /* answer for channel binding */
        unsigned int channelbind_number=0;
        int idx;
        OList *elem;
        struct CandidatePair *cand_pair=NULL;
        idx=0;
        for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
          cand_pair=(struct CandidatePair*) elem->data;
          if (memcmp(&(cand_pair->createpermission.tid), &(resp.msgHdr.tr_id), sizeof(resp.msgHdr.tr_id))==0)
          {
            break;
          }
          cand_pair = NULL;
        }



        if (!resp.hasLifetimeAttributes) {
          ortp_warning("%s: (%8X) got Refresh success WITHOUT lifetime!", FUNC, ntohl(*val));
          resp.lifetimeAttributes.lifetime = TURN_PERMISSION_LIFETIME;
        }

        if (cand_pair==NULL) {
          if (memcmp(&(data->createpermission.tid), &(resp.msgHdr.tr_id), sizeof(resp.msgHdr.tr_id))!=0) {
            ortp_warning("%s: (%8X) no matching tid for createpermission or channelbind response!", FUNC, ntohl(*val));
            return -1;
          }
#if defined(_WIN32_WCE)
          data->channel_expires = timeGetTime()/1000 + TURN_PERMISSION_LIFETIME;
#else
          data->channel_expires = time(NULL) + TURN_PERMISSION_LIFETIME;
#endif
        } else {
#if defined(_WIN32_WCE)
          cand_pair->channel_expires = timeGetTime()/1000 + TURN_PERMISSION_LIFETIME;
#else
          cand_pair->channel_expires = time(NULL) + TURN_PERMISSION_LIFETIME;
#endif
        }

        if (cand_pair!=NULL)
          channelbind_number = cand_pair->channelbind_number;
        else
          channelbind_number = data->channelbind_number;

        data->bad_nonce_count = 0;
        data->state = TurnSessionStable;
        ortp_warning("%s: (%8X) successful ChannelBind/CreatePermission on fd %d for channel %x with lifetime %d", FUNC, ntohl(*val), data->fd, channelbind_number, resp.lifetimeAttributes.lifetime);
        return 0;
    } else if (resp.msgHdr.msgType==(TURN_METHOD_CHANNELBIND|STUN_ERR_RESP)
      ||resp.msgHdr.msgType==(TURN_METHOD_CREATEPERMISSION|STUN_ERR_RESP)) { /* answer for channel binding */

        unsigned int channelbind_number=0;
        int idx;
        OList *elem;
        struct CandidatePair *cand_pair=NULL;
        idx=0;
        for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem),idx++){
          cand_pair=(struct CandidatePair*) elem->data;
          if (memcmp(&(cand_pair->createpermission.tid), &(resp.msgHdr.tr_id), sizeof(resp.msgHdr.tr_id))==0)
          {
            break;
          }
          cand_pair = NULL;
        }

        if (resp.hasErrorCode==TRUE) {
          err = (100 * resp.errorCode.errorClass) + resp.errorCode.number;
          err_reason = resp.errorCode.reason;
        }

        ortp_debug("%s: (%8X) got err %d in msg type %x on fd %d", FUNC, ntohl(*val), err, resp.msgHdr.msgType, data->fd);

        if (err == 438 && resp.hasNonce == TRUE && data->bad_nonce_count == 0) {

          memcpy(data->auth_nonce, resp.nonceName.value, resp.nonceName.sizeValue);
          data->auth_nonce[resp.nonceName.sizeValue] = '\0';
          data->bad_nonce_count++;

          rc = 0;
          if (cand_pair==NULL) {
            if (memcmp(&(data->createpermission.tid), &(resp.msgHdr.tr_id), sizeof(resp.msgHdr.tr_id))!=0) {
              ortp_warning("%s: (%8X) no matching tid for createpermission or channelbind response!", FUNC, ntohl(*val));
              return -1;
            }
            if (data->channelbind_state==1)
              rc = turn_queue_createpermission(data, 0, 0x4000, checklist->remote_conn_ip, checklist->remote_conn_port, &data->createpermission);
            else if (data->channelbind_state==2)
              rc = turn_queue_createpermission(data, 1, 0x4000, checklist->remote_conn_ip, checklist->remote_conn_port, &data->createpermission);
          } else {
            if (cand_pair->channelbind_state==1)
              rc = turn_queue_createpermission(data, 0, cand_pair->channelbind_number, cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &cand_pair->createpermission);
            else if (cand_pair->channelbind_state==2)
              rc = turn_queue_createpermission(data, 1, cand_pair->channelbind_number, cand_pair->remote_candidate.conn_addr, cand_pair->remote_candidate.conn_port, &cand_pair->createpermission);
          }
          if (rc)
            turn_flush_sendq(data, 0);
          return 0;
        }

        if (cand_pair!=NULL)
          channelbind_number = cand_pair->channelbind_number;
        else
          channelbind_number = data->channelbind_number;
        ortp_warning("%s: (%8X) unexpected response %x to ChannelBind/CreatePermission on fd %d for channel %x (err %d, reason %s)", FUNC, ntohl(*val), resp.msgHdr.msgType, data->fd, channelbind_number, err, err_reason ? err_reason : "<NONE>");
        data->state = TurnSessionFailed;
        return -1;
    } else if (resp.msgHdr.msgType==(TURN_METHOD_REFRESH|STUN_SUCCESS_RESP)) { /* answer for refresh request */

      if (!resp.hasLifetimeAttributes) {
        ortp_warning("%s: (%8X) got Refresh success WITHOUT lifetime!", FUNC, ntohl(*val));
        resp.lifetimeAttributes.lifetime = TURN_DEFAULT_ALLOCATION_LIFETIME;
      }

      if (memcmp(&(data->refresh.tid), &(resp.msgHdr.tr_id), sizeof(resp.msgHdr.tr_id))!=0) {
        ortp_warning("%s: (%8X) no matching tid for refresh response!", FUNC, ntohl(*val));
        return -1;
      }
#if defined(_WIN32_WCE)
      data->alloc_expires = timeGetTime()/1000 + resp.lifetimeAttributes.lifetime;
#else
      data->alloc_expires = time(NULL) + resp.lifetimeAttributes.lifetime;
#endif

      data->bad_nonce_count = 0;
      data->state = TurnSessionStable;
      ortp_warning("%s: (%8X) successful Refresh on fd %d with lifetime %d", FUNC, ntohl(*val), data->fd, resp.lifetimeAttributes.lifetime);
      return 0;
    } else if (resp.msgHdr.msgType==(TURN_METHOD_REFRESH|STUN_ERR_RESP)) { /* answer for allocate request */
      if (resp.hasErrorCode==TRUE) {
        err = (100 * resp.errorCode.errorClass) + resp.errorCode.number;
        err_reason = resp.errorCode.reason;
      }

      ortp_debug("%s: (%8X) got err %d in msg type %x on fd %d", FUNC, ntohl(*val), err, resp.msgHdr.msgType, data->fd);

      if (memcmp(&(data->refresh.tid), &(resp.msgHdr.tr_id), sizeof(resp.msgHdr.tr_id))!=0) {
        ortp_warning("%s: (%8X) no matching tid for refresh response!", FUNC, ntohl(*val));
        return -1;
      }

      if (err == 438 && resp.hasNonce == TRUE && data->bad_nonce_count == 0) {

        memcpy(data->auth_nonce, resp.nonceName.value, resp.nonceName.sizeValue);
        data->auth_nonce[resp.nonceName.sizeValue] = '\0';
        data->bad_nonce_count++;

        rc = turn_queue_refresh(data, TURN_DEFAULT_ALLOCATION_LIFETIME);
        if (rc)
          turn_flush_sendq(data, 0);
        return 0;
      }

      ortp_warning("%s: (%8X) unexpected response %x to ChannelBind/CreatePermission on fd %d (err %d, reason %s)", FUNC, ntohl(*val), resp.msgHdr.msgType, data->fd, err, err_reason ? err_reason : "<NONE>");
      data->state = TurnSessionFailed;
      return -1;
    } else if (resp.msgHdr.msgType==(TURN_MEDHOD_ALLOCATE|STUN_SUCCESS_RESP)) { /* answer for allocate request */
      ortp_warning("%s: (%8X) unexpected allocate sucesss response %x on fd %d, (err %d, reason %s)", FUNC, ntohl(*val), resp.msgHdr.msgType, data->fd, err, err_reason ? err_reason : "<NONE>");
      return 0;
    } else if (resp.msgHdr.msgType==(TURN_MEDHOD_ALLOCATE|STUN_ERR_RESP)) { /* answer for allocate request */
      ortp_warning("%s: (%8X) unexpected allocate error response %x on fd %d, (err %d, reason %s)", FUNC, ntohl(*val), resp.msgHdr.msgType, data->fd, err, err_reason ? err_reason : "<NONE>");
      return 0;
    } else {
      if (resp.hasErrorCode==TRUE) {
        err = (100 * resp.errorCode.errorClass) + resp.errorCode.number;
        err_reason = resp.errorCode.reason;
      }
      ortp_warning("%s: (%8X) unexpected stun response %x on fd %d, (err %d, reason %s)", FUNC, ntohl(*val), resp.msgHdr.msgType, data->fd, err, err_reason ? err_reason : "<NONE>");
      return 0;
    }
  } else if (STUN_IS_REQUEST(resp.msgHdr.msgType)) { /* request */
    if (resp.msgHdr.msgType==(STUN_METHOD_BINDING|STUN_REQUEST)) { /* binding request */
      ortp_warning("%s: unexpected stun binding on fd %d ", FUNC, data->fd);
      return 0;
    } else {
      ortp_warning("%s: unexpected stun request %x on fd %d, (err %d, reason %s)", FUNC, resp.msgHdr.msgType, data->fd, err, err_reason ? err_reason : "<NONE>");
      return 0;
    }
  } else { /* indication */
    if (resp.msgHdr.msgType==(TURN_INDICATION_DATA|STUN_INDICATION)) { /* binding indication */
      /* we have to deliver this data */
      if (resp.hasData) {
        if (resp.data.sizeValue>0) {
          memcpy(m->b_rptr, resp.data.value, resp.data.sizeValue);

          if (resp.hasXorPeerAddress) {
            StunIPAddress peerAddress;
            char tmpip[512];
            struct addrinfo *addrinfo;
            int res;
            stunDecodeXorAddress(&resp.xorPeerAddress, &peerAddress, &resp.msgHdr.tr_id);
            stunIPAddress_ntop(&peerAddress, tmpip, sizeof(tmpip));
            res = stunGetaddrinfo(&addrinfo, tmpip, peerAddress.addr.ipv4.port, 0, (resp.xorPeerAddress.family==0x01)?AF_INET:AF_INET6);
            if (res) {
              ortp_warning("%s: error in stun indication %x on fd %d, channel %x (reason: bad xorPeerAddress?)", FUNC, resp.msgHdr.msgType, data->fd, data->channelbind_number);
            } else {
              *fromlen = addrinfo->ai_addrlen;
              memcpy(from, addrinfo->ai_addr, addrinfo->ai_addrlen);
              freeaddrinfo(addrinfo);
            }
          }
          return resp.data.sizeValue;
        }
      }
    } else {
      ortp_warning("%s: unexpected stun indication %x on fd %d, channel %x (err %d, reason %s)", FUNC, resp.msgHdr.msgType, data->fd, data->channelbind_number, err, err_reason ? err_reason : "<NONE>");
      return 0;
    }
  }

  return 0;
}

static int
  turn_flush_sendq(struct TurnSocketData *data, int flags)
{
  mblk_t *mp;
  int len, slen, n;
  
  mp = peekq(&data->send_q);

  len = 0;
  while (mp != NULL) {
    mblk_t *tmp;

    msgpullup(mp, -1);
    slen = msgdsize(mp);
    n = (int)send(data->fd, (const char *)mp->b_rptr, slen, flags);
    if (n < 0) {
      int errnum = getSocketErrorCode();
      if (!is_would_block_error(errnum)) {
        ortp_warning("%s: socket error with %d/%d bytes sent: %s",
          FUNC, n, slen, getSocketError());
        
        tmp = getq(&data->send_q);
        freemsg(tmp);
        return -1;
      }

      n = 0;
    }

    len += n;
    mp->b_rptr += n; 

    /* Partial write, need more data to flush */
    if (n < slen) return len;

    tmp = getq(&data->send_q);
    freemsg(tmp);
    mp = peekq(&data->send_q);
  }

  return len;
}

static int
  rtp_refresh_turn(struct _RtpSession *session)
{
  RtpTurnData *_data = (RtpTurnData *)session->rtp.tr->data;
  struct TurnSocketData *data = (struct TurnSocketData *)&_data->rtp;
  if (session->rtp_icechecklist==NULL)
    return 0;

  {
    uint64_t now = turn_get_cur_time(NULL);
    if (now>_data->rtp_cur_time+20) {
      _data->rtp_cur_time = now;
      ice_send_stun_request(data, session, session->rtp_icechecklist, FALSE, _data->rtp_cur_time);
    }
  }

  turn_session_check_for_refresh(data, session->rtp_icechecklist);
  turn_flush_sendq(data, 0);
  return 0;
}

static int
  rtp_turn_sendto(struct _RtpSession *session, mblk_t *m, int flags, const struct sockaddr *to, socklen_t tolen)
{
  RtpTurnData *_data = (RtpTurnData *)session->rtp.tr->data;
  struct TurnSocketData *data = (struct TurnSocketData *)&_data->rtp;
  TurnChannelDataHdr chanDataHdr;
  int len;
  ortp_socket_t fd;
  mblk_t *header;

  rtp_refresh_turn(session);

  fd = data->fd;
  if (fd > 0) {

    if (data->use_relay==1 && !TurnSessionCanSend(data)) {
      ortp_warning("%s: discarding %d bytes of data since no TURN channel is active (state = %d)", FUNC, msgdsize(m), data->state);
      /* send data over TURN relay except RTP data */
      turn_flush_sendq(data, flags);
      return -1;
    } else if (data->use_relay==1) {

      if (data->peerAddress.addr.ipv4.port==0) {
        int ret;
        /* keep data->use_relay = 0; if it has been set by symmetric RTP operation */
        ret = stunIPToStunIPAddress(session->rtp_icechecklist->remote_conn_ip, session->rtp_icechecklist->remote_conn_port, &data->peerAddress);
        if (ret != TRUE) {
          ortp_error("turn: bad peer address for channel bind");
        }
      }

      if (0==1) {
        /* channel mode */

        // Save original (un-padded) length
        len = msgdsize(m);

        header = allocb(sizeof(chanDataHdr), 0);
        chanDataHdr.chanId = htons(data->channelbind_number);
        chanDataHdr.msgLength = htons(len);
        memcpy(header->b_rptr, &chanDataHdr, sizeof(chanDataHdr));
        header->b_wptr+=sizeof(chanDataHdr);
        concatb(header, dupb(m));
        // Pad the message to multiple of 4 bytes
        msgappend(header, NULL, 0, TRUE);

        // coalesce msg buffers
        msgpullup(header, -1);

        // Re-check length in case padding affected it
        len = msgdsize(header);
        putq(&data->send_q, header);
      } else {
        /* send indication */
        StunMessage req;
        mblk_t *output;
        len = msgdsize(m);
        if (len>sizeof(req.data.value)) {
          ortp_error("ice.c(%x): RTP data doesn't fit into send indication");
          return -1;
        }
        if (m->b_cont!=NULL)
          msgpullup(m,-1);
        memset(&req, 0, sizeof(StunMessage));

        stunBuildReqSimple( &req, NULL, FALSE , FALSE , 0 );
        req.msgHdr.msgType= (TURN_INDICATION_SEND|STUN_INDICATION);

        req.hasXorPeerAddress=TRUE;
        stunEncodeXorAddress(&data->peerAddress, &req.xorPeerAddress, &req.msgHdr.tr_id);
        req.hasData = TRUE;
        memcpy( req.data.value, m->b_rptr, len);
        req.data.sizeValue = len;
        len = (len + 51) & ~3; /* 20 (STUN HEADER) + 24 (XORPAd) + 4 (TA_DATA+len) + 3 (POSS. PADDING) = 51 */
        output = allocb(len, 0);
        len = stunEncodeMessage( &req, (char *)output->b_rptr, len, NULL );
        output->b_wptr = output->b_rptr+len;
        putq(&data->send_q, output);
      }

      /* send RTP data and other data over TURN relay  */
      return turn_flush_sendq(data, flags);
    }

    /* send data over TURN relay except RTP data */
    turn_flush_sendq(data, flags);
  }


  if (data->use_relay==0) {
    ortp_socket_t media_socket_tmp=-1;
    int error;
    if (m->b_cont!=NULL)
      msgpullup(m,-1);

#ifdef FAKE_LOST
    {
      int val = rand() % 100 + 1;
      if  (val < 2 ) {
        ortp_warning("introduce lost packet");
        return (int) (m->b_wptr - m->b_rptr);
      }
    }
#endif
    if (to->sa_family==AF_INET)
      media_socket_tmp = _stun_choose_socket(session, FALSE, 1);
    else
      media_socket_tmp = _stun_choose_socket(session, FALSE, 2);
    error = sendto (media_socket_tmp, (char*)m->b_rptr, (int) (m->b_wptr - m->b_rptr), 0,to,tolen);
    if (error>0 && session->FEC_N_interval>0) {
      if (session->rtp.stats.packet_sent>0 && session->rtp.stats.packet_sent%session->FEC_N_interval==0)
        sendto(media_socket_tmp, (char*)m->b_rptr, (int)(m->b_wptr - m->b_rptr), 0, to, tolen);
    }
    return error;
  }

  /* we must use a relay, but none exist? */
  return -1;
}

static int
  rtcp_refresh_turn(struct _RtpSession *session)
{
  RtpTurnData *_data = (RtpTurnData *)session->rtp.tr->data;
  struct TurnSocketData *data = (struct TurnSocketData *)&_data->rtcp;
  struct IceCheckList *icechecklist = session->rtcp_icechecklist;

  if (session->rtp_icechecklist==NULL)
    return 0;

  if (!session->rtcp_mux) {
    uint64_t now = turn_get_cur_time(NULL);
    if (now>_data->rtcp_cur_time+20) {
      _data->rtcp_cur_time = now;
      ice_send_stun_request(data, session, icechecklist, TRUE, _data->rtcp_cur_time);
    }
  }

  if (session->rtcp_mux) {
    /* rtp-mux is enabled: use same socket */
    data = (struct TurnSocketData *)&_data->rtp;
    icechecklist = session->rtp_icechecklist;
  }

  turn_session_check_for_refresh(data, icechecklist);
  turn_flush_sendq(data, 0);
  return 0;
}

static int
  rtcp_turn_sendto(struct _RtpSession *session, mblk_t *m, int flags, const struct sockaddr *to, socklen_t tolen)
{
  RtpTurnData *_data = (RtpTurnData *)session->rtp.tr->data;
  struct TurnSocketData *data = (struct TurnSocketData *)&_data->rtcp;
  struct IceCheckList *icechecklist = session->rtcp_icechecklist;
  TurnChannelDataHdr chanDataHdr;
  int len;
  ortp_socket_t fd;
  mblk_t *header;

  if (session->rtcp_mux) {
    /* rtp-mux is enabled: use same socket */
    data = (struct TurnSocketData *)&_data->rtp;
    icechecklist = session->rtp_icechecklist;
  }

  rtcp_refresh_turn(session);

  fd = data->fd;
  if (fd > 0) {

    if (data->use_relay==1 && !TurnSessionCanSend(data)) {
      ortp_warning("%s: discarding %d bytes of data since no TURN channel is active (state = %d)", FUNC, msgdsize(m), data->state);
      /* send data over TURN relay except RTP data */
      turn_flush_sendq(data, flags);
      return -1;
    } else if (data->use_relay==1) {

      if (data->peerAddress.addr.ipv4.port==0) {
        int ret;
        /* keep data->use_relay = 0; if it has been set by symmetric RTP operation */
        ret = stunIPToStunIPAddress(icechecklist->remote_conn_ip, icechecklist->remote_conn_port, &data->peerAddress);
        if (ret != TRUE) {
          ortp_error("turn: bad peer address for channel bind");
          return -1;
        }
      }

      if (0==1) {
        /* channel mode */

        // Save original (un-padded) length
        len = msgdsize(m);

        header = allocb(sizeof(chanDataHdr), 0);
        chanDataHdr.chanId = htons(data->channelbind_number);
        chanDataHdr.msgLength = htons(len);
        memcpy(header->b_rptr, &chanDataHdr, sizeof(chanDataHdr));
        header->b_wptr+=sizeof(chanDataHdr);
        concatb(header, dupb(m));
        // Pad the message to multiple of 4 bytes
        msgappend(header, NULL, 0, TRUE);

        // coalesce msg buffers
        msgpullup(header, -1);

        // Re-check length in case padding affected it
        len = msgdsize(header);
        putq(&data->send_q, header);
      } else {
        /* send indication */
        StunMessage req;
        mblk_t *output;
        len = msgdsize(m);
        if (len>sizeof(req.data.value)) {
          ortp_error("ice.c(%x): RTP data doesn't fit into send indication");
          return -1;
        }
        if (m->b_cont!=NULL)
          msgpullup(m,-1);
        memset(&req, 0, sizeof(StunMessage));

        stunBuildReqSimple( &req, NULL, FALSE , FALSE , 0 );
        req.msgHdr.msgType= (TURN_INDICATION_SEND|STUN_INDICATION);

        req.hasXorPeerAddress=TRUE;
        stunEncodeXorAddress(&data->peerAddress, &req.xorPeerAddress, &req.msgHdr.tr_id);
        req.hasData = TRUE;
        memcpy( req.data.value, m->b_rptr, len);
        req.data.sizeValue = len;
        len = (len + 51) & ~3; /* 20 (STUN HEADER) + 24 (XORPAd) + 4 (TA_DATA+len) + 3 (POSS. PADDING) = 51 */
        output = allocb(len, 0);
        len = stunEncodeMessage( &req, (char *)output->b_rptr, len, NULL );
        output->b_wptr = output->b_rptr+len;
        putq(&data->send_q, output);
      }

      /* send RTP data and other data over TURN relay  */
      return turn_flush_sendq(data, flags);
    }

    /* send data over TURN relay except RTP data */
    turn_flush_sendq(data, flags);
  }


  if (data->use_relay==0) {
    ortp_socket_t media_socket_tmp=-1;
    int error;
    if (m->b_cont!=NULL)
      msgpullup(m,-1);
    if (to->sa_family==AF_INET)
      media_socket_tmp = _stun_choose_socket(session, TRUE, 1);
    else
      media_socket_tmp = _stun_choose_socket(session, TRUE, 2);
    error = sendto (media_socket_tmp, (char*)m->b_rptr, (int) (m->b_wptr - m->b_rptr), 0,to,tolen);
    return error;
  }

  /* we must use a relay, but none exist? */
  return -1;
}

static int
  turn_extract_rtp_packet(struct TurnSocketData *data, struct IceCheckList *checklist, mblk_t *m, struct sockaddr *from, socklen_t *fromlen)
{
  /* unfragment data->recv_m */
  //msgpullup(data->recv_m, -1);

  // If we're not in the middle of a packet, extract the header and go
  // from there..
  if (data->recv_need == 0) {
    TurnChannelDataHdr chanDataHdr;
    if (data->recv_left < sizeof(TurnChannelDataHdr)) return 0;

    memcpy(&chanDataHdr, data->recv_m->b_rptr, sizeof(chanDataHdr));
    chanDataHdr.chanId = ntohs(chanDataHdr.chanId);
    chanDataHdr.msgLength = ntohs(chanDataHdr.msgLength);

    // If the high bit is set, this is neither a standard STUN message
    // nor a valid TURN ChannelData message; skip through the data 
    // until we find a a header that could pass for one of the above.
    //
    if (chanDataHdr.chanId & 0x8000) {
      int skiplen = 0;

      do {
        skiplen += sizeof(TurnChannelDataHdr);
        data->recv_left -= sizeof(TurnChannelDataHdr);
        data->recv_m->b_rptr += sizeof(TurnChannelDataHdr);

        if (data->recv_left < sizeof(TurnChannelDataHdr))
          return 0;

        memcpy(&chanDataHdr, data->recv_m->b_rptr, sizeof(chanDataHdr));
        chanDataHdr.chanId = ntohs(chanDataHdr.chanId);
        chanDataHdr.msgLength = ntohs(chanDataHdr.msgLength);
      } while ((chanDataHdr.chanId & 0x8000) == 0 &&
        ((chanDataHdr.chanId & 0x4000) == 0x4000 ||
        (chanDataHdr.msgLength & 0x03) == 0));

      ortp_warning("%s: received unexpected message starting with high bit set, skipped %d bytes", FUNC, skiplen);
    }

    if ((chanDataHdr.chanId & 0x4000) == 0x4000) {

      if (chanDataHdr.chanId==0x4000) {
        struct sockaddr_in *addr;
        addr = (struct sockaddr_in*)from;
        addr->sin_addr.s_addr = inet_addr (checklist->remote_conn_ip);
        addr->sin_port = htons(checklist->remote_conn_port);
        addr->sin_family = AF_INET;
        *fromlen = sizeof (struct sockaddr_in);
      } else {
        OList *elem;
        struct CandidatePair *cand_pair=NULL;
        for (elem=checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
          cand_pair=(struct CandidatePair*) elem->data;
          if (cand_pair->channelbind_number==chanDataHdr.chanId)
          {
            struct sockaddr_in *addr;
            addr = (struct sockaddr_in*)from;
            addr->sin_addr.s_addr = inet_addr (cand_pair->remote_candidate.conn_addr);
            addr->sin_port = htons(cand_pair->remote_candidate.conn_port);
            addr->sin_family = AF_INET;
            *fromlen = sizeof (struct sockaddr_in);
            break;
          }
          cand_pair = NULL;
        }
        if (cand_pair==NULL) {
          data->skipping=1;
          ortp_warning("%s: chn_id %x doesn't match any channel, skipping %d bytes", FUNC, chanDataHdr.chanId, data->recv_need);
        }
      }
      chanDataHdr.msgLength += sizeof(chanDataHdr);
    } else {
      chanDataHdr.msgLength += sizeof(StunMsgHdr);
      data->have_turn_msg = 1;
    }

    if (!data->skipping && chanDataHdr.msgLength > 1500) {
      ortp_warning("%s: len too big, skipping %d bytes", FUNC, chanDataHdr.msgLength);
      data->skipping = 1;
    }

    // NB: over TCP all channel data messages must be padded to
    // multiple of 4 bytes; normal STUN/TURN mesages are also a
    // multiple of 4 bytes since the header is 20 bytes and all
    // attributes are required to be padded to 4 bytes.
    data->recv_size = chanDataHdr.msgLength;
    data->recv_need = (chanDataHdr.msgLength + 3) & ~3;
  }

  // Now check to see if we can pass a complete packet up the stack
  if (data->recv_need > 0 && data->recv_left >= data->recv_need) {
    int sz = 0;

    if (!data->skipping) {
      if (data->have_turn_msg) {
        sz = turn_session_process_turn_message(data, checklist, (char*)data->recv_m->b_rptr, data->recv_size, m, from, fromlen);
        if (sz>0) {
        } else {
          ortp_message("%s: received TURN message on fd %d, len %d", FUNC, data->fd, data->recv_size);
          sz=0;
        }
      } else {
        sz = data->recv_size - sizeof(TurnChannelDataHdr);
        msgappend(m, (char*)data->recv_m->b_rptr + sizeof(TurnChannelDataHdr), sz, FALSE);
        msgpullup(m, -1);
        m->b_wptr = m->b_rptr;
#ifdef TRACEFILE
        if (f && data->channelbind_number == 0x4000) fprintf(f, "%s: %lu.%06ld %lu.%06ld extract %d (skipping %d, left: was %d, will be %d)\n", FUNC, tv.tv_sec, tv.tv_usec, diff.tv_sec, diff.tv_usec, sz, data->skipping, data->recv_left, data->recv_left - data->recv_need);
#endif
      }
    }

    data->recv_m->b_rptr += data->recv_need;
    data->recv_left -= data->recv_need;
    
    /* any race condition if we use memmove here? */
    /* use maximum buffer size */
    if (data->recv_left>0) {
      memmove(data->recv_m->b_datap->db_base, data->recv_m->b_rptr, data->recv_left);
      data->recv_m->b_wptr-=(data->recv_m->b_rptr-data->recv_m->b_datap->db_base);
    }
    data->recv_m->b_rptr = data->recv_m->b_datap->db_base;

    data->recv_size = data->recv_need = 0;
    data->have_turn_msg = 0;
    data->skipping = 0;
    return sz;
  }

  return 0;
}

static int
  rtp_turn_recvfrom(struct _RtpSession *session, mblk_t *m, int flags, struct sockaddr *from, socklen_t *fromlen)
{
  int error;
  RtpTurnData *_data = (RtpTurnData *)session->rtp.tr->data;
  struct TurnSocketData *data = (struct TurnSocketData *)&_data->rtp;

  int r;
  int i;
  ortp_socket_t fd = data->fd;

#ifdef TRACEFILE
  if (!f) { f = fopen(TRACEFILE,"w"); }
  if (f && data->channelbind_number == 0x4000) { 
    gettimeofday(&tv,NULL);
    if (last_tv.tv_sec > 0) {
      diff.tv_sec = tv.tv_sec - last_tv.tv_sec;
      diff.tv_usec = tv.tv_usec - last_tv.tv_usec;
      if (diff.tv_usec < 0) {
        diff.tv_usec += 1000000;
        --diff.tv_sec;
      }
    } else {
      diff.tv_sec = diff.tv_usec = 0;
    }
    last_tv.tv_sec = tv.tv_sec;
    last_tv.tv_usec = tv.tv_usec;
  }
#endif

  {
    uint64_t now = turn_get_cur_time(NULL);
    if (now>_data->rtp_cur_time+20) {
      _data->rtp_cur_time = now;
      ice_send_stun_request(data, session, session->rtp_icechecklist, FALSE, _data->rtp_cur_time);
    }
  }

  for (i=0;i<10;i++) {
    if (sockisvalid(session->rtp.socketv4[i])) {
      int bufsz=(int) (m->b_datap->db_lim - m->b_datap->db_base);
      error = (int)recvfrom(session->rtp.socketv4[i], (char*)m->b_wptr,
        bufsz, 0,
        (struct sockaddr *) from,
        fromlen);
      if (error>0) {
        ice_process_stun_message(data, session, session->rtp_icechecklist, session->rtp.socketv4[i], m, error, from, *fromlen);
        return error; /* data available */
      }
    }
  }

  if (data->fd == -1)
    return error;

  *fromlen=0;

  /* Check if we need to refresh, and if so, send the queued requests */
  turn_session_check_for_refresh(data, session->rtp_icechecklist);
  turn_flush_sendq(data, 0);

  // First check if we can pull any complete messages out of buffered receive data
  if (data->recv_left > 0) {
    r = turn_extract_rtp_packet(data, session->rtp_icechecklist, m, from, fromlen);
#ifdef TRACEFILE
    if (f && data->channelbind_number == 0x4000 && r > 0) fprintf(f, "%s: %lu.%06ld %lu.%06ld no-recv-return %d\n", FUNC, tv.tv_sec, tv.tv_usec, diff.tv_sec, diff.tv_usec, r);
#endif
    if (r > 0) {
      ice_process_stun_message(data, session, session->rtp_icechecklist, 0, m, r, from, *fromlen);
      return r;
    }
  }

  if (!TurnSessionCanRecv(data)) {
    static int val=0;
    if (val%200==0)
      ortp_debug("%s: ignoring read since no TURN channel is active (state = %d)", FUNC, data->state);
    val++;
    return 0;
  }

  // Rewind buffer back to beginning if we don't have any data in it..
  if (data->recv_left == 0) {
    data->recv_m->b_rptr = data->recv_m->b_wptr = data->recv_m->b_datap->db_base;
  }
  if (data->recv_left > 0 && data->recv_need>0 && data->recv_need>data->recv_left) {
    if (data->recv_m->b_datap->db_lim-data->recv_m->b_wptr < data->recv_need-data->recv_left) {
      /* we don't have enough buffer to read the complete TURN packet */
      int old_size = (int)(data->recv_m->b_datap->db_lim-data->recv_m->b_datap->db_base);
      int new_size = old_size + data->recv_need;
      mblk_t * new_recv_m = allocb(new_size, 0);
      ortp_message("ice.c(%x): increase TURN buffer size (old=%i new=%i)", session, old_size, new_size);
      memcpy(new_recv_m->b_rptr, data->recv_m->b_rptr, data->recv_left);
      new_recv_m->b_wptr += data->recv_left;
      freemsg(data->recv_m);
      data->recv_m = new_recv_m;
    }
  }

  /* recv into data->recv_m */
  r = (int)recv(fd, (char *)data->recv_m->b_wptr, data->recv_m->b_datap->db_lim-data->recv_m->b_wptr, flags);

#if 0
  ortp_message("ice.c(%x): recv(want: %d: read:%i) recv_left=%i recv_need=%i recv_size=%i", session, data->recv_m->b_datap->db_lim-data->recv_m->b_wptr, r,
    data->recv_left, data->recv_need, data->recv_size);
#endif

#ifdef TRACEFILE
  if (f && data->channelbind_number == 0x4000) fprintf(f, "%s: %lu.%06ld %lu.%06ld recv(%d) returns %d\n", FUNC, tv.tv_sec, tv.tv_usec, diff.tv_sec, diff.tv_usec, data->recv_m->b_datap->db_lim-data->recv_m->b_wptr, r);
#endif
  if (r > 0)  {
    data->recv_m->b_wptr += r;
    data->recv_left += r;
  }

  r = turn_extract_rtp_packet(data, session->rtp_icechecklist, m, from, fromlen);
#ifdef TRACEFILE
  if (f && data->channelbind_number == 0x4000) fprintf(f, "%s: %lu.%06ld %lu.%06ld return %d\n", FUNC, tv.tv_sec, tv.tv_usec, diff.tv_sec, diff.tv_usec, r);
#endif
  if (r > 0) {
    ice_process_stun_message(data, session, session->rtp_icechecklist, 0, m, r, from, *fromlen);
  }
  return r;
}

static int
  rtcp_turn_recvfrom(struct _RtpSession *session, mblk_t *m, int flags, struct sockaddr *from, socklen_t *fromlen)
{
  int error;
  RtpTurnData *_data = (RtpTurnData *)session->rtp.tr->data;
  struct TurnSocketData *data = (struct TurnSocketData *)&_data->rtcp;

  int r;
  int i;
  ortp_socket_t fd = data->fd;

#ifdef TRACEFILE
  if (!f) { f = fopen(TRACEFILE,"w"); }
  if (f && data->channelbind_number == 0x4000) { 
    gettimeofday(&tv,NULL);
    if (last_tv.tv_sec > 0) {
      diff.tv_sec = tv.tv_sec - last_tv.tv_sec;
      diff.tv_usec = tv.tv_usec - last_tv.tv_usec;
      if (diff.tv_usec < 0) {
        diff.tv_usec += 1000000;
        --diff.tv_sec;
      }
    } else {
      diff.tv_sec = diff.tv_usec = 0;
    }
    last_tv.tv_sec = tv.tv_sec;
    last_tv.tv_usec = tv.tv_usec;
  }
#endif

  if (!session->rtcp_mux) {
    uint64_t now = turn_get_cur_time(NULL);
    if (now>_data->rtcp_cur_time+20) {
      _data->rtcp_cur_time = now;
      ice_send_stun_request(data, session, session->rtcp_icechecklist, TRUE, _data->rtcp_cur_time);
    }
  }

  for (i=0;i<10;i++) {
    if (sockisvalid(session->rtcp.socketv4[i])) {
      int bufsz=(int) (m->b_datap->db_lim - m->b_datap->db_base);
      error = (int)recvfrom(session->rtcp.socketv4[i], (char*)m->b_wptr,
        bufsz, 0,
        (struct sockaddr *) from,
        fromlen);
      if (error>0) {
        ice_process_stun_message(data, session, session->rtcp_icechecklist, session->rtcp.socketv4[i], m, error, from, *fromlen);
        return error; /* data available */
      }
    }
  }

  if (data->fd == -1)
    return error;

  *fromlen=0;

  /* Check if we need to refresh, and if so, send the queued requests */
  if (!session->rtcp_mux) 
    turn_session_check_for_refresh(data, session->rtcp_icechecklist);
  turn_flush_sendq(data, 0);

  // First check if we can pull any complete messages out of buffered receive data
  if (data->recv_left > 0) {
    r = turn_extract_rtp_packet(data, session->rtcp_icechecklist, m, from, fromlen);
#ifdef TRACEFILE
    if (f && data->channelbind_number == 0x4000 && r > 0) fprintf(f, "%s: %lu.%06ld %lu.%06ld no-recv-return %d\n", FUNC, tv.tv_sec, tv.tv_usec, diff.tv_sec, diff.tv_usec, r);
#endif
    if (r > 0) {
      ice_process_stun_message(data, session, session->rtcp_icechecklist, 0, m, r, from, *fromlen);
      return r;
    }
  }

  if (!TurnSessionCanRecv(data)) {
    static int val=0;
    if (val%200==0)
      ortp_debug("%s: ignoring read since no TURN channel is active (state = %d)", FUNC, data->state);
    val++;
    return 0;
  }

  // Rewind buffer back to beginning if we don't have any data in it..
  if (data->recv_left == 0)
    data->recv_m->b_rptr = data->recv_m->b_wptr = data->recv_m->b_datap->db_base;

  /* recv into data->recv_m */
  r = (int)recv(fd, (char *)data->recv_m->b_wptr, data->recv_m->b_datap->db_lim-data->recv_m->b_wptr, flags);
#ifdef TRACEFILE
  if (f && data->channelbind_number == 0x4000) fprintf(f, "%s: %lu.%06ld %lu.%06ld recv(%d) returns %d\n", FUNC, tv.tv_sec, tv.tv_usec, diff.tv_sec, diff.tv_usec, data->recv_m->b_datap->db_lim-data->recv_m->b_wptr, r);
#endif
  if (r > 0)  {
    data->recv_m->b_wptr += r;
    data->recv_left += r;
  }

  r = turn_extract_rtp_packet(data, session->rtcp_icechecklist, m, from, fromlen);
#ifdef TRACEFILE
  if (f && data->channelbind_number == 0x4000) fprintf(f, "%s: %lu.%06ld %lu.%06ld return %d\n", FUNC, tv.tv_sec, tv.tv_usec, diff.tv_sec, diff.tv_usec, r);
#endif
  if (r > 0) {
    ice_process_stun_message(data, session, session->rtcp_icechecklist, 0, m, r, from, *fromlen);
  }
  return r;
}

static ortp_socket_t 
  turn_rtp_get_socket(const struct _RtpSession *session)
{
  RtpTurnData *_data = (RtpTurnData *)session->rtp.tr->data;
  struct TurnSocketData *data = (struct TurnSocketData *)&_data->rtp;
  return data->fd;
}

static ortp_socket_t 
  turn_rtcp_get_socket(const struct _RtpSession *session)
{
  RtpTransport *transport = session->rtp.tr;
  RtpTurnData *_data = (RtpTurnData *)transport->data;
  struct TurnSocketData *data = (struct TurnSocketData *)&_data->rtcp;

  return data->fd;
}


static void
  turn_socket_data_init(struct TurnSocketData *data, int srcport)
{
  data->fd = -1;
  data->state = TurnSessionInit;
  qinit(&data->send_q);
  data->send_remain = 0;
  data->recv_m = allocb(4096, 0);
  data->recv_need = 0;
  data->recv_size = 0;
  data->recv_left = 0;
  data->skipping = 0;
  data->have_turn_msg = 0;

  data->bad_nonce_count = 0;

  data->channelbind_number = 0;
  data->max_channelbind_number = 0x4000;

  data->srcport = srcport;
}

static void
  turn_socket_data_destroy(struct TurnSocketData *data)
{
  if (data->fd != -1) {
    int rc = turn_queue_refresh(data, 0);
    if (rc)
      turn_flush_sendq(data, 0);
  }

  if (data->fd != -1) closesocket(data->fd);
  flushq(&data->send_q, FLUSHALL);
  freeb(data->recv_m);
}

static int
  turn_connect_udp(struct TurnSocketData *data, struct addrinfo *relay_server_addr)
{
  struct addrinfo hints;
  int err;
  char strport[8];
  struct addrinfo *res;

  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_flags    = AI_PASSIVE;
  hints.ai_family   = relay_server_addr->ai_family;
  hints.ai_socktype = SOCK_DGRAM;

  snprintf(strport, sizeof(strport), "%i", data->srcport);

  err = getaddrinfo(NULL, strport, &hints, &res);
  if (err <0) {
    ortp_warning("turn_connect_udp: getaddrinfo error");
    goto done;
  }

  // create socket
  data->fd = socket(relay_server_addr->ai_family, SOCK_DGRAM, IPPROTO_UDP);
  if (data->fd < 0) {
    ortp_warning("turn_connect_udp: unable to create socket");
    goto done;
  }

  {
    int optval = 1;
    err = setsockopt (data->fd, SOL_SOCKET, SO_REUSEADDR, (SOCKET_OPTION_VALUE)&optval, sizeof (optval));

#ifdef IPV6_V6ONLY
    if (relay_server_addr->ai_family==AF_INET6) {
      optval = 1;
      setsockopt (data->fd, IPPROTO_IPV6, IPV6_V6ONLY, (SOCKET_OPTION_VALUE) &optval, sizeof (int));
    }
#endif

#if SO_NOSIGPIPE
    optval = 1;
    setsockopt (data->fd, SOL_SOCKET, SO_NOSIGPIPE, (SOCKET_OPTION_VALUE) &optval, sizeof (int));
#endif
  }

  err = bind (data->fd, (struct sockaddr *)res->ai_addr, res->ai_addrlen);
  if (err < 0) {
    ortp_warning("turn_connect_udp: initial bind() failed: %s", getSocketError());

    freeaddrinfo(res);
    snprintf(strport, sizeof(strport), "%i", 0);
    err = getaddrinfo(NULL, strport, &hints, &res);
    if (err <0) {
      ortp_warning("turn_connect_udp: getaddrinfo error (strport=0)");
      closesocket(data->fd);
      data->fd = -1;
      goto done;
    }
    err = bind (data->fd, (struct sockaddr *)res->ai_addr, res->ai_addrlen);
    if (err < 0) {
      ortp_warning("turn_connect_udp: second bind() failed: %s", getSocketError());
      closesocket(data->fd);
      data->fd = -1;
      goto done;
    }
  }

  // connect
  err = connect(data->fd, relay_server_addr->ai_addr, relay_server_addr->ai_addrlen);
  if (err < 0) {
    ortp_warning("turn_connect_udp: connect() failed: %s", getSocketError());
    closesocket(data->fd);
    data->fd = -1;
    goto done;
  }

  set_non_blocking_socket(data->fd);

done:
  if (data->fd == -1) {
    data->state = TurnSessionFailed;
  } else {
    data->state = TurnSessionConnected;
  }

  freeaddrinfo(res);
  data->proto=SOCK_DGRAM;
  if (data->fd < 0) {
    return -1;
  }
  return 0;
}

static int
  turn_connect_tcp(struct TurnSocketData *data, struct addrinfo *relay_server_addr)
{
  struct addrinfo hints;
  int err;
  char strport[8];
  struct addrinfo *res;

  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_flags    = AI_PASSIVE;
  hints.ai_family   = relay_server_addr->ai_family;
  hints.ai_socktype = SOCK_STREAM;

  snprintf(strport, sizeof(strport), "%i", data->srcport);

  err = getaddrinfo(NULL, strport, &hints, &res);
  if (err <0) {
    ortp_warning("turn_connect_tcp: getaddrinfo error");
    goto done;
  }

  // create socket
  data->fd = socket(relay_server_addr->ai_family, SOCK_STREAM, IPPROTO_TCP);
  if (data->fd < 0) {
    ortp_warning("turn_connect_tcp: unable to create socket");
    goto done;
  }

  {
    int optval = 1;
    err = setsockopt (data->fd, SOL_SOCKET, SO_REUSEADDR, (SOCKET_OPTION_VALUE)&optval, sizeof (optval));

#ifdef IPV6_V6ONLY
    if (relay_server_addr->ai_family==AF_INET6) {
      optval = 1;
      setsockopt (data->fd, IPPROTO_IPV6, IPV6_V6ONLY, (SOCKET_OPTION_VALUE) &optval, sizeof (int));
    }
#endif

#if SO_NOSIGPIPE
    optval = 1;
    setsockopt (data->fd, SOL_SOCKET, SO_NOSIGPIPE, (SOCKET_OPTION_VALUE) &optval, sizeof (int));
#endif
  }

  
  err = bind (data->fd, (struct sockaddr *)res->ai_addr, res->ai_addrlen);
  if (err < 0) {
    ortp_warning("turn_connect_tcp: initial bind() failed: %s", getSocketError());

    freeaddrinfo(res);
    snprintf(strport, sizeof(strport), "%i", 0);
    err = getaddrinfo(NULL, strport, &hints, &res);
    if (err <0) {
      ortp_warning("turn_connect_tcp: getaddrinfo error (strport=0)");
      closesocket(data->fd);
      data->fd = -1;
      goto done;
    }
    err = bind (data->fd, (struct sockaddr *)res->ai_addr, res->ai_addrlen);
    if (err < 0) {
      ortp_warning("turn_connect_tcp: second bind() failed: %s", getSocketError());
      closesocket(data->fd);
      data->fd = -1;
      goto done;
    }
  }

  {
    int on = 1;
    unsigned int sndbufsz = 65535; //0x10000;

    if (setsockopt(data->fd, IPPROTO_TCP, TCP_NODELAY, (const char *)&on, sizeof(on)) != 0) {
      ortp_warning("turn_connect_tcp: unable to set TCP_NODELAY on socket");
    }

    err = setsockopt(data->fd, SOL_SOCKET, SO_SNDBUF, (void *)&sndbufsz, sizeof(sndbufsz));
    if (err == -1) {
      ortp_error("turn_connect_tcp: Fail to increase socket's send buffer size with SO_SNDBUF: %s.", getSocketError());
    }
  }

  // connect
  err = connect(data->fd, relay_server_addr->ai_addr, relay_server_addr->ai_addrlen);
  if (err < 0) {
    ortp_warning("turn_connect_tcp: connect() failed: %s", getSocketError());
    closesocket(data->fd);
    data->fd = -1;
    goto done;
  }

  set_non_blocking_socket(data->fd);

done:
  if (data->fd == -1) {
    data->state = TurnSessionFailed;
  } else {
    data->state = TurnSessionConnected;
  }

  freeaddrinfo(res);
  data->proto=SOCK_STREAM;
  if (data->fd<0)
    return -1;
  return 0;
}

RtpTurnData *
  rtp_turn_data_new(const char *relay_server, char *username, char *password, int srcport, bool_t prefer_tcp)
{
  RtpTurnData *data = ortp_new0(RtpTurnData,1);
  TurnAtrReservationToken resToken;
  int ret1;
  int ret2;

  struct addrinfo *addrinfo = NULL;
  struct addrinfo *curinfo;
  struct addrinfo *curinfo_ipv4_udp = NULL;
  struct addrinfo *curinfo_ipv4_tcp = NULL;
  struct addrinfo *curinfo_ipv6_udp = NULL;
  struct addrinfo *curinfo_ipv6_tcp = NULL;
  char host[512];
  int port=3478;
  int counter=0;

  memset(host, 0, sizeof(host));

  turn_socket_data_init(&data->rtp, srcport);
  if (srcport==0)
    turn_socket_data_init(&data->rtcp, 0);
  else
    turn_socket_data_init(&data->rtcp, srcport+1);

  if (username!=NULL) {
    snprintf(data->rtp.auth_username, sizeof(data->rtp.auth_username), "%s", username);
    snprintf(data->rtcp.auth_username, sizeof(data->rtp.auth_username), "%s", username);
  }
  if (password!=NULL) {
    snprintf(data->rtp.auth_password, sizeof(data->rtp.auth_password), "%s", password);
    snprintf(data->rtcp.auth_password, sizeof(data->rtcp.auth_password), "%s", password);
  }

  if (relay_server==NULL)
    return data;


  if (relay_server[0] == '[') {
    int port;
    const char * p = strchr(relay_server, ']');
    size_t len = p ? (size_t) (p - relay_server - 1): 0;
    if (p==NULL || len>sizeof(host))
      return data;
    memcpy(host, relay_server + 1, len);

    p++;
    if (p[0]==':') {
      p++;
      port = atoi(p);
      if (port>65535 || port<0)
        port=3478;
    }
  } else {
    const char * p = strchr(relay_server, ':');
    size_t len = p ? (size_t) (p - relay_server): strlen(relay_server);
    if (len>sizeof(host))
      return data;
    memcpy(host, relay_server, len);
    if (p!=NULL) {
      p++;
      port = atoi(p);
      if (port>65535 || port<0)
        port=3478;
    }
  }

  {
    int res;
    res = stunGetaddrinfo (&addrinfo, host, port, 0, PF_UNSPEC);
    if (res)
      return data;

    for (curinfo = addrinfo; curinfo; curinfo = curinfo->ai_next) {

      if (curinfo->ai_protocol && curinfo->ai_protocol != IPPROTO_UDP && curinfo->ai_protocol != IPPROTO_TCP) {
        continue;
      }
      if (curinfo->ai_family != AF_INET && curinfo->ai_family != AF_INET6) {
        continue;
      }
      if (curinfo->ai_family == AF_INET) {
        curinfo_ipv4_udp = curinfo;
        counter++;
      }
      if (curinfo->ai_family == AF_INET) {
        curinfo_ipv4_tcp = curinfo;
        counter++;
      }
      if (curinfo->ai_family == AF_INET6) {
        curinfo_ipv6_udp = curinfo;
        counter++;
      }
      if (curinfo->ai_family == AF_INET6) {
        curinfo_ipv6_tcp = curinfo;
        counter++;
      }
    }
  }

  //TEST
  //curinfo_ipv4_udp=NULL;
  //curinfo_ipv4_tcp=NULL;
  //TEST
  
  /* if ipv6 only, then use it */
  if (curinfo_ipv4_udp==NULL && curinfo_ipv6_udp!=NULL) {
    curinfo_ipv4_udp=curinfo_ipv6_udp;
  }
  if (curinfo_ipv4_tcp==NULL && curinfo_ipv6_tcp!=NULL) {
    curinfo_ipv4_tcp=curinfo_ipv6_tcp;
  }

  //TEST TCP
  //prefer_tcp=TRUE;
  //TEST
  
  if (prefer_tcp==FALSE && curinfo_ipv4_udp!=NULL) {
    ret1 = turn_connect_udp(&data->rtp, curinfo_ipv4_udp);
    
    if (ret1!=0) {
      if (curinfo_ipv4_udp!=curinfo_ipv6_udp && curinfo_ipv4_udp!=NULL && curinfo_ipv6_udp!=NULL) {
        /* IPv4 failed, try IPv6 */
        curinfo_ipv4_udp=curinfo_ipv6_udp;
      }
      if (curinfo_ipv4_tcp!=curinfo_ipv6_tcp && curinfo_ipv4_tcp!=NULL && curinfo_ipv6_tcp!=NULL) {
        /* IPv4 failed, try IPv6 */
        curinfo_ipv4_tcp=curinfo_ipv6_tcp;
      }
      ret1 = turn_connect_udp(&data->rtp, curinfo_ipv4_udp);
    }
    
    ret2 = 0;
    if (ret1==0)
      ret2 = turn_connect_udp(&data->rtcp, curinfo_ipv4_udp);
    
    if (ret1!=0 || ret2!=0) {
      if (data->rtp.fd>0) closesocket(data->rtp.fd);
      data->rtp.fd = -1;
      data->rtp.state = TurnSessionFailed;
      if (data->rtcp.fd>0) closesocket(data->rtcp.fd);
      data->rtcp.fd = -1;
      data->rtcp.state = TurnSessionFailed;
    }

    if (ret1==0 && ret2==0) {
      ret1 = turn_allocate(&data->rtp, &resToken, TRUE, curinfo_ipv4_udp->ai_family==AF_INET6?0x02:0x01);
      if (ret1 == 0)
        ret2 = turn_allocate(&data->rtcp, &resToken, FALSE, curinfo_ipv4_udp->ai_family==AF_INET6?0x02:0x01);

      if ( ret1 == 0 && ret2 == 0 ) {
        data->rtp.use_relay=1;
        data->rtcp.use_relay=1;

        /* revert srcport to REAL local port: it must be verified by application after success */
        {
          struct sockaddr_storage ai_addr;
          socklen_t len;
          len = sizeof (struct sockaddr_storage);
          
          getsockname (data->rtp.fd, (struct sockaddr *) &ai_addr, &len);
          if (curinfo_ipv4_udp->ai_family==AF_INET)
            data->rtp.srcport = ntohs (((struct sockaddr_in *) &ai_addr)->sin_port);
          else
            data->rtp.srcport = ntohs (((struct sockaddr_in6 *) &ai_addr)->sin6_port);

          getsockname (data->rtcp.fd, (struct sockaddr *) &ai_addr, &len);
          if (curinfo_ipv4_udp->ai_family==AF_INET)
            data->rtcp.srcport = ntohs (((struct sockaddr_in *) &ai_addr)->sin_port);
          else
            data->rtcp.srcport = ntohs (((struct sockaddr_in6 *) &ai_addr)->sin6_port);
        }
        
        ortp_message("rtp_turn_data_new: local ports (TURN/UDP/%s) -> rtp=%i rtcp=%i", (curinfo_ipv4_udp->ai_family==AF_INET)?"IPv4":"IPv6", data->rtp.srcport, data->rtcp.srcport);
        
        freeaddrinfo(addrinfo);
        return data;
      }

      if (ret1!=0 || ret2!=0) {
        if (data->rtp.fd>0) closesocket(data->rtp.fd);
        data->rtp.fd = -1;
        data->rtp.state = TurnSessionFailed;
        if (data->rtcp.fd>0) closesocket(data->rtcp.fd);
        data->rtcp.fd = -1;
        data->rtcp.state = TurnSessionFailed;
      }
    }
  }

  /* revert to TCP */
  if (curinfo_ipv4_tcp!=NULL) {
    ret1 = turn_connect_tcp(&data->rtp, curinfo_ipv4_tcp);

    if (ret1!=0) {
      if (curinfo_ipv4_udp!=curinfo_ipv6_udp && curinfo_ipv4_udp!=NULL && curinfo_ipv6_udp!=NULL) {
        /* IPv4 failed, try IPv6 */
        curinfo_ipv4_udp=curinfo_ipv6_udp;
      }
      if (curinfo_ipv4_tcp!=curinfo_ipv6_tcp && curinfo_ipv4_tcp!=NULL && curinfo_ipv6_tcp!=NULL) {
        /* IPv4 failed, try IPv6 */
        curinfo_ipv4_tcp=curinfo_ipv6_tcp;
      }
      ret1 = turn_connect_tcp(&data->rtp, curinfo_ipv4_tcp);
    }
    
    ret2 = 0;
    if (ret1==0)
      ret2 = turn_connect_tcp(&data->rtcp, curinfo_ipv4_tcp);

    if (ret1!=0 || ret2!=0) {
      if (data->rtp.fd>0) closesocket(data->rtp.fd);
      data->rtp.fd = -1;
      data->rtp.state = TurnSessionFailed;
      if (data->rtcp.fd>0) closesocket(data->rtcp.fd);
      data->rtcp.fd = -1;
      data->rtcp.state = TurnSessionFailed;
    }

    if (ret1==0 && ret2==0) {
      ret1 = turn_allocate(&data->rtp, &resToken, TRUE, curinfo_ipv4_udp->ai_family==AF_INET6?0x02:0x01);
      if (ret1 == 0)
        ret2 = turn_allocate(&data->rtcp, &resToken, FALSE, curinfo_ipv4_udp->ai_family==AF_INET6?0x02:0x01);

      if ( ret1 == 0 && ret2 == 0 ) {
        data->rtp.use_relay=1;
        data->rtcp.use_relay=1;
        
        /* revert srcport to REAL local port: it must be verified by application after success */
        {
          struct sockaddr_storage ai_addr;
          socklen_t len;
          len = sizeof (struct sockaddr_storage);
          
          getsockname (data->rtp.fd, (struct sockaddr *) &ai_addr, &len);
          if (curinfo_ipv4_udp->ai_family==AF_INET)
            data->rtp.srcport = ntohs (((struct sockaddr_in *) &ai_addr)->sin_port);
          else
            data->rtp.srcport = ntohs (((struct sockaddr_in6 *) &ai_addr)->sin6_port);
          
          getsockname (data->rtcp.fd, (struct sockaddr *) &ai_addr, &len);
          if (curinfo_ipv4_udp->ai_family==AF_INET)
            data->rtcp.srcport = ntohs (((struct sockaddr_in *) &ai_addr)->sin_port);
          else
            data->rtcp.srcport = ntohs (((struct sockaddr_in6 *) &ai_addr)->sin6_port);
        }
        
        ortp_message("rtp_turn_data_new: local ports (TURN/TCP/%s) -> rtp=%i rtcp=%i", (curinfo_ipv4_tcp->ai_family==AF_INET)?"IPv4":"IPv6", data->rtp.srcport, data->rtcp.srcport);
        
        freeaddrinfo(addrinfo);
        return data;
      }

      if (ret1!=0 || ret2!=0) {
        if (data->rtp.fd>0) closesocket(data->rtp.fd);
        data->rtp.fd = -1;
        data->rtp.state = TurnSessionFailed;
        if (data->rtcp.fd>0) closesocket(data->rtcp.fd);
        data->rtcp.fd = -1;
        data->rtcp.state = TurnSessionFailed;
      }
    }
  }

  rtp_turn_data_destroy(data);
  freeaddrinfo(addrinfo);
  ortp_error("ice: cannot allocate TURN channel using UDP and TCP");
  return NULL;
}

void
  rtp_turn_data_destroy(RtpTurnData *data)
{
  turn_socket_data_destroy(&data->rtp);
  turn_socket_data_destroy(&data->rtcp);
  ortp_free(data);
}

static void
  turn_destroy(struct _RtpTransport *transport)
{
  transport->data=NULL;
}

static int turn_global_init(void)
{
  return 0;
}

static int turn_global_shutdown(void)
{
  return 0;
}

#ifdef __cplusplus
extern "C" int turn_transport_setup(void);
#else
int turn_transport_setup(void);
#endif

int turn_transport_setup(void)
{
  RtpTransport *tr = ortp_new0(RtpTransport,1);
  snprintf(tr->name, sizeof(tr->name), "TURN");

  tr->t_global_init = turn_global_init;
  tr->t_global_shutdown = turn_global_shutdown;
  tr->t_set_option = NULL;

  tr->t_rtp_getsocket=turn_rtp_get_socket;
  tr->t_rtcp_getsocket=turn_rtcp_get_socket;

  tr->t_rtp_sendto=rtp_turn_sendto;
  tr->t_rtp_recvfrom=rtp_turn_recvfrom;
  tr->t_rtcp_sendto=rtcp_turn_sendto;
  tr->t_rtcp_recvfrom=rtcp_turn_recvfrom;

  tr->t_rtp_refresh=rtp_refresh_turn;
  tr->t_rtcp_refresh=rtcp_refresh_turn;

  tr->t_destroy=turn_destroy;

  ortp_transport_add(tr);

  return 0;
}

int
  rtp_turn_get_relay_addr(RtpTurnData *data, StunIPAddress *relay_addr)
{
  if (data && data->rtp.relay_addr.addr.ipv4.port > 0) {
    memcpy(relay_addr, &data->rtp.relay_addr, sizeof(*relay_addr));
    return 0;
  } else {
    ortp_warning("%s: no relayed rtp port", FUNC);
    return -1;
  }
}

static int
  sendTurnMessageTcp(ortp_socket_t fd, const char* buf, unsigned int len)
{
  int n, flags = 0;
  const char* bufptr = buf;
  int buflen = len, bufsent = 0;

  do {
    n = (int)send(fd, bufptr, buflen, flags);
    if (n < 0) {
      int errnum = getSocketErrorCode();
      if (!is_would_block_error(errnum)) {
        ortp_warning("%s: socket error with %d/%d bytes sent: %s",
          FUNC, n, buflen, getSocketError());
        return -1;
      }

      /* Buffer below */
      n = 0;
    }

    bufptr += n;
    bufsent += n;
    buflen -= n;
  } while (bufsent < len);

  return bufsent;
}

static int
  readTurnMessageUdp(ortp_socket_t fd, char* buf_input, unsigned int buflen_input, char* buf_output, unsigned int buflen_output)
{
  int flags = 0;
  int r;
  uint16_t mtype, mlen;
  int nextlen;

  struct timeval tv;
  fd_set fdSet; 
  int err;

  int restransmit=100*1000;

  ortp_socket_t fdSetSize;

  if (buflen_output < sizeof(uint32_t)) {
    ortp_warning("readTurnMessageUdp: buf size too small for Stun header!");
    return -1;
  }

again:

  FD_ZERO(&fdSet);
  fdSetSize=0;
  FD_SET(fd,&fdSet);
  fdSetSize = (fd+1>fdSetSize) ? fd+1 : fdSetSize;
  tv.tv_sec=0;
  tv.tv_usec=restransmit; /* 100 ms * restransmit */
  err = select(fdSetSize, &fdSet, NULL, NULL, &tv);
  if ( err == SOCKET_ERROR ) {
    ortp_warning("readTurnMessageUdp: select error: %s", getSocketError());
    return -1;
  }
  if (err==0) {
    /* resend? */
    if (restransmit==800000) { //100+200+400+800=1500
      ortp_warning("readTurnMessageUdp: no answer for TURN Allocate");
      return -1;
    }
    err = sendTurnMessageTcp(fd, buf_input, buflen_input);
    if (err < buflen_input) {
      ortp_warning("readTurnMessageUdp: cant send retransmission");
      return -1;
    }
    restransmit=restransmit<<1;
    goto again;
  }
  r = (int)recv(fd, buf_output, buflen_output, flags);
  if (r < 0) {
    int errnum = getSocketErrorCode();
    if (!is_would_block_error(errnum)) {
      ortp_warning("readTurnMessageUdp: recv failure: %s", getSocketError());
      goto again;
    }
    ortp_warning("readTurnMessageUdp: recv failure: %s", getSocketError());
    return -1;
  }
  if (r == 0) {
    ortp_warning("readTurnMessageUdp: recv failure: timeout");
    return -1;
  }

  memcpy(&mtype, buf_output, sizeof(uint16_t));
  mtype = ntohs(mtype);
  memcpy(&mlen, buf_output + sizeof(uint16_t), sizeof(uint16_t));
  mlen = ntohs(mlen);

  if ((mtype & 0xc000) == 0) {
    /* Standard STUN message */
    nextlen = mlen + sizeof(StunMsgHdr) - sizeof(uint32_t);
  } else if ((mtype & 0xc000) == 0x4000) {
    /* ChanData message */
    nextlen = ((mlen + sizeof(uint32_t) + 3) & ~3) - sizeof(uint32_t);
  } else {
    /* neither... need to re-synch */
    ortp_warning("readTurnMessageUdp: unexpected mtype %x!", mtype);
    return -1;
  }

  if (buflen_output < nextlen+sizeof(uint32_t)) {
    ortp_warning("readTurnMessageUdp: buf size too small for Stun message!");
    return -1;
  }

  return r;
}

static int
  readTurnMessageTcp(ortp_socket_t fd, char* buf, unsigned int buflen)
{
  int r;
  int flags = 0;
  int hdrlen, nextlen, numread = 0;
  char* bufptr = buf;
  uint16_t mtype, mlen;
  int counter=0;

  hdrlen = sizeof(uint32_t);
  if (buflen < hdrlen) {
    ortp_warning("readTurnMessageTcp: buf size too small for Stun header!");
    return -1;
  }

  do {
    struct timeval tv;
    fd_set fdSet;
    int err;
    ortp_socket_t fdSetSize;
    
    FD_ZERO(&fdSet);
    fdSetSize=0;
    FD_SET(fd,&fdSet);
    fdSetSize = (fd+1>fdSetSize) ? fd+1 : fdSetSize;
    tv.tv_sec=2;
    tv.tv_usec=0; /* 2sec? */
    err = select(fdSetSize, &fdSet, NULL, NULL, &tv);
    if ( err == SOCKET_ERROR ) {
      ortp_warning("readTurnMessageTcp: select error: %s", getSocketError());
      return -1;
    }
    if (err==0) {
      /* timeout */
      ortp_warning("readTurnMessageTcp: timeout for TCP relay (no data)");
      return -1;
    }
    r = (int)recv(fd, bufptr, hdrlen - numread, flags);
    if (r < 0) {
      int errnum = getSocketErrorCode();
      if (!is_would_block_error(errnum)) {
        ortp_warning("%s: socket error with %d/%d bytes read: %s",
          FUNC, numread, hdrlen, getSocketError());
        return -1;
      }
      counter++;
      ortp_warning("counter: readTurnMessageTcp: %i %i", numread, counter);
    } else {
      numread += r;
      bufptr += r;
      buflen -= r;
    }
  } while (numread < hdrlen);

  // XXXrkb: use ChanDataHdr here?
  memcpy(&mtype, buf, sizeof(uint16_t));
  mtype = ntohs(mtype);
  memcpy(&mlen, buf + sizeof(uint16_t), sizeof(uint16_t));
  mlen = ntohs(mlen);

  if ((mtype & 0xc000) == 0) {
    /* Standard STUN message */
    nextlen = mlen + sizeof(StunMsgHdr) - numread;
  } else if ((mtype & 0xc000) == 0x4000) {
    /* ChanData message */
    nextlen = ((mlen + sizeof(uint32_t) + 3) & ~3) - numread;
  } else {
    /* neither... need to re-synch */
    ortp_warning("readTurnMessageTcp: unexpected mtype %x!", mtype);
    return -1;
  }

  /* Got everything we needed... must have been a 0-length ChanData msg */
  if (nextlen == 0) 
    return numread;

  if (buflen < nextlen+numread) {
    ortp_warning("readTurnMessageTcp: buf size too small for Stun message!");
    return -1;
  }

  do {
    struct timeval tv;
    fd_set fdSet;
    int err;
    ortp_socket_t fdSetSize;
    
    FD_ZERO(&fdSet);
    fdSetSize=0;
    FD_SET(fd,&fdSet);
    fdSetSize = (fd+1>fdSetSize) ? fd+1 : fdSetSize;
    tv.tv_sec=2;
    tv.tv_usec=0; /* 2sec? */
    err = select(fdSetSize, &fdSet, NULL, NULL, &tv);
    if ( err == SOCKET_ERROR ) {
      ortp_warning("readTurnMessageTcp: select error: %s", getSocketError());
      return -1;
    }
    if (err==0) {
      /* timeout */
      ortp_warning("readTurnMessageTcp: timeout for TCP relay (non complete data)");
      return -1;
    }
    
    r = (int)recv(fd, bufptr, nextlen, flags);
    if (r < 0) {
      int errnum = getSocketErrorCode();
      if (!is_would_block_error(errnum)) {
        ortp_warning("%s: socket error with %d/%d bytes read: %s",
          FUNC, numread, hdrlen, getSocketError());
        return -1;
      }

      r = 0;
    }

    numread += r;
    bufptr += r;
    nextlen -= r;
  } while (nextlen > 0);

  return numread;
}

static int 
  turn_allocate(struct TurnSocketData *data, TurnAtrReservationToken* resToken, bool_t reservePortPair, int requestIpv6)
{
  int slen;
  StunMessage resp;
  char buf_input[STUN_MAX_MESSAGE_SIZE];
  char buf_output[STUN_MAX_MESSAGE_SIZE];
  int len = STUN_MAX_MESSAGE_SIZE;
  StunAtrString username, *pUsername = NULL;
  StunAtrString password, *pPassword = NULL;
  StunAtrString realm, *pRealm = NULL;
  StunAtrString nonce, *pNonce = NULL;
  int mismatch=0; /* old allocation is not yet closed... */
  StunTrId out_tr_id;

  if (data == NULL) {
    ortp_warning("%s: no TURN socket data", FUNC);
    return -1;
  }

  if (data->auth_username[0] != '\0') {
    snprintf(username.value, sizeof(username.value), "%s", data->auth_username);
    username.sizeValue = (uint16_t)strlen(username.value);
    pUsername = &username;
  }

  if (data->auth_password[0] != '\0') {
    snprintf(password.value, sizeof(password.value), "%s", data->auth_password);
    password.sizeValue = (uint16_t)strlen(password.value);
    pPassword = &password;
  }

again:
  if (data->auth_realm[0] != '\0') {
    snprintf(realm.value, sizeof(realm.value), "%s", data->auth_realm);
    realm.sizeValue = (uint16_t)strlen(realm.value);
    pRealm = &realm;
  }

  if (data->auth_nonce[0] != '\0') {
    snprintf(nonce.value, sizeof(nonce.value), "%s", data->auth_nonce);
    nonce.sizeValue = (uint16_t)strlen(nonce.value);
    pNonce = &nonce;
  }

  len = turnBuildAllocate(buf_input, sizeof(buf_input), pUsername, pPassword, pRealm, pNonce, resToken, reservePortPair, requestIpv6 , &out_tr_id);
  ortp_debug("%s: About to send msg of len %i to fd %d", FUNC, len, data->fd);

  slen = sendTurnMessageTcp(data->fd, buf_input, len);
  if (slen < len) {
    data->state = TurnSessionFailed;
    return -1;
  }

  data->state = TurnSessionSentAllocate;

readagain:
  if (data->proto==SOCK_DGRAM)
    len = readTurnMessageUdp(data->fd, buf_input, slen, buf_output, sizeof(buf_output));
  else
    len = readTurnMessageTcp(data->fd, buf_output, sizeof(buf_output));
  ortp_debug("%s: Read msg of len %i from fd %d", FUNC, len, data->fd);

  if (len<=0) {
    data->state = TurnSessionFailed;
    return -1;
  }

  if (!stunParseMessage(buf_output, len, &resp)) {
    ortp_warning("%s: read=%d // stunParseMessage() failed", FUNC, len);
    data->state = TurnSessionFailed;
    return -1;
  }

  if (memcmp(&out_tr_id, &resp.msgHdr.tr_id, sizeof(StunTrId))!=0) {
    ortp_warning("%s: skip STUN answer (probably restransmission) id=%i", FUNC, (int)(resp.msgHdr.tr_id.octet[0]));
    goto readagain;
  }
      
  if (STUN_IS_ERR_RESP(resp.msgHdr.msgType))
  {
    int err = -1;

    if (resp.hasErrorCode==TRUE)
      err = (100 * resp.errorCode.errorClass) + resp.errorCode.number;

    /* check if we need to authenticate */
    if (err == 401 && resp.hasNonce == TRUE && resp.hasRealm == TRUE)
    {
      ortp_warning("%s: got err %d in msg type %x on fd %d", FUNC, err, resp.msgHdr.msgType, data->fd);

      /* If we have realm already, this is a bad-password failure */
      if (data->auth_realm[0] != '\0') {
        ortp_warning("%s: got 401 after auth, failing", FUNC);
        return -1;
      }

      memcpy(data->auth_nonce, resp.nonceName.value, resp.nonceName.sizeValue);
      data->auth_nonce[resp.nonceName.sizeValue] = '\0';

      memcpy(data->auth_realm, resp.realmName.value, resp.realmName.sizeValue);
      data->auth_realm[resp.realmName.sizeValue] = '\0';

      data->bad_nonce_count = 0;
      goto again;
    }

    if (err == 438 && resp.hasNonce == TRUE && data->bad_nonce_count == 0) {
      ortp_warning("%s: got err %d in msg type %x on fd %d", FUNC, err, resp.msgHdr.msgType, data->fd);

      memcpy(data->auth_nonce, resp.nonceName.value, resp.nonceName.sizeValue);
      data->auth_nonce[resp.nonceName.sizeValue] = '\0';
      data->bad_nonce_count++;
      goto again;
    }

    if (err == 437 && data->bad_nonce_count == 0) {
      StunTrId out_tr_id;
      ortp_warning("%s: got err %d in msg type %x on fd %d", FUNC, err, resp.msgHdr.msgType, data->fd);

      if (mismatch==1) {
        ortp_warning("%s: allocation mismatch // previous allocation still exists", FUNC);
        data->state = TurnSessionFailed;
        return -1;
      }
      mismatch++;
refreshagain:
      if (data->auth_realm[0] != '\0') {
        snprintf(realm.value, sizeof(realm.value), "%s", data->auth_realm);
        realm.sizeValue = (uint16_t)strlen(realm.value);
        pRealm = &realm;
      }

      if (data->auth_nonce[0] != '\0') {
        snprintf(nonce.value, sizeof(nonce.value), "%s", data->auth_nonce);
        nonce.sizeValue = (uint16_t)strlen(nonce.value);
        pNonce = &nonce;
      }

      len = turnBuildRefresh(buf_input, sizeof(buf_input), pUsername, pPassword, pRealm, pNonce, 0, &out_tr_id);
      ortp_debug("%s: About to send msg of len %i to fd %d", FUNC, len, data->fd);

      slen = sendTurnMessageTcp(data->fd, buf_input, len);
      if (slen < len) {
        data->state = TurnSessionFailed;
        return -1;
      }

      data->state = TurnSessionSentRefresh;

    refreshreadagain:
      if (data->proto==SOCK_DGRAM)
        len = readTurnMessageUdp(data->fd, buf_input, slen, buf_output, sizeof(buf_output));
      else
        len = readTurnMessageTcp(data->fd, buf_output, sizeof(buf_output));
      ortp_debug("%s: Read msg of len %i from fd %d", FUNC, len, data->fd);

      if (len<=0) {
        data->state = TurnSessionFailed;
        return -1;
      }

      if (!stunParseMessage(buf_output, len, &resp)) {
        ortp_warning("%s: read=%d // stunParseMessage() failed", FUNC, len);
        data->state = TurnSessionFailed;
        return -1;
      }

      if (memcmp(&out_tr_id, &resp.msgHdr.tr_id, sizeof(StunTrId))!=0) {
        ortp_warning("%s: skip STUN answer (probably restransmission) id=%i", FUNC, (int)(resp.msgHdr.tr_id.octet[0]));
        goto refreshreadagain;
      }

      if (STUN_IS_ERR_RESP(resp.msgHdr.msgType)) {
        if (resp.hasErrorCode==TRUE)
          err = (100 * resp.errorCode.errorClass) + resp.errorCode.number;

        ortp_warning("%s: error %d from refresh on fd %d with lifetime 600", FUNC, err, data->fd);

        /* check if we need to authenticate */
        if (err == 401 && resp.hasNonce == TRUE && resp.hasRealm == TRUE)
        {
          ortp_warning("%s: got err %d in msg type %x on fd %d", FUNC, err, resp.msgHdr.msgType, data->fd);

          /* If we have realm already, this is a bad-password failure */
          if (data->auth_realm[0] != '\0') {
            ortp_warning("%s: got 401 after auth, failing", FUNC);
            return -1;
          }

          memcpy(data->auth_nonce, resp.nonceName.value, resp.nonceName.sizeValue);
          data->auth_nonce[resp.nonceName.sizeValue] = '\0';

          memcpy(data->auth_realm, resp.realmName.value, resp.realmName.sizeValue);
          data->auth_realm[resp.realmName.sizeValue] = '\0';

          data->bad_nonce_count = 0;
          goto refreshagain;
        }

      }
      else if (STUN_IS_SUCCESS_RESP(resp.msgHdr.msgType)) {

        ortp_warning("%s: got success with for refresh with lifetime 600", FUNC);
        goto again;
      }
      return -1;
    }

    ortp_warning("%s: error %d from allocate on fd %d", FUNC, err, data->fd);
    data->state = TurnSessionFailed;
    return -1;
  }
  else if (STUN_IS_SUCCESS_RESP(resp.msgHdr.msgType))
  {
    if (!resp.hasXorRelayedAddress || 
      (reservePortPair && !resp.hasReservationToken)) {
        ortp_warning("%s: got success BUT without required attributes (relayed addr, reservation token, lifetime)",  FUNC);
        data->state = TurnSessionFailed;
        return -1;
    }

    if (resp.hasXorMappedAddress) {
      stunDecodeXorAddress(&resp.xorMappedAddress, &data->mapped_addr, &resp.msgHdr.tr_id);
    }

    /* If we don't get lifetime stick in default lifetime and try and go on */
    if (!resp.hasLifetimeAttributes) {
      ortp_warning("%s: got Allocate success WITHOUT lifetime!");
      resp.lifetimeAttributes.lifetime = TURN_DEFAULT_ALLOCATION_LIFETIME;
    }

    if (reservePortPair) {
      memcpy(resToken, &resp.reservationToken, sizeof(resp.reservationToken));
    }

    data->bad_nonce_count = 0;

    stunDecodeXorAddress(&resp.xorRelayedAddress, &data->relay_addr, &resp.msgHdr.tr_id);

#if defined(_WIN32_WCE)
    data->alloc_expires = timeGetTime()/1000 + resp.lifetimeAttributes.lifetime;
#else
    data->alloc_expires = time(NULL) + resp.lifetimeAttributes.lifetime;
#endif
    ortp_warning("%s: got success with XorRelayedAddress, port %d, lifetime %d", FUNC, data->relay_addr.addr.ipv4.port, resp.lifetimeAttributes.lifetime);
    return 0;
  }

  ortp_warning("%s: unknown STUN type %x", FUNC, resp.msgHdr.msgType);
  data->state = TurnSessionFailed;
  return -1;
}

int
  turnBuildAllocate(char* buf, int buf_size,
  const StunAtrString *username, const StunAtrString *password,
  const StunAtrString *realm, const StunAtrString *nonce,
  const TurnAtrReservationToken *resToken, bool_t reservePortPair, int requestIpv6, StunTrId *out_tr_id)
{
  StunMessage req;
  const char serverName[] = "aoRTP  " STUN_VERSION; /* must pad to mult of 4 */

  memset(&req, 0, sizeof(StunMessage));

  stunBuildReqSimple( &req, username, FALSE , FALSE , 0 );
  memcpy(out_tr_id, &req.msgHdr.tr_id, sizeof(StunTrId));
  req.msgHdr.msgType = (TURN_MEDHOD_ALLOCATE|STUN_REQUEST);

  req.hasSoftware = TRUE;
  memcpy( req.softwareName.value, serverName, sizeof(serverName));
  req.softwareName.sizeValue = sizeof(serverName);

  req.hasRequestedTransport = TRUE;
  memset(&req.requestedTransport, 0, sizeof(req.requestedTransport));
  req.requestedTransport.proto = IPPROTO_UDP;

  req.hasLifetimeAttributes = TRUE;
  req.lifetimeAttributes.lifetime = TURN_DEFAULT_ALLOCATION_LIFETIME;
  /* req.hasDontFragment = TRUE; */ /* interop issue with restund */

  /* 
  * NB: EVEN-PORT (which besides forcing an even port to be returned also
  * allocates the following port and returns us a grant to alocate it via
  * the returned reservation token) and use of the reservationt token are
  * mutually exclusive.  We can either use the reservation token from a
  * prior EVEN-PORT allocation or we can request a new allocation (with or
  * without the EVEN-PORT attribute).
  */
  if (reservePortPair) 
  {
    req.hasEvenPort = TRUE;
    req.evenPort.value = ReserveNextPort;
  }
  else if (resToken != NULL) 
  {
    req.hasReservationToken = TRUE;
    memcpy(&req.reservationToken, resToken, sizeof(*resToken));
  }

  if (req.hasReservationToken!=TRUE && requestIpv6==0x02) {
    req.hasRequestAddressFamily = TRUE;
    req.requestAddressFamily.family=0x2;
    req.requestAddressFamily.family=0x1;
  } else {
#if 0
    /* TODO? */
    req.hasRequestAddressFamily = TRUE;
    req.requestAddressFamily.family=0x1
#endif
  }

  if (username!=NULL && username->sizeValue>0
    && password!=NULL && password->sizeValue>0
    && realm!=NULL && realm->sizeValue>0
    && nonce!=NULL && nonce->sizeValue>0)
  {
    req.hasUsername = TRUE;
    memcpy( req.username.value, username->value, username->sizeValue );
    req.username.sizeValue = username->sizeValue;

    req.hasRealm = TRUE;
    memcpy( req.realmName.value, realm->value, realm->sizeValue );
    req.realmName.sizeValue = realm->sizeValue;

    req.hasNonce = TRUE;
    memcpy( req.nonceName.value, nonce->value, nonce->sizeValue );
    req.nonceName.sizeValue = nonce->sizeValue;

    req.hasMessageIntegrity = TRUE;
  }

  return stunEncodeMessage( &req, buf, buf_size, password );
}

int 
  turnBuildChannelBindOrPermission(int use_channelbind, char* buf, int buf_size,
  const StunAtrString *username, const StunAtrString *password,
  unsigned int chanNumber, const StunIPAddress *peerAddress,
  const StunAtrString *realm, const StunAtrString *nonce, StunTrId *tid)
{ 
  StunMessage req;
  const char serverName[] = "oRTP   " STUN_VERSION; /* must pad to mult of 4 */

  memset(&req, 0, sizeof(StunMessage));

  stunBuildReqSimple( &req,  NULL, FALSE, FALSE, 0 );
  if (use_channelbind)
    req.msgHdr.msgType = (TURN_METHOD_CHANNELBIND|STUN_REQUEST);
  else
    req.msgHdr.msgType = (TURN_METHOD_CREATEPERMISSION|STUN_REQUEST);

  if (use_channelbind) {
    memset(&req.channelNumberAttributes, 0, sizeof(TurnAtrChannelNumber));
    req.channelNumberAttributes.channelNumber = chanNumber;
    req.hasChannelNumberAttributes = TRUE;
  }

  req.hasSoftware = TRUE;
  memcpy( req.softwareName.value, serverName, sizeof(serverName));
  req.softwareName.sizeValue = sizeof(serverName);

  req.hasXorPeerAddress = TRUE;
  stunEncodeXorAddress(peerAddress, &req.xorPeerAddress, &req.msgHdr.tr_id);

  if (username!=NULL && username->sizeValue>0
    && password!=NULL && password->sizeValue>0
    && realm!=NULL && realm->sizeValue>0
    && nonce!=NULL && nonce->sizeValue>0)
  {
    req.hasUsername = TRUE;
    memcpy( req.username.value, username->value, username->sizeValue );
    req.username.sizeValue = username->sizeValue;

    req.hasRealm = TRUE;
    memcpy( req.realmName.value, realm->value, realm->sizeValue );
    req.realmName.sizeValue = realm->sizeValue;

    req.hasNonce = TRUE;
    memcpy( req.nonceName.value, nonce->value, nonce->sizeValue );
    req.nonceName.sizeValue = nonce->sizeValue;

    req.hasMessageIntegrity = TRUE;
  }

  memcpy(tid , &(req.msgHdr.tr_id), sizeof(req.msgHdr.tr_id));
  return stunEncodeMessage( &req, buf, buf_size, password );
}

int 
  turnBuildRefresh(char* buf, int buf_size,
  const StunAtrString *username, const StunAtrString *password,
  const StunAtrString *realm, const StunAtrString *nonce, int lifetime, StunTrId *tid)
{ 
  StunMessage req;
  const char serverName[] = "oRTP   " STUN_VERSION; /* must pad to mult of 4 */

  memset(&req, 0, sizeof(StunMessage));

  stunBuildReqSimple( &req,  NULL, FALSE, FALSE, 0 );
  req.msgHdr.msgType = (TURN_METHOD_REFRESH|STUN_REQUEST);

  req.hasSoftware = TRUE;
  memcpy( req.softwareName.value, serverName, sizeof(serverName));
  req.softwareName.sizeValue = sizeof(serverName);

  req.hasLifetimeAttributes = TRUE;
  req.lifetimeAttributes.lifetime = lifetime;

  if (username!=NULL && username->sizeValue>0
    && password!=NULL && password->sizeValue>0
    && realm!=NULL && realm->sizeValue>0
    && nonce!=NULL && nonce->sizeValue>0)
  {
    req.hasUsername = TRUE;
    memcpy( req.username.value, username->value, username->sizeValue );
    req.username.sizeValue = username->sizeValue;

    req.hasRealm = TRUE;
    memcpy( req.realmName.value, realm->value, realm->sizeValue );
    req.realmName.sizeValue = realm->sizeValue;

    req.hasNonce = TRUE;
    memcpy( req.nonceName.value, nonce->value, nonce->sizeValue );
    req.nonceName.sizeValue = nonce->sizeValue;

    req.hasMessageIntegrity = TRUE;
  }

  memcpy(tid , &(req.msgHdr.tr_id), sizeof(req.msgHdr.tr_id));
  return stunEncodeMessage( &req, buf, buf_size, password );
}

